# coding=utf-8

import pygraphviz as pgv
import numpy as np
import matplotlib.pyplot as plt
from PIL import Image

import json
import os
import random
import sys
from matplotlib2tikz import save as tikz_save

W_SCALE_FACTOR = 10
H_SCALE_FACTOR = 20

COLORS = 768
SHUFFLE_COLORS = False
TIKZ_SAVE = True


def display_progress(uuid, current_task, num_processed, total_files):
    """
    Displays a single line of progress information that updates in-place.
    :param uuid: Full UUID of the file being processed
    :param current_task: Current task being processed
    :param num_processed: The current number of files that have been processed
    :param total_files: The total number of files that are going to be processed
    :return:
    """
    sys.stdout.write('\rProcessing %s / %s [%d/%d] (%.1f%%) complete...' % (
        uuid[:8],
        current_task.ljust(16),
        num_processed,
        total_files,
        (float(i) / float(total_files)) * 100)
    )
    sys.stdout.flush()


def list_directory(base_path):
    """
    List all json files in a directory.
    :param base_path: The base path of the directory.
    :return: A list of file paths to each json file
    """

    def full_path(f):
        return os.path.join(base_path, f)

    return [full_path(f) for f in os.listdir(base_path) if os.path.isfile(full_path(f)) and '.json' in f]


def read_candidate(fp):
    """
    Reads a Candidate json file from disk and returns a dict
    :param fp: The path to the candidate file
    :return: A dictionary representation of the JSON file
    """

    with open(fp) as f:
        return json.loads(''.join(f))


def translate_color(intensity):
    """
    Translates a intensity 0 <= x < 768 into an RGB color tuple
    :param intensity: An intensity measure, ranging from 0 up to 756
    :return: A tuple of the form (Red, Green, Blue)
    """

    intensity = int(intensity)

    if intensity < 256:
        r = intensity
        g = intensity
        b = intensity
    elif 256 <= intensity < 512:
        r = 255
        g = 255
        b = 511 - intensity
    else:
        r = 255
        g = 767 - intensity
        b = 0

    return r, g, b


def generate_raw_output_chart(f, data):
    """
    Creates a plot of the raw output from this candidate
    :param f: The filename of the candidate
    :param data: The output array / matrix
    """


def generate_raw_output_chart_2d(f, data):
    """
    Creates a plot of the raw output from this candidate
    :param f: The filename of the candidate
    :param data: The output array / matrix
    """

    data = np.array(data).transpose()

    # Plot and save the graph
    for column in data:
        plt.subplot(111)
        plt.plot(column)

    plt.ylabel('Value')
    plt.xlabel('Iteration')
    plt.title('Raw output')

    plt.savefig("%s-rawoutput.png" % os.path.splitext(f)[0])
    # Clear the figure buffer for next plot
    plt.clf()


def generate_state_history_plot(f, data):
    """
    Creates an image displaying the state history of the candidate
    :param f: The filename of the candidate
    :param data: The state history matrix
    """

    rows = len(data)
    cols = len(data[0])

    # Make a lookup table for the different kinds of state hash codes seen in the data
    codes = dict()

    # Create an equally spread out list of colors based on an RGB translation scheme
    num_colors = len(set(np.array(data).flatten()))
    step = COLORS / float(num_colors)
    colors = [translate_color(x * step) for x in range(0, COLORS)]

    # Either draw colors at random or gradually pull colors as new states appear
    # The latter will cause gradient patterns to emerge in many cases.
    if SHUFFLE_COLORS:
        random.shuffle(colors)
        for key, x in enumerate(codes.keys()):
            codes[key] = colors[x]
    else:
        x = 0
        for row in data:
            for col in row:
                if col not in codes:
                    codes[col] = colors[x]
                    x += 1

    img = Image.new('RGB', (cols, rows))

    for y in range(rows):
        for x in range(cols):
            val = data[y][x]
            img.putpixel((x, y), codes.get(val))

    img = img.resize((img.size[0] * W_SCALE_FACTOR, img.size[1] * H_SCALE_FACTOR))

    img.save("%s-statehistory.png" % os.path.splitext(f)[0])


def generate_state_transition_graph(f, data):
    """
    Creates a NetworkX plot displaying a directed graph of the possible state transitions from the state history
    :param f: The filename of the candidate
    :param data: The state history matrix
    """

    g = pgv.AGraph(directed=True)

    g.graph_attr['label'] = 'State transitions'
    g.node_attr['shape'] = 'circle'
    g.node_attr['color'] = 'red'
    g.edge_attr['color'] = 'black'

    state_id = 1
    states = dict()
    for y in range(0, len(data)):
        for x in range(0, len(data[y]) - 1):
            from_state = data[y][x]
            to_state = data[y][x+1]

            if from_state not in states:
                states[from_state] = state_id
                state_id += 1
            if to_state not in states:
                states[to_state] = state_id
                state_id += 1

            g.add_edge(states[from_state], states[to_state])
            g.get_node(states[from_state]).attr['label'] = 's%d' % states[from_state]
            g.get_node(states[to_state]).attr['label'] = 's%d' % states[to_state]

        # Add initial state, since it is missing from state history matrix
        g.add_edge(0, states[data[y][0]])
        g.get_node(0).attr['label'] = 's0'

    # g.draw("%s-stategraph.dot" % os.path.splitext(f)[0], prog='dot')
    g.draw("%s-stategraph.png" % os.path.splitext(f)[0], prog='dot')


def generate_power_spectrum_chart(f, data, frequency=2.0, over_sampling_factor=32.0, seconds=2.0):
    """
    Creates an image displaying the power spectrums of each column signal in the data matrix in a multi-line chart.
    :param f: The filename of the candidate
    :param data: The power spectrum matrix
    :param frequency: The frequency of the signal
    :param over_sampling_factor: The over sampling factor used in the samples
    :param seconds: The period over which the signal was captured
    """

    if not data:
        return

    f_s = frequency * over_sampling_factor
    fft_n = f_s * seconds
    delta_f = f_s / fft_n

    x_values = np.arange(0, (fft_n / 2.0) * delta_f, delta_f, float)

    # Plot and save the graph
    for spectrum in data:
        if not spectrum:
            return
        ax = plt.subplot(111)
        ax.set_yscale('log')
        plt.plot(x_values, spectrum)

    plt.ylabel('Power')
    plt.xlabel('Frequency (Hz)')
    plt.title('Power spectrum')

    if TIKZ_SAVE:
        tikz_save("%s-powerspectrum.tikz" % os.path.splitext(f)[0])
    plt.savefig("%s-powerspectrum.png" % os.path.splitext(f)[0])

    # Clear the figure buffer for next plot
    plt.clf()


def generate_sine_output_chart(f, data, seconds=2.0):
    """
    Creates an image displaying the raw sine wave output in a multi-line chart. One line for each dimension.
    :param f: The filename of the candidate
    :param data: The raw sine output matrix
    :param seconds: Number of seconds the signal was recorded over
    """
    data = np.array(data).transpose()

    # Plot and save the graph
    for sine_wave in data:
        if sine_wave is None:
            continue
        step = seconds / float(len(sine_wave))
        x_values = np.arange(0.0, seconds, step)

        # Generate x values (time)
        plt.subplot(111)
        plt.plot(x_values, sine_wave)

    plt.ylabel('Output')
    plt.xlabel('Time (s)')
    plt.title('Raw sine wave output')

    if TIKZ_SAVE:
        tikz_save("%s-sinewave.tikz" % os.path.splitext(f)[0])
    plt.savefig("%s-sinewave.png" % os.path.splitext(f)[0])

    # Clear the figure buffer for next plot
    plt.clf()


def generate_power_spectrum_chart_with_raw_sine_output(f, sine_series, power_spectrum):
    """
    Creates an image displaying both the power spectrum and the raw sine wave output of a candidate,
    in a double-figured multi-line plit.
    :param f: The filename of the candidate
    :param sine_series: The raw sine output matrix
    :param power_spectrum: The power spectrum matrix
    """

    # Plot and save the graph
    plt.figure(1)
    for sine in sine_series:
        if None in sine:
            continue
        plt.subplot(211)
        plt.plot(sine)
    plt.ylabel('Sine output')
    plt.xlabel('Frequency (Hz)')
    plt.title('Power spectrum')

    for spectrum in power_spectrum:
        if 'None' in spectrum:
            continue
        plt.subplot(212)
        plt.plot(spectrum)
    plt.ylabel('Output')
    plt.xlabel('Time')
    plt.title('Sine output')

    if TIKZ_SAVE:
        tikz_save("%s-powerspectrum-with-sine.tikz" % os.path.splitext(f)[0])
    plt.savefig("%s-powerspectrum-with-sine.png" % os.path.splitext(f)[0])

    # Clear the figure buffer for next plot
    plt.clf()


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print("Usage: python %s path/to/candidates" % sys.argv[0])
        sys.exit(1)

    file_path = sys.argv[1]
    json_files = list_directory(file_path)

    num_files = len(json_files)
    i = 0
    print("Processing %d candidates..." % len(json_files))
    for json_file in json_files:
        candidate = read_candidate(json_file)

        candidate_id = candidate['uuid']
        keys = candidate.keys()

        if 'powerSpectrum' in keys:
            display_progress(candidate_id, 'power_spectrum', i, num_files)
            power_spectrum = candidate.get('powerSpectrum')

            generate_power_spectrum_chart(
                json_file,
                power_spectrum,
                frequency=candidate.get('sineWaveFrequency', 2.0),
                over_sampling_factor=candidate.get('sineWaveOverSamplingFactor', 32.0),
                seconds=candidate.get('sineWaveSeconds', 2.0)
            )

        if 'rawSineOutput' in keys:
            display_progress(candidate_id, 'raw_sine_output', i, num_files)
            generate_sine_output_chart(
                json_file,
                candidate.get('rawSineOutput'),
                seconds=candidate.get('sineWaveSeconds')
            )

        if 'rawOutput' in keys:
            display_progress(candidate_id, 'raw_output', i, num_files)
            generate_raw_output_chart(json_file, candidate['rawOutput'])

        if 'rawOutput2D' in keys:
            display_progress(candidate_id, 'raw_output', i, num_files)
            generate_raw_output_chart_2d(json_file, candidate['rawOutput2D'])

        if 'stateHistory' in keys:
            display_progress(candidate_id, 'state_history', i, num_files)
            try:
                generate_state_history_plot(json_file, candidate['stateHistory'])
            except:
                pass

            display_progress(candidate_id, 'state_graph', i, num_files)
            generate_state_transition_graph(json_file, candidate['stateHistory'])

        i += 1
        display_progress(candidate['uuid'], 'complete', i, num_files)

    print('')
