# coding=utf-8

import argparse
import json
import os
import random
import shutil
import statistics
import sys


def read_candidate(fp):
    """
    Reads a Candidate json file from disk and returns a dict
    :param fp: The path to the candidate file
    :return: A dictionary representation of the JSON file
    """

    with open(fp) as f:
        return json.loads(''.join(f))


def list_directory(base_path):
    """
    List all json files in a directory.
    :param base_path: The base path of the directory.
    :return: A list of file paths to each json file
    """

    def full_path(f):
        return os.path.join(base_path, f)

    return [full_path(f) for f in os.listdir(base_path) if os.path.isfile(full_path(f)) and '.json' in f]


def sort_on_matrix_eval(candidates):
    """
    Sorts a list of candidates on their matrix evaluation scores.
    :param candidates: A list of candidate objects
    :return: A sorted list
    """

    def extract_pivots(candidate):
        me = candidate.get('matrixEvaluation', [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]])

        return me[0][2], me[1][2], me[0][1], me[1][1], me[0][0], me[0][1]

    return sorted(candidates, key=extract_pivots)


def sort_on_narma10(candidates):
    """
    Sorts a list of candidates on their NARMA10 scores.
    :param candidates: A list of candidate objects
    :return: A sorted list
    """

    def extract_pivots(candidate):
        n10 = candidate.get('narma', [[0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0]])

        return n10[0][1], n10[0][2]

    return sorted(candidates, key=extract_pivots, reverse=True)


def print_fitness(candidates, args):
    """
    Prints the candidate fitnesses as a list
    :param candidates: A list of candidate json objects
    :param args: The arguments object from argparser
    """

    scores = [c['testSetFitness'] for c in candidates]

    print(scores)


def print_statistics(candidates):
    """
    Prints a brief statistics summary of the candidate list with mean, median and population std dev and variance
    :param candidates: A list of candidte objects
    """

    scores = [c.get('testSetFitness', 0.0) for c in candidates]

    mean_fitness = statistics.mean(scores)
    median_fitness = statistics.median(scores)
    fitness_pstd_dev = statistics.pstdev(scores)
    fitness_pvariance = statistics.pvariance(scores, mean_fitness)

    print('Mean: %.3f, Median: %.3f, Population StdDev: %.3f, Population Var: %.3f' % (
        mean_fitness,
        median_fitness,
        fitness_pstd_dev,
        fitness_pvariance,
    ))


def print_narma10_statistics(candidates, latex_format=False):
    """
    Prints a statistics summary of the narma 10 error rates for each candidate, with mean and median error as well
    as std dev.
    :param candidates: A list of candidate objects
    """

    candidate_groups = {
        '5-1': [],
        '7-1': [],
        '7-2': [],
        '10-1': [],
        '10-3': [],
        '13-1': [],
        '16-3': [],
        '16-5': [],
    }

    for c in candidates:
        o = c.get('numOutputNodes')
        ief = c.get('inputExpansionFactor', 1)

        candidate_groups['%d-%d' % (o, ief)].append(c)

    def print_group_stats(candidate_set, group_label=None):
        scores = [c.get('narma', [[0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0]]) for c in candidate_set]

        mse = filter(lambda x: x < 1.0, [score[0][0] for score in scores])
        rmse = filter(lambda x: x < 1.0, [score[0][1] for score in scores])
        nrmse = filter(lambda x: x < 1.0, [score[0][2] for score in scores])
        rnmse = filter(lambda x: x < 1.0, [score[0][3] for score in scores])
        mae = filter(lambda x: x < 1.0, [score[0][4] for score in scores])

        measures = ['MSE', 'RMSE', 'NRMSE', 'RNMSE', 'MAE']
        results = [
                {'mean': 0.0, 'median': 0.0, 'stdev': 0.0},
                {'mean': 0.0, 'median': 0.0, 'stdev': 0.0},
                {'mean': 0.0, 'median': 0.0, 'stdev': 0.0},
                {'mean': 0.0, 'median': 0.0, 'stdev': 0.0},
                {'mean': 0.0, 'median': 0.0, 'stdev': 0.0},
        ]

        for i, measure in enumerate((mse, rmse, nrmse, rnmse, mae)):
            results[i]['mean'] = statistics.mean(measure)
            results[i]['median'] = statistics.median(measure)
            results[i]['stdev'] = statistics.pstdev(measure)

        if latex_format:
            latex_fmt_string = '\t\t\t\\textbf{%s} & %.6f$\\pm$%.6f & %.6f$\\pm$%.6f & '
            latex_fmt_string += '%.6f$\\pm$%.6f & %.6f$\\pm$%.6f \\\\'

            print(latex_fmt_string % (
                group_label,
                results[0]['mean'],
                results[0]['stdev'],
                results[1]['mean'],
                results[1]['stdev'],
                results[2]['mean'],
                results[2]['stdev'],
                results[4]['mean'],
                results[4]['stdev']
            ))
        else:
            for i, measure in enumerate(measures):
                print('%s: Mean: %.5f Median: %.5f StDev: %.5f' % (
                    measure.ljust(6),
                    results[i]['mean'],
                    results[i]['median'],
                    results[i]['stdev'],
                ))
    if latex_format:
        print('\t\t\t\\textbf{Group} & \\textbf{MSE} & \\textbf{RMSE} & \\textbf{NRMSE} & \\textbf{MAE} \\\\')
    for key in sorted(candidate_groups.keys(), key=lambda x: tuple(map(int, x.split('-')))):
        if latex_format:
            print_group_stats(candidate_groups[key], group_label=key)
        else:
            print('--- %s ---' % key)
            print_group_stats(candidate_groups[key])


def print_candidate_details(candidates, args, total_files):
    """
    Prints a list with detailed information about each candidate, based on options and filters from commandline
    :param candidates: A list of candidate objects
    :param args: The argument object from argparser
    :param total_files: The total number of files read from disk
    """

    found = 0
    standard_fmt_string = '%s.json (%.2f) [O:%d TS:%d IE:%.1s SIF:%d]'
    verbose_fmt_string = '%s.json (%.2f) [O:%d TS:%d IE:%.1s SIF:%.1s IEF:%d WS:%d D:%d H1:%d H2:%d '
    verbose_fmt_string += 'LB:%d RL:%.1f RU:%.1f GF:%.1f MD:%d]'
    very_verbose_fmt_string = verbose_fmt_string + '\nME:[%.2f %.2f %.2f | %.2f %.2f %.2f] '
    very_verbose_fmt_string += 'NA10:[%.5f %.5f %.5f %.5f %.5f]\nF:%s'
    for candidate in candidates:
        matrix_evaluation = candidate.get('matrixEvaluation', [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]])
        narma = candidate.get('narma', [[0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0]])
        params = [
            candidate.get('uuid', 'null'),
            candidate.get('testSetFitness', 0.0),
            candidate.get('numOutputNodes', -1),
            candidate.get('numCachedTimeSteps', -1),
            candidate.get('useInputExpansion', False),
            candidate.get('useSimpleInputFunction', False),
            candidate.get('inputExpansionFactor', 0.0),
            candidate.get('windowSize', -1),
            candidate.get('signalDelay', -1),
            candidate.get('hiddenLayerOneNodeCount', -1),
            candidate.get('hiddenLayerTwoNodeCount', -1),
            candidate.get('levelsBack', -1),
            candidate.get('realNumbersLower', -0.0),
            candidate.get('realNumbersUpper', 0.0),
            candidate.get('maxGrowthFactor', 0.0),
            candidate.get('maxNodesMovedOrDuplicated', 0),
            matrix_evaluation[0][0],
            matrix_evaluation[0][1],
            matrix_evaluation[0][2],
            matrix_evaluation[1][0],
            matrix_evaluation[1][1],
            matrix_evaluation[1][2],
            narma[0][0],
            narma[0][1],
            narma[0][2],
            narma[0][3],
            narma[0][4],
            candidate.get('functionSet', 'null'),
        ]

        if args.very_verbose:
            print(very_verbose_fmt_string % tuple(params[:28]))
        elif args.verbose:
            print(verbose_fmt_string % tuple(params[:16]))
        else:
            print(standard_fmt_string % tuple(params[:6]))
        found += 1

    print('Found %d (%.1f%%) candidates' % (
        found,
        (found / float(total_files)) * 100.0
    ))


def copy_candidates(candidates, args):
    """
    Copies the candidates in the provided candidate list to the copy argument in the arguments namespace
    :param candidates: A list of candidate objects
    :param args: The arguent object from argparser
    """

    if os.path.exists(args.copy):
        choice = raw_input('Are you sure you want to copy these files? (y/N) ')
        if choice.lower() != 'y':
            print('Aborting')
            sys.exit(0)

        for c in candidates:
            print('Copying %s.json to %s' % (c.get('uuid'), os.path.join(args.copy, '%s.json' % c.get('uuid'))))
            shutil.copy(os.path.join(args.path, '%s.json' % c.get('uuid')), args.copy)
    else:
        print('The path %s does not exist' % args.copy)


def move_candidates(candidates, args):
    """
    Moves the candidates in the provided candidate list to the move argument in the arguments namespace
    :param candidates: A list of candidate objects
    :param args: The arguent object from argparser
    """

    if os.path.exists(args.move):
        choice = raw_input('Are you sure you want to move these files? (y/N) ')
        if choice.lower() != 'y':
            print('Aborting')
            sys.exit(0)

        for c in candidates:
            print('Moving %s.json to %s' % (c.get('uuid'), os.path.join(args.move, '%s.json' % c.get('uuid'))))
            shutil.move(os.path.join(args.path, '%s.json' % c.get('uuid')), args.move)
    else:
        print('The path %s does not exist' % args.move)


def delete_candidates(candidates, args):
    """
    Deletes the candidates in the provided candidate list
    :param candidates: A list of candidate objects
    :param args: The arguent object from argparser
    """

    if os.path.exists(args.path):
        choice = raw_input('Are you sure you want to delete these files? (y/N) ')
        if choice.lower() != 'y':
            print('Aborting')
            sys.exit(0)

        for c in candidates:
            file_path = os.path.join(args.path, '%s.json' % c.get('uuid'))
            print('Deleting %s' % file_path)
            os.remove(file_path)
    else:
        print('The path %s does not exist' % args.path)


def main():
    """
    Main program task logic
    :return: Program return code
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("path",
                        help='The directory to load candidates from')
    parser.add_argument('-m', '--move', type=str, default=None,
                        help='Move candidates to target folder')
    parser.add_argument('-c', '--copy', type=str, default=None,
                        help='Copy candidates to target folder')
    parser.add_argument('-d', '--delete', action='store_true',
                        help='Deletes the selected candidates')
    parser.add_argument('-mf', '--min-fitness', type=float, default=0.0,
                        help='Select candidates with fitness higher than the provided number')
    parser.add_argument('-r', '--random', type=int, default=None,
                        help='Selects N random candidates')
    parser.add_argument('-o', '--outputs', type=int, default=None,
                        help='Select candidates with number of outputs equal to the provided number')
    parser.add_argument('-ws', '--window-size', type=int, default=None,
                        help='Select candidates with window size equal to the provided number')
    parser.add_argument('-ief', '--input-expansion-factor', type=int, default=None,
                        help='Select candidates with input expansion factor equal to the provided number')
    parser.add_argument('-im', '--incomplete-matrix', type=int, default=None,
                        help='Select candidates that have at least the given number of incomplete matrix evaluations')
    parser.add_argument('-in10', '--incomplete-narma10', type=int, default=None,
                        help='Select candidates that have at least the given number of incomplete narma10')
    parser.add_argument('-ifs', '--incomplete-function-set', action='store_true',
                        help='Select candidates that have missing or malconfigured function set')
    parser.add_argument('-pf', '--print-fitness', action='store_true',
                        help='Prints all the fitness scores as a list')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Verbose output')
    parser.add_argument('-vv', '--very-verbose', action='store_true',
                        help='Very verbose output')
    parser.add_argument('-s', '--sorted', action='store_true',
                        help='Sorts the candidates by fitness')
    parser.add_argument('-S', '--sorted-alpha', action='store_true',
                        help='Sorts the candidates alphabetically')
    parser.add_argument('-sme', '--sorted-matrix_eval', action='store_true',
                        help='Sorts the candidates by matrix evaluation scores, TP7 > TD7 > TP5 > TD5 > TP3 > TD3')
    parser.add_argument('-sn', '--sorted-narma', action='store_true',
                        help='Sorts the candidates by NARMA10 score, RMSE > NRMSE')
    parser.add_argument('-st', '--statistics', action='store_true',
                        help='Prints mean, median, standard deviation and variance of the candidates')
    parser.add_argument('-stn', '--statistics-narma10', action='store_true',
                        help='Prints mean, median and standard deviation for the NARMA10 error measures')
    parser.add_argument('-stnl', '--statistics-narma10-latex', action='store_true',
                        help='Prints mean and standard deviation as latex table for the NARMA10 error measures')

    args = parser.parse_args()

    file_path = args.path
    min_fitness = args.min_fitness
    json_files = list_directory(file_path)

    sys.stdout.write('Loading candidates...')
    sys.stdout.flush()

    # Read all files from disk, filtering by minimum fitness
    candidates = filter(lambda x: x.get('testSetFitness', 0.0) >= min_fitness, (read_candidate(c) for c in json_files))

    sys.stdout.write('\r')
    sys.stdout.flush()

    # Apply additional argument filters
    if args.random is not None:
        candidates = random.sample(candidates, args.random)
    if args.outputs is not None:
        candidates = filter(lambda x: x.get('numOutputNodes', -1) == args.outputs, candidates)
    if args.input_expansion_factor is not None:
        candidates = filter(lambda x: x.get('inputExpansionFactor', 0) == args.input_expansion_factor, candidates)
    if args.window_size is not None:
        candidates = filter(lambda x: x.get('windowSize', 0) == args.window_size, candidates)
    if args.incomplete_matrix is not None:
        def incomplete_me(cand):
            me = cand.get('matrixEvaluation', [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]])
            tot = 0
            for task in me:
                for result in task:
                    if result == 0.0:
                        tot += 1
            if tot >= args.incomplete_matrix:
                return True
            return False
        candidates = filter(incomplete_me, candidates)
    if args.incomplete_narma10 is not None:
        def incomplete_narma10(cand):
            narma = cand.get('narma', [[0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0]])
            tot = 0
            for result in narma[0]:
                if result == 0.0:
                    tot += 1
            if tot >= args.incomplete_narma10:
                return True
            return False
        candidates = filter(incomplete_narma10, candidates)
    if args.incomplete_function_set:
        candidates = filter(
            lambda x: x.get('functionSet', None) is None or x.get('functionSet', None) == 'all',
            candidates
        )

    # Order the candidates accordingly if specified
    if args.sorted:
        candidates = sorted(candidates, key=lambda x: x['testSetFitness'])
    elif args.sorted_alpha:
        candidates = sorted(candidates, key=lambda x: x['uuid'])
    elif args.sorted_matrix_eval:
        candidates = sort_on_matrix_eval(candidates)
    elif args.sorted_narma:
        candidates = sort_on_narma10(candidates)

    # Abort if we have nothing to do, to prevent errors in statistics etc
    if len(candidates) == 0:
        print('No candidates found')
        sys.exit(0)

    if args.copy or args.move or args.delete:
        print_candidate_details(candidates, args, len(json_files))

        # An optional print of statistics can be appended
        if args.statistics:
            print_statistics(candidates)

        # Copying is a standalone operation and must have presedence as well as return early
        if args.copy:
            copy_candidates(candidates, args)
        # Moving is a standalone operation and must have presedence as well as return early
        elif args.move:
            move_candidates(candidates, args)
        # Deleting is a standalone operation and must have presedence as well as return early
        elif args.delete:
            delete_candidates(candidates, args)

        sys.exit(0)

    # The two main modes of operation are either printing a fitness array or candidate details
    if args.print_fitness:
        print_fitness(candidates, args)
    else:
        print_candidate_details(candidates, args, len(json_files))

    # An optional print of statistics can be appended
    if args.statistics:
        print_statistics(candidates)
    if args.statistics_narma10:
        print_narma10_statistics(candidates)
    if args.statistics_narma10_latex:
        print_narma10_statistics(candidates, latex_format=True)


if __name__ == '__main__':
    main()
