# -*- coding: utf8 -*-
#
# Created by 'myth' on 5/25/17

import argparse
import json
import os
import random
import sys


def read_candidate(fp):
    """
    Reads a Candidate json file from disk and returns a dict
    :param fp: The path to the candidate file
    :return: A dictionary representation of the JSON file
    """

    with open(fp) as f:
        return json.loads(''.join(f))


def list_directory(base_path):
    """
    List all json files in a directory.
    :param base_path: The base path of the directory.
    :return: A list of file paths to each json file
    """

    def full_path(f):
        return os.path.join(base_path, f)

    return [full_path(f) for f in os.listdir(base_path) if os.path.isfile(full_path(f)) and '.json' in f]


def copy_fitness_to_tp7(candidate, args):
    """
    Copies the current testset fitness to tp7
    :return:
    """

    me = candidate.get('matrixEvaluation', [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]])
    me[0][2] = candidate['testSetFitness']
    candidate['matrixEvaluation'] = me

    with open(os.path.join(args.path, '%s.json' % candidate.get('uuid')), 'w') as f:
        f.write(json.dumps(candidate))
        f.write('\n')
        f.flush()


def fix_function_set(candidate, args):
    """
    Fixes the function set of a candidate to use proper one
    :param candidate: Candidate object
    :param args: Option parser args
    :return:
    """
    candidate['functionSet'] = 'Add,Constant,Cos,Divide,Max,Mean,Min,Modulo,Multiply,Pow,Sin,Square,Subtract,Tan,Tanh'
    for genome in candidate['genotype']:
        if genome[0] > 24:
            genome[0] = random.randint(0, 24)

    with open(os.path.join(args.path, '%s.json' % candidate.get('uuid')), 'w') as f:
        f.write(json.dumps(candidate))
        f.write('\n')
        f.flush()


def main():
    """
    Main program tasklogic
    :return: Program
    return code
    """

    parser = argparse.ArgumentParser()
    parser.add_argument('path', help='The directory to load candidates from')
    parser.add_argument('-ffs', '--fix-function-set', action='store_true', help='Sets proper function set')
    parser.add_argument('-cfttp7', '--copy-fitness-to-tp7', action='store_true', help='Copies value of fitness to TP7')
    args = parser.parse_args()

    candidates = [read_candidate(fp) for fp in list_directory(args.path)]
    n = len(candidates)
    for i, candidate in enumerate(candidates):
        if args.fix_function_set:
            fix_function_set(candidate, args)
        elif args.copy_fitness_to_tp7:
            copy_fitness_to_tp7(candidate, args)

        sys.stdout.write('\rProcessing candidate %d/%d (%.1f%%)' % (
            i+1,
            n,
            100.0 * (float(i+1) / float(n))
        ))
        sys.stdout.flush()

    if args.fix_function_set:
        task = 'fixed function set'
    elif args.copy_fitness_to_tp7:
        task = 'hammered testSetFitness into the TP7 matrix evaluation slot'
    else:
        task = 'did nothing'
    print('\nSuccessfully %s on %d candidates!' % (task, n))


if __name__ == '__main__':
    main()
