# coding=utf-8

from collections import OrderedDict
import matplotlib.pyplot as plt

import json
import os
import sys
import re

# CONFIG
PARAMS = {'legend.fontsize': 'x-large',
          'figure.figsize': (10, 10),
          'axes.labelsize': '22',
          'axes.titlesize': 'x-large',
          'xtick.labelsize': '16',
          'ytick.labelsize': '16',
          'font.family': 'Times New Roman'
          }
# Box Plot
BOX_COLOR = 'grey'
MEDIAN_COLOR = 'black'
WHISKER_COLOR = 'black'
EDGE_COLOR = 'black'
MARKER_SIZE_BOX = 10
MARKER_COLOR_BOX = 'grey'
MARKER_EDGE = 'black'

# Scatter Plot
MARKER_SIZE_SCATTER = 100
MARKER_COLOR_SCATTER = 'grey'

# General
AXIS_LABEL = 'Accuracy'
MARKER_TYPE = '.'



def list_directory(base_path):
    """
    List all json files in a directory.
    :param base_path: The base path of the directory.
    :return: A list of file paths to each json file
    """

    def full_path(f):
        return os.path.join(base_path, f)

    return [full_path(f) for f in os.listdir(base_path) if os.path.isfile(full_path(f)) and '.json' in f]


def read_candidate(fp):
    """
    Reads a Candidate json file from disk and returns a dict
    :param fp: The path to the candidate file
    :return: A dictionary representation of the JSON file
    """

    with open(fp) as f:
        return json.loads(''.join(f))


def generate_box_plot_multi(data, labels):
    plt.rcParams.update(PARAMS)

    fig, ax = plt.subplots()
    plt.subplots_adjust(left=0.075, right=0.95, top=0.9, bottom=0.25)
    bp = plt.boxplot(data, labels=labels, notch=1, widths=0.3, sym='.', vert=1, patch_artist=True,
                     boxprops=dict(facecolor=BOX_COLOR, color=EDGE_COLOR),
                     capprops=dict(color=BOX_COLOR),
                     whiskerprops=dict(color=WHISKER_COLOR),
                     flierprops=dict(markerfacecolor=MARKER_COLOR_BOX, markeredgecolor=MARKER_EDGE, marker=MARKER_TYPE,
                                     markersize=MARKER_SIZE_BOX),
                     medianprops=dict(color=MEDIAN_COLOR))

    ax.yaxis.grid(True, linestyle='-', which='major', color='lightgrey', alpha=0.9)
    ax.set_ylabel(AXIS_LABEL)

    #plt.savefig('%s-boxplot-%s.png' % ''.join(labels), EX_TYPE+EX_WIN)
    plt.savefig('%s-boxplot-%s%s.svg' % ('all', EX_TYPE, str(EX_WIN)), format='svg', dpi=1200)

def generate_scatter_plot(name, data):
    plt.rcParams.update(PARAMS)

    y = data
    x = list(range(len(data)))

    fig, ax = plt.subplots()
    plt.subplots_adjust(left=0.075, right=0.95, top=0.9, bottom=0.25)
    bp = plt.scatter(x, y, s=MARKER_SIZE_SCATTER, marker=MARKER_TYPE, color=MARKER_COLOR_SCATTER)

    ax.yaxis.grid(True, linestyle='-', which='major', color='lightgrey', alpha=0.9)
    ax.set_axisbelow(True)
    ax.set_ylabel(AXIS_LABEL)

    plt.xticks([], [])
    ax.set_xlabel(name)

    plt.savefig('%s-scatter.png' % name)


def generate_pie_chart(data, labels):
    fig1, ax1 = plt.subplots()
    ax1.pie(data, labels=labels, autopct='%1.1f%%',
        shadow=False, startangle=90)
    ax1.axis('equal')  # Equal aspect ratio ensures that pie is drawn as a circle.
    plt.savefig('pie-chart.png')

if __name__ == '__main__':
    if len(sys.argv) < 2:
        print("Usage: python %s path/to/candidates" % sys.argv[0])
        sys.exit(1)

    # PIE CHART
    if len(sys.argv) >= 2 and sys.argv[1] == 'pie':
        if len(sys.argv) >= 4 and len(sys.argv[2]) > 1 and len(sys.argv[3]) > 1:
            data = map(int, sys.argv[2].split(','))
            labels = sys.argv[3].split(',')
            print data, labels
            generate_pie_chart(data, labels)
        else:
            print 'Pie Chart requires a list with data and a list with labels!!'
        sys.exit()

    # SCATTER AND BOX PLOT
    file_path = sys.argv[1]
    json_files = list_directory(file_path)
    configs = OrderedDict()
    for json_path in json_files:
        candidate = read_candidate(json_path)

        if 'inputExpansionFactor' not in candidate:
            candidate['inputExpansionFactor'] = 1

        configuration = "%s-%s" % (candidate['numOutputNodes'], candidate['inputExpansionFactor'])

        if configuration in configs.keys():
            configs[configuration].append(candidate['testSetFitness'])
        else:
            configs[configuration] = [candidate['testSetFitness']]

    if len(sys.argv) >= 3:
        if sys.argv[2] == 'scatter':
            # Draw scatter plots
            for conf, fitness in configs.iteritems():
                generate_scatter_plot(conf, fitness)

        elif sys.argv[2] == 'box':
            # Draw box plot
            generate_box_plot_multi(configs.values(), configs.keys())
    else:
        # Draw both scatter and box
        generate_box_plot_multi(configs.values(), configs.keys())

        for conf, fitness in configs.iteritems():
            generate_scatter_plot(conf, fitness)
