# -*- coding: utf8 -*-
#
# Created by 'myth' on 5/25/17

import argparse
import json
import math
import numpy as np
import os
from sklearn import linear_model, model_selection, metrics
import sys
import statistics
import time


def read_candidate(fp):
    """
    Reads a Candidate json file from disk and returns a dict
    :param fp: The path to the candidate file
    :return: A dictionary representation of the JSON file
    """

    with open(fp) as f:
        return json.load(f)


def list_directory(base_path, full_path=True):
    """
    List all json files in a directory.
    :param base_path: The base path of the directory.
    :param full_path: True if concatenating file name with base path
    :return: A list of file paths to each json file
    """

    def concat(f):
        return os.path.join(base_path, f)

    return [
        concat(f) if full_path else f for f in os.listdir(base_path) if os.path.isfile(concat(f)) and '.json' in f
    ]


def perform_ridge_regression(fp, args):
    """
    Performs ridge regression on a candidate
    :param fp: Candidate file path
    :param args: Option parser args
    """

    uuid = os.path.splitext(fp)[0]
    input_json = os.path.join(args.path, fp)
    output_csv = os.path.join(args.path, '%s-out.csv' % uuid)
    target_csv = os.path.join(args.path, '%s-target.csv' % uuid)

    if not os.path.isfile(output_csv) or not os.path.isfile(target_csv):
        return

    candidate = read_candidate(input_json)

    output_data = np.genfromtxt(output_csv, delimiter=',')
    target_data = np.genfromtxt(target_csv, delimiter=',')

    if not np.isfinite(np.sum(output_data)):
        return

    mse = []
    rmse = []
    nrmse = []
    rnmse = []
    mae = []

    for i in range(args.iterations):
        x_train, x_test, y_train, y_test = model_selection.train_test_split(output_data, target_data, test_size=0.25)
        model = linear_model.Ridge(alpha=0.1)
        model.fit(x_train, y_train)

        predictions = model.predict(x_test)
        max_target = max(y_test)
        min_target = min(y_test)
        target_range = max_target - min_target

        mse_value = metrics.mean_squared_error(y_test, predictions)
        rmse_value = math.sqrt(mse_value)

        mse.append(mse_value)
        rmse.append(rmse_value)
        nrmse.append(rmse_value / target_range)
        rnmse.append(math.sqrt(mse_value / target_range))
        mae.append(metrics.mean_absolute_error(y_test, predictions))

    mse_mean = statistics.mean(mse)
    rmse_mean = statistics.mean(rmse)
    nrmse_mean = statistics.mean(nrmse)
    rnmse_mean = statistics.mean(rnmse)
    mae_mean = statistics.mean(mae)

    narma = candidate.get('narma', [[0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0]])
    narma[0][0] = mse_mean
    narma[0][1] = rmse_mean
    narma[0][2] = nrmse_mean
    narma[0][3] = rnmse_mean
    narma[0][4] = mae_mean
    candidate['narma'] = narma

    with open(os.path.join(args.path, '%s.json' % candidate.get('uuid')), 'w') as f:
        json.dump(candidate, f)


def main():
    """
    Main program tasklogic
    :return: Program
    return code
    """

    parser = argparse.ArgumentParser()
    parser.add_argument('path', help='The directory to load candidates from')
    parser.add_argument('-i', '--iterations', type=int, default=20,
                        help='Make <i> passes over the data set, before averaging the errors (default 20)')
    args = parser.parse_args()

    sys.stdout.write('Loading candidates...')
    sys.stdout.flush()

    files = list_directory(args.path, full_path=False)

    sys.stdout.write('\r')
    sys.stdout.flush()

    n = len(files)
    start_time = time.clock()
    timestamps = [time.clock() for i in range(10)]
    for i, candidate in enumerate(files):
        perform_ridge_regression(candidate, args)

        progress = '%d/%d (%.1f%%)' % (i+1, n, 100.0 * (float(i+1) / float(n)))
        timestamps[i % 10] = time.clock()
        speed = 10.0 / (time.clock() - timestamps[abs(i - 9) % 10])
        speed_format = '%.1f c/s' % speed
        sys.stdout.write('\rProcessing candidate %s at %s' % (progress, speed_format.ljust(10)))
        sys.stdout.flush()

    print('\nNARMA10 Ridge regression done for %d candidates in %.1fs (%.1f c/s)' % (
        n,
        time.clock() - start_time,
        float(n) / (time.clock() - start_time)
    ))


if __name__ == '__main__':
    main()
