package no.ntnu.cerebrum.medulla;


import com.google.gson.Gson;
import no.ntnu.cerebrum.medulla.config.Config;
import no.ntnu.cerebrum.medulla.data.*;
import no.ntnu.cerebrum.medulla.reservoir.SMCGPReservoir;
import no.ntnu.cerebrum.medulla.reservoir.readout.MultiLayerClassifier;
import no.ntnu.cerebrum.medulla.reservoir.readout.MultiLayerRegression;
import no.ntnu.cerebrum.medulla.reservoir.readout.SingleLayerLinearRegression;
import no.ntnu.cerebrum.medulla.smcgp.Genotype2;
import no.ntnu.cerebrum.medulla.smcgp.Phenotype2;
import no.ntnu.cerebrum.medulla.smcgp.evolution.Evolve;
import no.ntnu.cerebrum.medulla.telemetry.EvolveStatsCollector;
import no.ntnu.cerebrum.medulla.telemetry.ReservoirStatsCollector;
import org.apache.commons.math3.stat.descriptive.DescriptiveStatistics;
import org.apache.commons.math3.stat.descriptive.SummaryStatistics;
import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.dataset.DataSet;
import org.nd4j.linalg.dataset.SplitTestAndTrain;
import org.nd4j.linalg.factory.Nd4j;
import org.nd4j.linalg.string.NDArrayStrings;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.LocalDateTime;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.TimeoutException;
import java.util.stream.Stream;


/**
 * Created 2017-02-15.
 * <p>
 * cerebrum-medulla is licenced under the MIT licence.
 */
public class TaskSelector {
    static Logger log = Logger.getLogger(TaskSelector.class);

    static int windowSize;
    static int delay;

    static List<Data> trainingData;
    static INDArray lastOutput;

    static boolean overwriteCandidates = false;
    static int timeout;
    static int numMatrixEvaluations = 0;

    /**
     * Creates a Phenotype of the provided genotype, spawns a SMCGPReservoir instance, stats collactor and
     * ReservoirDataSet instance for pre-buffering the data through the SMCGP. Performs basic analysis of the raw
     * reservoir output before potentially allowing the smcgp output to train a readout network.
     * @param phenotype A Phenotype2 instance.
     * @return A Phenotype2 instance if the genotype was acceptable and used to train the readout layer. null otherwise.
     */
    private static Candidate evaluateGenotype(Phenotype2 phenotype) {
        // Set up SMCGPReservoir with readout layer
        SMCGPReservoir r = new SMCGPReservoir(phenotype);
        String dataSetType = Config.get("no.ntnu.cerebrum.medulla.dataset.type");
        if (dataSetType == null) throw new IllegalStateException("Failed to identify data set type from config!");

        if (dataSetType.equals("narma10") || dataSetType.equals("narma30")) {
            if (Config.getBoolean("no.ntnu.cerebrum.medulla.reservoir.readout.useMultilayerForRegression")) {
                MultiLayerRegression sllr = new MultiLayerRegression(Config.getInt("no.ntnu.cerebrum.medulla.smcgp.outputs"), 1);
                r.readoutLayer(sllr);
            } else {
                SingleLayerLinearRegression sllr = new SingleLayerLinearRegression(Config.getInt("no.ntnu.cerebrum.medulla.smcgp.outputs"), 1);
                r.readoutLayer(sllr);
            }
        } else {
            MultiLayerClassifier slc = new MultiLayerClassifier(Config.getInt("no.ntnu.cerebrum.medulla.smcgp.outputs"), 2);
            r.readoutLayer(slc);
        }

        // Set up a stats collector and register for SMCGP data processing and readout layer training events.
        ReservoirStatsCollector rsc = new ReservoirStatsCollector();
        r.addSMCGPPreProcessingListener(rsc);
        r.addReservoirTrainingListener(rsc);
        phenotype.addSMCGPProcessingListener(rsc);

        // Run the training and test data through the reservoir and build ND4j data sets from it, ready for readout training.
        org.nd4j.linalg.dataset.api.DataSet trainSet = buildDataSet(phenotype, r, trainingData);
        if (trainSet == null) return null;
        // Print to console to see what it looks like
        NDArrayStrings outputFormatter = new NDArrayStrings(", ", 4);
        System.out.println(outputFormatter.format(trainSet.getRange(0, 32).getFeatures()));
        trainSet.shuffle();

        if (Config.getBoolean("no.ntnu.cerebrum.medulla.dataset.dumpToCSV")) {
            CSV.writeToFile(phenotype.getUuid() + "-out.csv", DataFactory.convertTo2DArray(trainSet.getFeatures()));
            CSV.writeToFile(phenotype.getUuid() + "-target.csv", DataFactory.convertTo2DArray(trainSet.getLabels()));
        }

        double min = trainSet.getFeatures().minNumber().doubleValue();
        double max = trainSet.getFeatures().maxNumber().doubleValue();
        if (Config.getBoolean("no.ntnu.cerebrum.medulla.dataset.normalize")) {
            if (max - min != 0.0) trainSet.multiplyBy(1.0 / (max - min));
        }

        org.nd4j.linalg.dataset.api.DataSet testSet;
        if (dataSetType.equals("narma10") || dataSetType.equals("narma30")) {
            SplitTestAndTrain stat = trainSet.splitTestAndTrain(0.5);
            trainSet = stat.getTrain();
            testSet = stat.getTest();

        } else {
            testSet = trainSet.sample(Config.getInt("no.ntnu.cerebrum.medulla.dataset.testSamples"));
        }

        r.train(
            Config.getInt("no.ntnu.cerebrum.medulla.reservoir.epochs"),
            trainSet,
            testSet,
            Config.getInt("no.ntnu.cerebrum.medulla.reservoir.evaluateEveryNumEpochs")
        );

        r.test(testSet);

        // Update with fitness gathered from stats collector
        phenotype.setFitness(rsc.getTestSetFitness());
        phenotype.setTrainingSetFitness(rsc.getTrainingSetFitness());

        lastOutput = trainSet.getRange(0, Config.getInt("no.ntnu.cerebrum.medulla.dataset.numSamplesPerBatch")).getFeatures();

        Candidate candidate = new Candidate(phenotype);

        if (lastOutput.shape()[1] == 1) candidate.rawOutput = DataFactory.convertTo1DArray(lastOutput);
        else candidate.rawOutput2D = DataFactory.convertTo2DArray(lastOutput);
        candidate.epochsTrained = rsc.getNumEpochsTrained();

        if (dataSetType.equals("narma10") || dataSetType.equals("narma30")) {
            double[] regressionStats;

            if (dataSetType.equals("narma10")) regressionStats = candidate.getNarma10();
            else regressionStats = candidate.getNarma30();

            regressionStats[0] = rsc.getTestSetMSE();
            regressionStats[1] = rsc.getTestSetRMSE();
            regressionStats[2] = rsc.getTestSetNRMSE();
            regressionStats[3] = rsc.getTestSetRNMSE();
            regressionStats[4] = rsc.getTestSetMAE();
        }

        return candidate;
    }

    private static double[] performPreProcessingWithAnalytics(Phenotype2 p, SMCGPReservoir r, double[] inputData) throws TimeoutException {
        p.reset();
        ReservoirStatsCollector rsc = new ReservoirStatsCollector();
        p.addSMCGPProcessingListener(rsc);
        r.addSMCGPPreProcessingListener(rsc);

        double[] results = r.preProcess(inputData);

        log.info(String.format(
            "Processed inputs: %d, SM Ops: %d, SM States: %d, AN: %.0f/%.2f/%.0f, AIP: %.2f Time: %.0f ms %s",
            rsc.getTotalInputsProcessed(),
            rsc.getNumSMFunctionsApplied(),
            rsc.getNumActiveNodeStates(),
            rsc.getMinActiveNodes(),
            rsc.getMeanActiveNodes(),
            rsc.getMaxActiveNodes(),
            rsc.getActiveInputPercentage(),
            rsc.getTotalPreProcessingTime(),
            rsc.getSelfModificationNodeActivationCounts()
        ));

        p.removeSMCGPProcessingListener(rsc);
        r.removeSMCGPPreProcessingListener(rsc);

        return results;
    }

    private static double[][] performPreProcessingWithAnalytics(Phenotype2 p, SMCGPReservoir r, double[][] inputData) throws TimeoutException {
        p.reset();
        ReservoirStatsCollector rsc = new ReservoirStatsCollector();
        p.addSMCGPProcessingListener(rsc);
        r.addSMCGPPreProcessingListener(rsc);

        double[][] results = r.preProcess(inputData);

        log.info(String.format(
            "Processed inputs: %d, SM Ops: %d, SM States: %d, AN: %.0f/%.2f/%.0f, AIP: %.2f Time: %.0f ms %s",
            rsc.getTotalInputsProcessed(),
            rsc.getNumSMFunctionsApplied(),
            rsc.getNumActiveNodeStates(),
            rsc.getMinActiveNodes(),
            rsc.getMeanActiveNodes(),
            rsc.getMaxActiveNodes(),
            rsc.getActiveInputPercentage(),
            rsc.getTotalPreProcessingTime(),
            rsc.getSelfModificationNodeActivationCounts()
        ));

        p.removeSMCGPProcessingListener(rsc);
        r.removeSMCGPPreProcessingListener(rsc);

        return results;
    }

    /**
     * Constructs a merged ND4j data set, pre-processed using a provided phenotype and reservoir.
     * Supports both single input/output and multi input/output.
     * @param p A Phenotype2 instance.
     * @param r A SMCGPReservoir instance that has been initialized with the Phenotype.
     * @param dataSets A list of Data objects containing data samples and expected output labels.
     * @return A processed and merged data set consisting of the raw reservoir output to expected label mapping.
     */
    private static org.nd4j.linalg.dataset.api.DataSet buildDataSet(Phenotype2 p, SMCGPReservoir r, List<Data> dataSets) {
        List<DataSet> processedDataSets = new ArrayList<>();
        String dataSetType = Config.get("no.ntnu.cerebrum.medulla.dataset.type");
        if (dataSetType == null) dataSetType = "temporalParity";

        try {
            for (Data data : dataSets) {
                DataSet rawOutput;

                if (Config.getInt("no.ntnu.cerebrum.medulla.smcgp.outputs") < 2) {
                    rawOutput = new DataSet(Nd4j.create(
                        performPreProcessingWithAnalytics(p, r, data.flatInput())).reshape(data.numSamples(), 1),
                        Nd4j.create(data.output())
                    );

                } else {
                    rawOutput = new DataSet(Nd4j.create(
                        performPreProcessingWithAnalytics(p, r, data.input())),
                        Nd4j.create(data.output())
                    );
                }

                // Kill infinities with fire
                if (Double.isInfinite(rawOutput.getFeatures().maxNumber().doubleValue())) return null;

                processedDataSets.add(rawOutput);
            }
        } catch (TimeoutException e) {
            timeout++;
            return null;
        }


        if (dataSetType.equals("narma10") || dataSetType.equals("narma30")) {
            int start = 9;
            if (dataSetType.equals("narma30")) start = 29;

            List<DataSet> slicedDataSets = new ArrayList<>();
            for (DataSet d : processedDataSets) {
                slicedDataSets.add((DataSet) d.getRange(start, d.numExamples()));
            }
            processedDataSets = slicedDataSets;
        }

        return DataSet.merge(processedDataSets);
    }

    /**
     * Creates new training data batches base on config settings
     */
    private static void generateNewTrainingData() {
        windowSize = Config.getInt("no.ntnu.cerebrum.medulla.dataset.temporal.windowSize");
        delay = Config.getInt("no.ntnu.cerebrum.medulla.dataset.temporal.delay");
        String dataSetType = Config.get("no.ntnu.cerebrum.medulla.dataset.type");
        if (dataSetType == null) throw new IllegalStateException("Failed to parse data set type from config!");

        int numTrainSamples = Config.getInt("no.ntnu.cerebrum.medulla.dataset.trainingSamples");
        int samplesPerBatch = Config.getInt("no.ntnu.cerebrum.medulla.dataset.numSamplesPerBatch");
        int numDataSets = numTrainSamples / samplesPerBatch;

        switch (dataSetType) {
            case "temporalParity":
                trainingData = DataFactory.createTemporalParityDataWithEvenParityBitDistribution(
                    numDataSets, windowSize, delay, numTrainSamples, true
                );
                break;
            case "temporalDensity":
                trainingData = DataFactory.createTemporalDensityDataWithEvenParityBitDistribution(
                    numDataSets, windowSize, delay, numTrainSamples, true
                );
                break;
            case "narma10":
                trainingData = DataFactory.generateNARMA10DataSet();
                break;
            case "narma30":
                trainingData = DataFactory.generateNARMA30DataSet();
                break;
        }

        if (dataSetType.equals("temporalParity") || dataSetType.equals("temporalDensity")) {
            log.info(String.format(
                "Generating %d %s training samples, in %d samples per batch, with window size %d and delay %d...",
                numTrainSamples,
                dataSetType,
                samplesPerBatch,
                windowSize,
                delay
            ));
        } else {
            log.info(String.format("Generating 10 %s training samples, in 1000 samples per batch.", dataSetType));
        }
    }

    public static void generateNewCandidates() {
        int i = 1;

        // Record global stats
        DescriptiveStatistics globalFitness = new DescriptiveStatistics();
        EvolveStatsCollector globalEsc = new EvolveStatsCollector();

        double minFitness = Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.phenotype.minFitness");
        double gigaByte = Math.pow(1024, 3);
        double maxMem = Runtime.getRuntime().maxMemory();

        while (true) {
            log.info("Starting evolution " + i + "...");

            double usedMem = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
            double allocation = (usedMem / maxMem) * 100.0;

            log.info(String.format(
                "Memory: %.2fGB / %.2fGB (%.1f%%)",
                usedMem / gigaByte,
                maxMem / gigaByte,
                allocation
            ));
            Evolve e = new Evolve();
            // Record per-evolution stats
            EvolveStatsCollector localEsc = new EvolveStatsCollector();
            e.addEvolutionListener(globalEsc);
            e.addEvolutionListener(localEsc);

            Phenotype2 p = e.evolve();
            System.out.println(new Candidate(p).toStringWithNodeTree());

            p.reset();
            Candidate result = evaluateGenotype(p);
            if (result == null) {
                log.warn("Phenotype failed full eval...");
                e.removeEvolutionListener(localEsc);
                continue;
            }

            // Store all fitness values and print stats for this round
            globalFitness.addValue(p.getFitness());
            log.info(String.format(
                "Evolved phenotypes: %d, Mean fitness: %.3f Fitness StdDev: %.3f",
                globalFitness.getN(),
                globalFitness.getMean(),
                globalFitness.getStandardDeviation()
            ));
            log.info("Last evolve stats: " + localEsc.toString());
            log.info("Global evolve stats: " + globalEsc.toString());

            if (result.getTestSetFitness() >= minFitness) {
                result.setEpochsTrained(result.epochsTrained);
                result.save();
            } else {
                log.info(String.format(
                    "Candidate had fitness score %.3f, lacking %.3f to be stored for later use.",
                    result.getTestSetFitness(),
                    minFitness - result.getTestSetFitness()
                ));
            }

            i++;

            e.removeEvolutionListener(localEsc);
            e.removeEvolutionListener(globalEsc);
        }
    }

    /**
     * Performs a new read out training on the candidate using data set from config file. Will overwrite
     * rawOutput, epochs trained, as well as train and test set fitness.
     */
    public static void reEvaluateCandidatesFromDisk() {
        String path = Storage.BASE_PATH + Storage.CANDIDATES;
        Gson g = new Gson();

        double gigaByte = Math.pow(1024, 3);
        double maxMem = Runtime.getRuntime().maxMemory();

        try (Stream<Path> paths = Files.walk(Paths.get(path))) {
            paths
                .filter(Files::isRegularFile)
                .filter(fp -> fp.toString().toLowerCase().endsWith("json"))
                .map(Storage::readFromFile)
                .map(c -> g.fromJson(c, Candidate.class))
                .forEach(c -> {
                    // Set current configuration settings to match the one from the candidate
                    Config.set("no.ntnu.cerebrum.medulla.smcgp.levelsback", c.levelsBack.toString());
                    Config.set("no.ntnu.cerebrum.medulla.smcgp.realnumbersUpper", c.realNumbersUpper.toString());
                    Config.set("no.ntnu.cerebrum.medulla.smcgp.realnumbersLower", c.realNumbersLower.toString());
                    Config.set("no.ntnu.cerebrum.medulla.smcgp.outputs", c.getNumOutputNodes().toString());
                    Config.set("no.ntnu.cerebrum.medulla.smcgp.maxGrowthFactor", c.maxGrowthFactor.toString());
                    Config.set("no.ntnu.cerebrum.medulla.smcgp.useSimpleInputFunction", c.useSimpleInputFunction.toString());
                    Config.set("no.ntnu.cerebrum.medulla.smcgp.maxNodesMovedOrDuplicated", c.maxNodesMovedOrDuplicated.toString());
                    Config.set("no.ntnu.cerebrum.medulla.reservoir.numCachedTimeSteps", c.numCachedTimeSteps.toString());
                    Config.set("no.ntnu.cerebrum.medulla.reservoir.useInputExpansion", c.useInputExpansion.toString());
                    Config.set("no.ntnu.cerebrum.medulla.reservoir.inputExpansionFactor", c.inputExpansionFactor.toString());

                    Candidate result = evaluateGenotype(c.toPhenotype());

                    double usedMem = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
                    double allocation = (usedMem / maxMem) * 100.0;

                    log.info(String.format(
                        "Memory: %.2fGB / %.2fGB (%.1f%%)",
                        usedMem / gigaByte,
                        maxMem / gigaByte,
                        allocation
                    ));

                    if (result == null) {
                        log.warn("Evaluate phenotype returned 'null'");
                    } else {
                        assert c.getUuid().equals(result.getUuid());
                        c.testSetFitness = result.testSetFitness;
                        c.trainingSetFitness = result.trainingSetFitness;
                        c.rawOutput2D = result.rawOutput2D;
                        c.epochsTrained = result.epochsTrained;
                        c.windowSize = Config.getInt("no.ntnu.cerebrum.medulla.dataset.temporal.windowSize");
                        c.signalDelay = Config.getInt("no.ntnu.cerebrum.medulla.dataset.temporal.delay");
                        c.hiddenLayerOneNodeCount = Config.getInt("no.ntnu.cerebrum.medulla.reservoir.readout.hiddenLayerOneNodeCount");
                        c.save();
                    }

                    numMatrixEvaluations++;
                    log.info("Completed " + numMatrixEvaluations + " evaluations...");
                });
        } catch (IOException e) {
            log.error(String.format(
                "Failed to read directory: %s (%s - %s)",
                path,
                e.getMessage(),
                e.getCause()
            ));
        }
    }

    public static void performMatrixEvaluationWithCandidatesFromDisk() {
        int[] windowSizes = new int[]{3, 5, 7};
        int[] delays = new int[]{0};
        int runsPerCandidate = 2;

        HashMap<String, DescriptiveStatistics> scores = new HashMap<>();
        String path = Storage.BASE_PATH + Storage.CANDIDATES;
        Gson g = new Gson();

        try (Stream<Path> paths = Files.walk(Paths.get(path))) {
            paths
                .filter(Files::isRegularFile)
                .filter(fp -> fp.toString().toLowerCase().endsWith("json"))
                .map(Storage::readFromFile)
                .map(c -> g.fromJson(c, Candidate.class))
                .forEach(c -> matrixEvaluation(c, scores, windowSizes, delays, runsPerCandidate));
        } catch (IOException e) {
            log.error(String.format(
                "Failed to read directory: %s (%s - %s)",
                path,
                e.getMessage(),
                e.getCause()
            ));
        }
    }

    public static void matrixEvaluation(Candidate c, HashMap<String, DescriptiveStatistics> scores,
                                        int[] windowSizes, int[] delays, int runsPerCandidate) {
        overwriteCandidates = false;

        // Set current configuration settings to match the one from the candidate
        Config.set("no.ntnu.cerebrum.medulla.smcgp.levelsback", c.levelsBack.toString());
        Config.set("no.ntnu.cerebrum.medulla.smcgp.realnumbersUpper", c.realNumbersUpper.toString());
        Config.set("no.ntnu.cerebrum.medulla.smcgp.realnumbersLower", c.realNumbersLower.toString());
        Config.set("no.ntnu.cerebrum.medulla.smcgp.outputs", c.getNumOutputNodes().toString());
        Config.set("no.ntnu.cerebrum.medulla.smcgp.maxGrowthFactor", c.maxGrowthFactor.toString());
        Config.set("no.ntnu.cerebrum.medulla.smcgp.useSimpleInputFunction", c.useSimpleInputFunction.toString());
        Config.set("no.ntnu.cerebrum.medulla.smcgp.maxNodesMovedOrDuplicated", c.maxNodesMovedOrDuplicated.toString());
        Config.set("no.ntnu.cerebrum.medulla.reservoir.numCachedTimeSteps", c.numCachedTimeSteps.toString());
        Config.set("no.ntnu.cerebrum.medulla.reservoir.useInputExpansion", c.useInputExpansion.toString());
        Config.set("no.ntnu.cerebrum.medulla.reservoir.inputExpansionFactor", c.inputExpansionFactor.toString());

        double gigaByte = Math.pow(1024, 3);
        double maxMem = Runtime.getRuntime().maxMemory();

        for (int i = 0; i < runsPerCandidate; i++) {
            String task = Config.get("no.ntnu.cerebrum.medulla.dataset.type");
            if (task != null && !task.equals("narma10") && !task.equals("narma30")) {
                task  = i == 0 ? "temporalParity" : "temporalDensity";
                if (i == 0) Config.set("no.ntnu.cerebrum.medulla.dataset.type", task);
                else Config.set("no.ntnu.cerebrum.medulla.dataset.type", task);
            }

            for (int ws = 0; ws < windowSizes.length; ws++) {
                for (int d = 0; d < delays.length; d++) {
                    if (!Config.getBoolean("no.ntnu.cerebrum.medulla.task.matrix.overWriteExisting") && c.matrixEvaluation[i][ws] != 0.0) {
                        log.info(String.format(
                            "Skipping %s %d-%d since %s has already done it",
                            task,
                            windowSizes[ws],
                            delays[d],
                            c
                        ));
                        continue;
                    }
                    String key = task + "-" + windowSizes[ws] + "-" + delays[d];

                    // Create the score collector key
                    if (!scores.containsKey(key)) {
                        scores.put(key, new DescriptiveStatistics());
                    }

                    log.info("Next evaluation: Task: " + task + " WS=" + windowSizes[ws] + " D=" + delays[d]);
                    log.info(c);

                    // Change data parameters and re-create the data sets
                    Config.set("no.ntnu.cerebrum.medulla.dataset.temporal.windowSize", Integer.toString(windowSizes[ws]));
                    Config.set("no.ntnu.cerebrum.medulla.dataset.temporal.delay", Integer.toString(delays[d]));
                    generateNewTrainingData();

                    Candidate result = evaluateGenotype(c.toPhenotype());

                    double usedMem = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
                    double allocation = (usedMem / maxMem) * 100.0;

                    log.info(String.format(
                        "Memory: %.2fGB / %.2fGB (%.1f%%)",
                        usedMem / gigaByte,
                        maxMem / gigaByte,
                        allocation
                    ));

                    if (result == null) {
                        log.warn("Evaluate phenotype returned 'null'");
                        continue;
                    } else {
                        assert c.getUuid().equals(result.getUuid());
                        scores.get(key).addValue(result.getTestSetFitness());
                        c.matrixEvaluation[i][ws] = result.getTestSetFitness();
                        c.save();
                    }

                    numMatrixEvaluations++;
                    log.info("Completed " + numMatrixEvaluations + " evaluations...");

                    printMatrixEvaluationScores(scores, windowSizes, delays);

                    // Force garbage collection
                    System.gc();
                }
            }
        }
    }

    public static void printMatrixEvaluationScores(HashMap<String, DescriptiveStatistics> scores, int[] windowSizes, int[] delays) {
        log.info(String.format(
            "--- SUMMARY (O: %d, IEF: %d) ---",
            Config.getInt("no.ntnu.cerebrum.medulla.smcgp.outputs"),
            Config.getInt("no.ntnu.cerebrum.medulla.reservoir.inputExpansionFactor")
        ));

        for (int i = 0; i < 2; i++) {
            String dataSet = i == 0 ? "temporalParity" : "temporalDensity";
            for (int ws = 0; ws < windowSizes.length; ws++) {
                for (int d = 0; d < delays.length; d++) {
                    String key = dataSet + "-" + windowSizes[ws] + "-" + delays[d];
                    if (!scores.containsKey(key)) continue;
                    log.info(String.format(
                        "%s %d-%d: %.3f/%.3f/%.3f-%.3f Results: %s",
                        dataSet,
                        windowSizes[ws],
                        delays[d],
                        scores.get(key).getMax(),
                        scores.get(key).getMean(),
                        scores.get(key).getMin(),
                        scores.get(key).getStandardDeviation(),
                        Arrays.toString(scores.get(key).getSortedValues())
                    ));
                }
            }
        }
    }

    private static void performNarmaWithCandidatesFromDisk() {
        String dataSetType = Config.get("no.ntnu.cerebrum.medulla.dataset.type");
        if (dataSetType == null) throw new IllegalStateException("DataSet type must be narma10 or narma30");

        ArrayList<DescriptiveStatistics> scores = new ArrayList<>();
        String path = Storage.BASE_PATH + Storage.CANDIDATES;
        Gson g = new Gson();

        scores.add(new DescriptiveStatistics());    // MSE
        scores.add(new DescriptiveStatistics());    // RMSE
        scores.add(new DescriptiveStatistics());    // NRMSE
        scores.add(new DescriptiveStatistics());    // RNMSE
        scores.add(new DescriptiveStatistics());    // MAE

        try (Stream<Path> paths = Files.walk(Paths.get(path))) {
            paths
                .filter(Files::isRegularFile)
                .filter(fp -> fp.toString().toLowerCase().endsWith("json"))
                .map(Storage::readFromFile)
                .map(c -> g.fromJson(c, Candidate.class))
                .forEach(c -> narmaEvaluation(c, scores, dataSetType));
        } catch (IOException e) {
            log.error(String.format(
                "Failed to read directory: %s (%s - %s)",
                path,
                e.getMessage(),
                e.getCause()
            ));
        }
    }

    public static void narmaEvaluation(Candidate c, ArrayList<DescriptiveStatistics> scores, String dataSetType) {
        int numRunsPerCandidate = Config.getInt("no.ntnu.cerebrum.medulla.narma.numRunsPerCandidate");
        boolean overWriteExisting = Config.getBoolean("no.ntnu.cerebrum.medulla.task.narma.overWriteExisting");

        // Set current configuration settings to match the one from the candidate
        Config.set("no.ntnu.cerebrum.medulla.smcgp.levelsback", c.levelsBack.toString());
        Config.set("no.ntnu.cerebrum.medulla.smcgp.realnumbersUpper", c.realNumbersUpper.toString());
        Config.set("no.ntnu.cerebrum.medulla.smcgp.realnumbersLower", c.realNumbersLower.toString());
        Config.set("no.ntnu.cerebrum.medulla.smcgp.outputs", c.getNumOutputNodes().toString());
        Config.set("no.ntnu.cerebrum.medulla.smcgp.maxGrowthFactor", c.maxGrowthFactor.toString());
        Config.set("no.ntnu.cerebrum.medulla.smcgp.useSimpleInputFunction", c.useSimpleInputFunction.toString());
        Config.set("no.ntnu.cerebrum.medulla.smcgp.maxNodesMovedOrDuplicated", c.maxNodesMovedOrDuplicated.toString());
        Config.set("no.ntnu.cerebrum.medulla.reservoir.numCachedTimeSteps", c.numCachedTimeSteps.toString());
        Config.set("no.ntnu.cerebrum.medulla.reservoir.useInputExpansion", c.useInputExpansion.toString());
        Config.set("no.ntnu.cerebrum.medulla.reservoir.inputExpansionFactor", c.inputExpansionFactor.toString());

        double gigaByte = Math.pow(1024, 3);
        double maxMem = Runtime.getRuntime().maxMemory();

        // Abort if this candidate contains results that are not going to be overwritten
        if (!overWriteExisting && c.getNarma10()[0] != 0.0) {
            log.info(String.format("Skipping %s as NARMA10 scores are already registered", c));
            return;
        }

        Candidate result = null;
        for (int i = 0; i < numRunsPerCandidate; i++) {
            log.info(String.format("Performing %s run %d", dataSetType, i + 1));
            log.info(c);
            if (result == null) result = evaluateGenotype(c.toPhenotype());
            else {
                Candidate tempResult = evaluateGenotype(c.toPhenotype());
                if (tempResult == null) continue;

                double[] sink;
                double[] narma;

                if (dataSetType.equals("narma10")) {
                    narma = tempResult.getNarma10();
                    sink = result.getNarma10();
                } else {
                    narma = tempResult.getNarma30();
                    sink = result.getNarma30();
                }

                sink[0] += narma[0];
                sink[1] += narma[1];
                sink[2] += narma[2];
                sink[3] += narma[3];
                sink[4] += narma[4];
            }

            double usedMem = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
            double allocation = (usedMem / maxMem) * 100.0;

            log.info(String.format(
                "Memory: %.2fGB / %.2fGB (%.1f%%)",
                usedMem / gigaByte,
                maxMem / gigaByte,
                allocation
            ));

            numMatrixEvaluations++;
            log.info(String.format("Completed %d %s evaluations", numMatrixEvaluations, dataSetType));

            System.gc();
        }

        if (result == null) return;
        double[][] narma = result.getNarma();
        for (int x = 0; x < 5; x++) {
            if (dataSetType.equals("narma10")) {
                narma[0][x] /= numRunsPerCandidate;
                scores.get(x).addValue(narma[0][x]);
            } else {
                narma[1][x] /= numRunsPerCandidate;
                scores.get(x).addValue(narma[1][x]);
            }
        }

        log.info(String.format(
            "--- SUMMARY (O: %d, IEF: %d) ---",
            Config.getInt("no.ntnu.cerebrum.medulla.smcgp.outputs"),
            Config.getInt("no.ntnu.cerebrum.medulla.reservoir.inputExpansionFactor")
        ));
        for (int z = 0; z < 5; z++) {
            String type = "MSE";
            if (z == 1) type = "RMSE";
            else if (z == 2) type = "NRMSE";
            else if (z == 3) type = "RNMSE";
            else if (z == 4) type = "MAE";

            log.info(String.format(
                "%s: %.5f/%.5f/%.5f±%.5f %s",
                type,
                scores.get(z).getMin(),
                scores.get(z).getMean(),
                scores.get(z).getMax(),
                scores.get(z).getStandardDeviation(),
                Arrays.toString(scores.get(z).getSortedValues())
            ));
        }

        c.setNarma(result.getNarma());
        c.save();

    }

    public static String toString(double[][] arr) {
        StringBuilder out = new StringBuilder();
        out.append("[");
        for (int i = 0; i < arr.length; i++) {
            if (i == 0) {
                out.append(Arrays.toString(arr[i]));
                out.append(",");
            }
            else if (i == arr.length - 1) {
                out.append(" ");
                out.append(Arrays.toString(arr[i]));
            }
            else {
                out.append(" ");
                out.append(Arrays.toString(arr[i]));
                out.append(",");
            }
        }
        out.append("]");

        return out.toString();
    }

    private static Phenotype2 miniEvolution() {
        long startTime = System.currentTimeMillis();
        double mutationRate = Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.ea.mutationRate");

        ArrayList<Phenotype2> pool = new ArrayList<>();
        Phenotype2 best = new Phenotype2(new Genotype2());
        pool.add(best);

        int evaluations = 0;

        while (true) {
            pool.add(new Phenotype2(new Genotype2(best.getGenotype().copyGenes())));
            while (pool.size() < 6) {
                Genotype2 g = new Genotype2(best.getGenotype().copyGenes());
                g.mutate(mutationRate, true);
                pool.add(new Phenotype2(g));
            }

            List<Future> tasks = new ArrayList<>();
            for (Phenotype2 phenotype : pool) {
                tasks.add(Evolve.es.submit(() -> {
                    phenotype.setFitness(phenotype.miniScore(16, 24, 48));
                }));
            }

            try {
                for (Future f : tasks) {
                    f.get();
                }
                tasks.clear();
            } catch (InterruptedException e) {
                continue;
            } catch (ExecutionException e) {
                continue;
            }

            SummaryStatistics stats = new SummaryStatistics();

            for (Phenotype2 p : pool) {
                double f = p.getFitness();
                stats.addValue(f);
                evaluations++;
                if (f >= best.getFitness()) best = p;

                log.info(String.format(
                    "[%d] Best: %.2f Population{%.2f/%.2f/%.2f±%.3f}",
                    evaluations,
                    best.getFitness(),
                    stats.getMax(),
                    stats.getMean(),
                    stats.getMin(),
                    stats.getStandardDeviation()
                ));

                if (f == 1.0) {
                    double evalTime = (System.currentTimeMillis() - startTime) / 1000.0;
                    double evalSpeed = evaluations * 1000.0 / (System.currentTimeMillis() - startTime);
                    log.info(String.format(
                        "Found candidate after %.1fs and %d attempts (%.1f e/s)",
                        evalTime,
                        evaluations,
                        evalSpeed
                    ));

                    return p;
                }
            }
            
            pool.clear();
        }
    }

    public static void fastLinearRegressionTester() {
        DescriptiveStatistics stats = new DescriptiveStatistics();
        SummaryStatistics time = new SummaryStatistics();
        int numRuns = 1001;
        int outputs = Config.getInt("no.ntnu.cerebrum.medulla.smcgp.outputs");

        double found = 0;
        for (int i = 1; i < numRuns; i++) {
            Long startTime = System.currentTimeMillis();

            Phenotype2 p = miniEvolution();
            SMCGPReservoir r = new SMCGPReservoir(p);
            r.readoutLayer(new SingleLayerLinearRegression(
                outputs,
                1)
            );

            Candidate c = evaluateGenotype(p);
            if (c == null) continue;
            double f = c.getTestSetFitness();
            c.save();

            stats.addValue(f);
            time.addValue(System.currentTimeMillis() - startTime);

            LocalDateTime done = LocalDateTime.now();
            int secondsPerEvaluation = (int) (time.getMean() / 1000.0);
            done = done.plus(secondsPerEvaluation * (numRuns - i), ChronoUnit.SECONDS);
            log.info(String.format(
                "[%d/%.1f%%] %.2f/%.2f/%.2f±%.3f [%.2f|%.1f%%] (%.1f p/min, eta %s)",
                i,
                i * 100.0 / numRuns,
                stats.getMax(),
                stats.getMean(),
                stats.getMin(),
                stats.getStandardDeviation(),
                stats.getGeometricMean(),
                (found / i) * 100.0,
                60.0 / secondsPerEvaluation,
                done
            ));
        }
        log.info(Arrays.toString(stats.getSortedValues()));
    }

    public static void warmUpBlas() {
        Phenotype2 p = new Phenotype2(new Genotype2());
        SMCGPReservoir r = new SMCGPReservoir(p);
        int readoutOutputs = 2;
        String dataSetType = Config.get("no.ntnu.cerebrum.medulla.dataset.type");
        if (dataSetType != null && (dataSetType.equals("narma10") || dataSetType.equals("narma30"))) readoutOutputs = 1;
        r.readoutLayer(new MultiLayerClassifier(Config.getInt("no.ntnu.cerebrum.medulla.smcgp.outputs"), readoutOutputs));
        org.nd4j.linalg.dataset.api.DataSet train = buildDataSet(p, r, trainingData);
        r.train(1, train);
    }

    public static void reCalculatePowerSpectrumDensity() {
        log.info("Loading all candidates");
        List<Candidate> candidates = Candidate.loadAll();
        log.info(String.format("Re-calculating power spectrum density for %d candidates", candidates.size()));
        for (Candidate c : candidates) {
            Config.set("no.ntnu.cerebrum.medulla.smcgp.levelsback", c.levelsBack.toString());
            Config.set("no.ntnu.cerebrum.medulla.smcgp.realnumbersUpper", c.realNumbersUpper.toString());
            Config.set("no.ntnu.cerebrum.medulla.smcgp.realnumbersLower", c.realNumbersLower.toString());
            Config.set("no.ntnu.cerebrum.medulla.smcgp.outputs", c.getNumOutputNodes().toString());
            Config.set("no.ntnu.cerebrum.medulla.smcgp.maxGrowthFactor", c.maxGrowthFactor.toString());
            Config.set("no.ntnu.cerebrum.medulla.smcgp.useSimpleInputFunction", c.useSimpleInputFunction.toString());
            Config.set("no.ntnu.cerebrum.medulla.smcgp.maxNodesMovedOrDuplicated", c.maxNodesMovedOrDuplicated.toString());
            Config.set("no.ntnu.cerebrum.medulla.reservoir.numCachedTimeSteps", c.numCachedTimeSteps.toString());
            Config.set("no.ntnu.cerebrum.medulla.reservoir.useInputExpansion", c.useInputExpansion.toString());
            Config.set("no.ntnu.cerebrum.medulla.reservoir.inputExpansionFactor", c.inputExpansionFactor.toString());

            try {
                c.generatePowerSpectrum();
                c.save();
            } catch (Exception e) {
                log.error("Caught exception: " + e.getMessage());
                c.powerSpectrum = new double[][]{};
                c.save();
            }
        }
    }

    public static void main(String[] args) {
        // Verify that configuration files are present, or create from defaults.
        Config.bootstrapConfigurationFiles();
        PropertyConfigurator.configure("config/log4j.properties");
        Config.loadConfiguration();

        generateNewTrainingData();

        String task = Config.get("no.ntnu.cerebrum.medulla.task");
        if (task == null) task = "generateNew";

        switch (task) {
            case "generateNew":
                while (true) {
                    try {
                        generateNewCandidates();
                    } catch (Exception e) {
                        e.printStackTrace();
                        log.error("Infinity value caused JSON serialization to fail. Restarting...");
                    }
                }
            case "rerun":
                if (!Config.getBoolean("no.ntnu.cerebrum.medulla.dataset.dumpToCSV")) {
                    warmUpBlas();
                    System.out.println("Warm-up complete");
                }
                reEvaluateCandidatesFromDisk();
                break;
            case "matrix":
                if (!Config.getBoolean("no.ntnu.cerebrum.medulla.dataset.dumpToCSV")) {
                    warmUpBlas();
                    System.out.println("Warm-up complete");
                }
                performMatrixEvaluationWithCandidatesFromDisk();
                break;
            case "narma":
                if (!Config.getBoolean("no.ntnu.cerebrum.medulla.dataset.dumpToCSV")) {
                    warmUpBlas();
                    System.out.println("Warm-up complete");
                }
                performNarmaWithCandidatesFromDisk();
                break;
            case "randomRegression":
                fastLinearRegressionTester();
                break;
            case "powerSpectrum":
                reCalculatePowerSpectrumDensity();
                break;
        }
    }
}
