package no.ntnu.cerebrum.medulla.config;

import org.slf4j.Logger;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

public class Config {
    private static Properties properties;
    private static Logger log = org.slf4j.LoggerFactory.getLogger(Config.class);

    /**
     * Retrieve the value of a configuration option.
     * If environment variables are set that match the specified configuration option, the environment variable value
     * will take precedence over the config file value.
     *
     * @param key The name of the config property.
     * @return * The value of the config property.
     */
    public static String get(String key) {
        if (properties != null) {
            if (!key.contains("no.ntnu.cerebrum.medulla.")) key = "no.ntnu.cerebrum.medulla." + key;

            // Return value from environment variable if it exists
            String envKey = System.getenv(translateToEnvVarKey(key));
            if (envKey != null) return envKey;

            return properties.getProperty(key);
        } else {
            log.warn("Attempting to fetch property from uninitialized config file");
            return null;
        }
    }

    /**
     * Overwrites a currently loaded config setting. NB: These changes are not reflected in the config file on disk,
     * and only last for the duration of the application.
     * @param key The config option name
     * @param value The config option value
     */
    public static void set(String key, String value) {
        properties.setProperty(key, value);
    }

    /**
     * Translates the
     * @param key A properties-style configuration parameter key
     * @return An environment variable version of the provided key
     */
    public static String translateToEnvVarKey(String key) {
        key = key.replace("no.ntnu.cerebrum.medulla.", "");
        key = key.replace(".", "_");
        key = key.toUpperCase();

        // Env vars will be namespaced
        return "MEDULLA_" + key;
    }

    /**
     * Retrieve the double value of a configuration option.
     *
     * @param key The name of the config property.
     * @return * The double value of the config property.
     */
    public static Double getDouble(String key) {
        return Double.parseDouble(Config.get(key));
    }

    /**
     * Retrieve the int value of a configuration option.
     *
     * @param key The name of the config property.
     * @return * The int value of the config property.
     */
    public static int getInt(String key) {
        return Integer.parseInt(Config.get(key));
    }

    /**
     * Retrieve the float value of a configuration option.
     *
     * @param key The name of the config property.
     * @return * The float value of the config property.
     */
    public static Float getFloat(String key) {
        return Float.parseFloat(Config.get(key));
    }

    /**
     * Retrieve the boolean value of a configuration option.
     *
     * @param key The name of the config property.
     * @return * The float value of the config property.
     */
    public static Boolean getBoolean(String key) {
        return Boolean.parseBoolean(Config.get(key));
    }


    /**
     * Retrieve the boolean value of a configuration option.
     *
     * @param key The name of the config property.
     * @return * The float value of the config property.
     */
    public static String[] getList(String key) {
        String list = get(key);
        if (list != null) {
            return list.split(",");
        }
        return null;
    }

    /**
     * Override config with supplied key, value pairs
     */
    public static void setConfig(Map<String, String> newConfig) {
        if (properties == null) {
            log.warn("Attempting to fetch property from uninitialized config file");
        }

        for (String key : newConfig.keySet()) {
            if (get(key) != null) {
                properties.setProperty(key, newConfig.get(key));
            } else {
                log.warn(String.format("KeyError: <%s> is not a valid configuration option", key));
            }
        }
    }

    /**
     * Get all current config
     */
    public static Map<String, String> getAllConfig() {
        if (properties == null) {
            log.warn("Attempting to fetch property from uninitialized config file");
        }

        Map<String, String> allConfig = new HashMap();
        for (Map.Entry<Object, Object> configItem : properties.entrySet()) {
            allConfig.put(configItem.getKey().toString(), configItem.getValue().toString());
        }
        return allConfig;
    }

    /**
     * Loads or reloads the configuration options in "config/medulla.properties".
     */
    public static void loadConfiguration() {

        if (Config.properties == null) {
            log.info("Loading configuration properties ...");
            Config.properties = new Properties();
        } else {
            log.info("Reloading configuration properties ...");
            Config.properties.clear();
        }

        try {
            Config.properties.load(new FileInputStream("config/medulla.properties"));
        } catch (IOException e) {
            log.error(
                    String.format("Failed to load configuration file \"config/medulla.properties\": %s", e.getMessage())
            );
        }
    }

    /**
     * Checks that the necessary configuration directories and files are present, or creates them based on the
     * provided defaults if they are missing.
     */
    public static void bootstrapConfigurationFiles() {
        // If the config directory does not exist, create it
        File configDir = new File("config");
        if (!configDir.exists()) {
            if (!configDir.mkdirs()) {
                System.err.println("Failed to create configuration directories, exiting...");
                System.exit(1);
            }
        }

        // Declare the config files
        File medullaConfig = new File("config/medulla.properties");
        File log4jConfig = new File("config/log4j.properties");

        try {
            // If medulla.properties does not exist, copy it from the classpath resources folder
            if (!medullaConfig.exists()) {
                InputStream baseMedullaConfig = Config.class.getResourceAsStream("/config/medulla.properties");
                Files.copy(baseMedullaConfig, medullaConfig.toPath());
            }
            // If log4j.properties does not exist, copy it from the classpath resources folder
            if (!log4jConfig.exists()) {
                InputStream baseLog4jConfig = Config.class.getResourceAsStream("/config/log4j.properties");
                Files.copy(baseLog4jConfig, log4jConfig.toPath());
            }
        } catch (IOException e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
}
