package no.ntnu.cerebrum.medulla.data;

import com.google.gson.Gson;
import no.ntnu.cerebrum.medulla.common.TreeNode;
import no.ntnu.cerebrum.medulla.config.Config;
import no.ntnu.cerebrum.medulla.reservoir.SMCGPReservoir;
import no.ntnu.cerebrum.medulla.smcgp.Genotype2;
import no.ntnu.cerebrum.medulla.smcgp.Phenotype2;
import no.ntnu.cerebrum.medulla.smcgp.graph.action.FunctionLookup;
import no.ntnu.cerebrum.medulla.telemetry.ReservoirStatsCollector;
import org.apache.commons.math3.complex.Complex;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.nio.file.Path;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.TimeoutException;
import java.util.stream.Collectors;

/**
 * Created 2017-04-07.
 * <p>
 * cerebrum-medulla is licenced under the MIT licence.
 */
public class Candidate {
    /**
     * Candidate objects contain all relevant information related to a specific phenotype, like its fitness on training
     * and test data sets, raw output, power spectrum density, and what window size and signal delay was used when
     * tested with the temporal data sets.
     */
    public static Logger log = LoggerFactory.getLogger(Candidate.class);

    public String uuid = null;
    public String timeStamp = null;
    public double[][] genotype = null;
    public String functionSet = "all";

    public Integer windowSize = 0;
    public Integer signalDelay = 0;

    public Double testSetFitness = 0.0;
    public Double trainingSetFitness = 0.0;

    public Integer epochsTrained = 0;
    public Integer hiddenLayerOneNodeCount = 0;
    public Integer batchSize = 0;
    public Integer numCachedTimeSteps = 0;

    public Double realNumbersLower = 0.0;
    public Double realNumbersUpper = 0.0;
    public Integer levelsBack = 0;
    public Integer numOutputNodes = 1;
    public Boolean useInputExpansion = false;
    public Integer inputExpansionFactor = 3;
    public Boolean useSimpleInputFunction = false;
    public Double maxGrowthFactor = 3.0;
    public Integer maxNodesMovedOrDuplicated = 50;
    public Double sineWaveFrequency = 1.0;
    public Double sineWaveOverSamplingFactor = 64.0;
    public Double sineWaveSeconds = 2.0;

    public double[] rawOutput = null;
    public double[][] rawOutput2D = null;
    public double[][] powerSpectrum = null;
    public double[][] rawSineOutput = null;
    public int[][] stateHistory = null;
    public double[][] matrixEvaluation = null;
    public double[][] narma = null;

    /**
     * Initializes a blank Candidate with a new randomly generated UUID.
     */
    public Candidate() {
        genotype = new double[][]{};
        rawOutput = new double[]{};
        rawOutput2D = new double[][]{};
        rawSineOutput = new double[][]{};
        powerSpectrum = new double[][]{};
        stateHistory = new int[][]{};
        matrixEvaluation = new double[2][3];
        narma = new double[2][5];
    }

    /**
     * Initializes a Candidate from a Phenotype2 instance.
     * @param phenotype An evolved phenotype.
     */
    public Candidate(Phenotype2 phenotype) {
        init();

        genotype = phenotype.getGenotype().copyGenes();
        testSetFitness = phenotype.getFitness();
        trainingSetFitness = phenotype.getTrainingSetFitness();
        uuid = phenotype.getUuid();
        narma = new double[2][5];
        matrixEvaluation = new double[2][3];

        try {
            generatePowerSpectrum();
            generateStateHistory();
        } catch (TimeoutException e) {
            log.error("Phenotype timed out!");
        }
    }

    /**
     * Initializes required values that need to be stored with the candidate data.
     */
    private void init() {
        timeStamp = LocalDateTime.now().format(DateTimeFormatter.ISO_LOCAL_DATE_TIME);
        functionSet = Config.get("no.ntnu.cerebrum.medulla.smcgp.functionMath.whiteList");
        windowSize = Config.getInt("no.ntnu.cerebrum.medulla.dataset.temporal.windowSize");
        signalDelay = Config.getInt("no.ntnu.cerebrum.medulla.dataset.temporal.delay");
        numCachedTimeSteps = Config.getInt("no.ntnu.cerebrum.medulla.reservoir.numCachedTimeSteps");
        hiddenLayerOneNodeCount = Config.getInt("no.ntnu.cerebrum.medulla.reservoir.readout.hiddenLayerOneNodeCount");
        batchSize = Config.getInt("no.ntnu.cerebrum.medulla.reservoir.readout.batchSize");
        realNumbersLower = Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.realnumbersLower");
        realNumbersUpper = Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.realnumbersUpper");
        levelsBack = Config.getInt("no.ntnu.cerebrum.medulla.smcgp.levelsback");
        numOutputNodes = Config.getInt("no.ntnu.cerebrum.medulla.smcgp.outputs");
        useInputExpansion = Config.getBoolean("no.ntnu.cerebrum.medulla.reservoir.useInputExpansion");
        inputExpansionFactor = Config.getInt("no.ntnu.cerebrum.medulla.reservoir.inputExpansionFactor");
        useSimpleInputFunction = Config.getBoolean("no.ntnu.cerebrum.medulla.smcgp.useSimpleInputFunction");
        maxGrowthFactor = Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.maxGrowthFactor");
        maxNodesMovedOrDuplicated = Config.getInt("no.ntnu.cerebrum.medulla.smcgp.maxNodesMovedOrDuplicated");
        sineWaveFrequency = Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.frequency");
        sineWaveOverSamplingFactor = Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.overSamplingFactor");
        sineWaveSeconds = Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.seconds");
    }

    /**
     * Loads a Candidate object from disk, searching for the specified uuid.
     * @param uuid The UUID of the file to search for.
     * @return A Candidate object parsed from JSON data if UUID is found, null otherwise.
     */
    public static Candidate load(String uuid) {
        Path p = Storage.search(uuid + ".json");
        if (p == null) return null;

        String data = Storage.readFromFile(p);
        Gson g = new Gson();

        return g.fromJson(data, Candidate.class);
    }

    /**
     * Saves this candidate as a JSON file on disk.
     * @return Returns this object instance.
     */
    public Candidate save() {
        Gson g = new Gson();
        String serializedData = g.toJson(this);

        Storage.writeJSON(this.uuid, serializedData);

        log.info(String.format(
            "Saved candidate %s with fitness %.5f (%.5f)!",
            uuid.substring(0, 8),
            testSetFitness,
            trainingSetFitness
        ));

        return this;
    }

    /**
     * Deletes this Candidate from disk
     */
    public void delete() {
        String path = Storage.BASE_PATH + Storage.CANDIDATES + uuid + ".json";
        File f = new File(path);
        if (f.delete()) {
            log.info("Candidate " + uuid.substring(0, 8) + " was deleted!");
        } else {
            log.error("Failed to delete candidate " + uuid.substring(0, 8) + ": " + path);
        }
    }

    /**
     * Traverses the storage directory on disk and loads all Candidates found.
     * @return A List of Candidates.
     */
    public static List<Candidate> loadAll() {
        String path = Storage.BASE_PATH + Storage.CANDIDATES;
        Gson g = new Gson();
        return Storage.readAllFiles(path, "json").stream()
            .map(c -> g.fromJson(c, Candidate.class))
            .collect(Collectors.toList());
    }

    public String getUuid() {
        return uuid;
    }

    public void setUuid(String uuid) {
        this.uuid = uuid;
    }

    public String getTimeStamp() {
        return timeStamp;
    }

    public void setTimeStamp(String timeStamp) {
        this.timeStamp = timeStamp;
    }

    public double[][] getGenotype() {
        return genotype;
    }

    public void setGenotype(double[][] genotype) {
        this.genotype = genotype;
    }

    public String getFunctionSet() { return this.functionSet; }

    public void setFunctionSet(String functionSet) { this.functionSet = functionSet; }

    public Integer getWindowSize() {
        return windowSize;
    }

    public void setWindowSize(Integer windowSize) {
        this.windowSize = windowSize;
    }

    public Integer getSignalDelay() {
        return signalDelay;
    }

    public void setSignalDelay(Integer signalDelay) {
        this.signalDelay = signalDelay;
    }

    public Double getTestSetFitness() {
        return testSetFitness;
    }

    public void setTestSetFitness(Double testSetFitness) {
        this.testSetFitness = testSetFitness;
    }

    public Double getTrainingSetFitness() {
        return trainingSetFitness;
    }

    public void setTrainingSetFitness(Double trainingSetFitness) {
        this.trainingSetFitness = trainingSetFitness;
    }

    public Integer getEpochsTrained() {
        return epochsTrained;
    }

    public void setEpochsTrained(Integer epochsTrained) {
        this.epochsTrained = epochsTrained;
    }

    public Integer getHiddenLayerOneNodeCount() {
        return hiddenLayerOneNodeCount;
    }

    public void setHiddenLayerOneNodeCount(Integer hiddenLayerOneNodeCount) {
        this.hiddenLayerOneNodeCount = hiddenLayerOneNodeCount;
    }

    public Integer getBatchSize() {
        return batchSize;
    }

    public void setBatchSize(Integer batchSize) {
        this.batchSize = batchSize;
    }

    public Integer getNumOutputNodes() {
        return numOutputNodes;
    }

    public void setNumOutpuNodes(Integer outputNodes) {
        this.numOutputNodes = outputNodes;
    }

    public Integer getNumCachedTimeSteps() {
        return numCachedTimeSteps;
    }

    public void setNumCachedTimeSteps(Integer numCachedTimeSteps) {
        this.numCachedTimeSteps = numCachedTimeSteps;
    }

    public Boolean isUsingInputExpansion() {
        return this.useInputExpansion;
    }

    public void setUsingInputExpansion(Boolean useInputExpansion) {
        this.useInputExpansion = useInputExpansion;
    }

    public double[] getRawOutput() {
        return rawOutput;
    }

    public void setRawOutput(double[] rawOutput) {
        this.rawOutput = rawOutput;
    }

    public double[][] getRawOutput2D() {
        return rawOutput2D;
    }

    public void setRawOutput2D(double[][] rawOutput2D) {
        this.rawOutput2D = rawOutput2D;
    }

    public double[][] getRawSineOutput() {
        return rawSineOutput;
    }

    public void setRawSineOutput(double[][] rawSineOutput) {
        this.rawSineOutput = rawSineOutput;
    }

    public double[][] getPowerSpectrum() {
        return powerSpectrum;
    }

    public void setPowerSpectrum(double[][] powerSpectrum) {
        this.powerSpectrum = powerSpectrum;
    }

    public int[][] getStateHistory() {
        return stateHistory;
    }

    public void setStateHistory(int[][] stateHistory) {
        this.stateHistory = stateHistory;
    }

    public double[][] getNarma() { return this.narma; }

    public void setNarma(double[][] narma) { this.narma = narma; }

    public double[] getNarma10() { return this.narma[0]; }

    public double[] getNarma30() { return this.narma[1]; }

    public String getNarma10AsString() {
        return String.format(
            "MSE: %.5f RMSE: %.5f NRMSE: %.5f RNMSE: %.5f MEA: %.5f",
            narma[0][0],
            narma[0][1],
            narma[0][2],
            narma[0][3],
            narma[0][4]
        );
    }

    public String getNarma30AsString() {
        return String.format(
            "MSE: %.5f RMSE: %.5f NRMSE: %.5f RNMSE: %.5f MEA: %.5f",
            narma[1][0],
            narma[1][1],
            narma[1][2],
            narma[1][3],
            narma[1][4]
        );
    }

    /**
     * Generates a power spectrum of this candidate, by transforming this Candidate into a Phenotype2, and then
     * spawning a SMCGP Reservoir and ReservoirDataSet, before feeding a sine series into the candidate Phenotype
     * and performing FFT on the output, and finally generating the power density spectrum.
     * @return An array of real values representing the signal strength in each frequency bin.
     */
    public double[][] generatePowerSpectrum() throws TimeoutException {
        // Make reservoir and create standard data set.
        SMCGPReservoir r = new SMCGPReservoir(this.toPhenotype());
        int numOutputs = Config.getInt("no.ntnu.cerebrum.medulla.smcgp.outputs");

        double[] sineData = DataFactory.generateSineSeries(
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.frequency"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.amplitude"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.phase"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.overSamplingFactor"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.seconds")
        );
        // Make a 2D version if we are using multi-output preProcessing
        double[][] sineData2D = new double[sineData.length][1];
        for (int j = 0; j < sineData.length; j++) {
            sineData2D[j][0] = sineData[j];
        }

        powerSpectrum = new double[numOutputs][];
        rawSineOutput = r.preProcess(sineData2D);
        for (int i = 0; i < numOutputs; i++) {
            Complex[] result = Fourier.forwardDFT(DataFactory.getColumn(i, rawSineOutput));
            // We drop the second half the output since we are only dealing with real valued numbers
            powerSpectrum[i] = Arrays.copyOf(Evaluation.powerSpectrumDensity(result, 1.0), result.length / 2);
        }

        sineWaveFrequency = Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.frequency");
        sineWaveOverSamplingFactor = Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.overSamplingFactor");
        sineWaveSeconds = Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.seconds");

        return powerSpectrum;
    }

    /**
     * Creates a SMCGPReservoir and runs some data sets through it to record state history.
     * @return A matrix consisting of the state histories.
     */
    public int[][] generateStateHistory() {
        return generateStateHistory(8, 128);
    }

    /**
     * Creates a SMCGPReservoir and runs some data sets through it to record state history.
     * @param runs The number of runs to generate state history of
     * @param length The length of each bit stream
     * @return A matrix consisting of the state histories.
     */
    public int[][] generateStateHistory(int runs, int length) {
        Phenotype2 p = this.toPhenotype();
        SMCGPReservoir r = new SMCGPReservoir(p);
        stateHistory = new int[runs][length];
        List<Data> dataSets = DataFactory.createTemporalParityDataWithEvenParityBitDistribution(runs, 3, 0, runs * length, true);

        try {
            for (int i = 0; i < dataSets.size(); i++) {
                ReservoirStatsCollector rsc = new ReservoirStatsCollector();
                p.addSMCGPProcessingListener(rsc);

                r.preProcess(dataSets.get(i).input());
                p.reset();

                p.removeSMCGPProcessingListener(rsc);
                stateHistory[i] = rsc.getActiveNodeStateHistoryAsArray();
            }
        } catch (TimeoutException e) {
            log.error("Phenotype timed out while generating state history!");
        }

        return stateHistory;
    }

    /**
     * Creates a SMCGPReservoir and runs uniform data (just 1's or just 0's) through it to record state history.
     * @param runs The number of runs to generate state history of
     * @param length The length of each bit stream
     * @return A matrix consisting of the state histories.
     */
    public int[][] generateStaticStateHistory(int runs, int length) {
        Phenotype2 p = this.toPhenotype();
        SMCGPReservoir r = new SMCGPReservoir(p);
        stateHistory = new int[runs][length];

        double[][] zeroes = new double[length][1];
        double[][] ones = new double[length][1];
        for (int i = 0; i < length; i++) ones[i] = new double[]{1.0};

        try {
            for (int i = 0; i < runs; i++) {
                ReservoirStatsCollector rsc = new ReservoirStatsCollector();
                p.addSMCGPProcessingListener(rsc);

                if (i < runs / 2) r.preProcess(zeroes);
                else r.preProcess(ones);
                p.reset();

                p.removeSMCGPProcessingListener(rsc);
                stateHistory[i] = rsc.getActiveNodeStateHistoryAsArray();
            }
        } catch (TimeoutException e) {
            log.error("Phenotype timed out while generating state history!");
        }

        return stateHistory;
    }

    /**
     * Creates a tree structure of String labels that can be printed to console.
     * @param rootNode The index from which to start the recursive depth-first search.
     * @return A TreeNode object of the root node, with all descendants.
     */
    public TreeNode genotypeTree(int rootNode) {
        // Out of bounds references yields zero-input
        if (rootNode < 0) return new TreeNode(String.format("ZeroInput (%d)", rootNode), new ArrayList<>());

        int funcType = (int) genotype[rootNode][0];
        ArrayList<TreeNode> children = new ArrayList<>();

        // Input nodes are terminated
        if (0 <= funcType && funcType < 3) {
            return new TreeNode(String.format(
                "%s (%d)",
                FunctionLookup.getFunction(funcType),
                rootNode),
                children
            );
        }

        // Wrap these two lines in an if check for funcType > 2 to stop at input function types,
        // rather than go all the way to index zero.
        children.add(genotypeTree(rootNode - ((int) genotype[rootNode][1])));
        children.add(genotypeTree(rootNode - ((int) genotype[rootNode][2])));

        return new TreeNode(String.format(
            "%s (%d)",
            FunctionLookup.getFunction(funcType),
            rootNode),
            children
        );
    }

    /**
     * Convert this Candidate to a Phenotype2 representation.
     * @return A Phenotype2 with the Genotype of this Candidate object.
     */
    public Phenotype2 toPhenotype() {
        Phenotype2 p = new Phenotype2(new Genotype2(genotype));

        p.setFitness(testSetFitness);
        p.setTrainingSetFitness(trainingSetFitness);
        p.setUuid(uuid);

        return p;
    }

    /**
     * String representation containing relevant info about this Candidate.
     * @return A string containing most relevant Candidate info.
     */
    public String toString() {
        return String.format(
            "Candidate{%s - F:%.4f/%.4f [O:%d IEF:%d/H1:%d/TS:%d] (WS:%d D:%d)}",
            uuid.substring(0, 8),
            testSetFitness,
            trainingSetFitness,
            numOutputNodes,
            inputExpansionFactor,
            hiddenLayerOneNodeCount,
            numCachedTimeSteps,
            windowSize,
            signalDelay
        );
    }

    /**
     * Multi-line string representation containing relevant info about this Candidate, including genotype node tree.
     * @return A String containing relevant Candidate info.
     */
    public String toStringWithNodeTree() {
        int root = -1;
        for (int i = genotype.length - 1; i >= 0; i--) {
            if (genotype[i][genotype[i].length - 1] == 1.0) {
                root = i;
                break;
            }
        }

        return String.format("%s\n%s", toString(), genotypeTree(root));
    }
}
