package no.ntnu.cerebrum.medulla.data;

import com.google.gson.Gson;
import org.apache.commons.math3.stat.descriptive.DescriptiveStatistics;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.api.rng.Random;
import org.nd4j.linalg.factory.Nd4j;

import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;

/**
 * Created 2016-12-14.
 * <p>
 * cerebrum-medulla is licenced under the MIT licence.
 */
public class DataFactory {
    /**
     * Populates an input array with bits to form a bit stream, and populates a target array with
     * the resulting parity bits based on the provided size of the sliding window and signal delay.
     * @param input An array to store generated input data bit stream
     * @param target An array to store resulting target bit stream
     * @param windowSize The size of the sliding window that should be used to calculate parity
     * @param delay How many bits back in time should the parity calculation be offset with
     */
    public static void createRandomTemporalParityData(double[][] input, double[][] target,
                                                      int windowSize, int delay, int numSamples, boolean oneHot) {

        if (windowSize < 1) throw new IllegalArgumentException("Sliding window size must be 1 or greater.");
        if (delay < 0) throw new IllegalArgumentException("Signal delay cannot be negative");
        if (input.length != target.length) throw new IllegalArgumentException("Input and target arrays must have equal sizes");

        // Fill the input array with an evenly distributed number of bits
        generateRandomBits(input, numSamples, true);

        shuffle(input);

        // Calculate target labels
        for (int i = 0; i < numSamples; i++) {
            int to = (i + 1 - delay);
            int from = to - windowSize;

            if (to <= 0) to = 1;
            if (from < 0) from = 0;

            int total = 0;
            for (int x = from; x < to; x++) {
                total += input[x][0];
            }

            if (total % 2 == 0) {
                if (oneHot) {
                    target[i] = new double[]{1.0, 0.0};
                } else {
                    target[i] = new double[]{0.0};
                }
            } else {
                if (oneHot) {
                    target[i] = new double[]{0.0, 1.0};
                } else {
                    target[i] = new double[]{1.0};
                }
            }
        }
    }

    /**
     * Populates an input array with bits to form a bit stream, and populates a target array with
     * the resulting density bits based on the provided size of the sliding window and signal delay.
     * @param input An array to store generated input data bit stream
     * @param target An array to store resulting target bit stream
     * @param windowSize The size of the sliding window that should be used to calculate density
     * @param delay How many bits back in time should the density calculation be offset with
     */
    public static void createRandomTemporalDensityData(double[][] input, double[][] target,
                                                      int windowSize, int delay, int numSamples, boolean oneHot) {

        if (windowSize < 1) throw new IllegalArgumentException("Sliding window size must be 1 or greater.");
        if (delay < 0) throw new IllegalArgumentException("Signal delay cannot be negative");
        if (input.length != target.length) throw new IllegalArgumentException("Input and target arrays must have equal sizes");
        if (windowSize % 2 == 0) throw new IllegalArgumentException("The window size must be odd for temporal density");

        generateRandomBits(input, numSamples, true);

        shuffle(input);

        // Calculate target labels
        for (int i = 0; i < numSamples; i++) {
            int to = (i + 1 - delay);
            int from = to - windowSize;

            if (to <= 0) to = 1;
            if (from < 0) from = 0;

            int total = 0;
            for (int x = from; x < to; x++) {
                total += input[x][0];
            }

            if (total > windowSize / 2.0) {
                if (oneHot) {
                    target[i] = new double[]{0.0, 1.0};
                } else {
                    target[i] = new double[]{0.0};
                }
            } else {
                if (oneHot) {
                    target[i] = new double[]{1.0, 0.0};
                } else {
                    target[i] = new double[]{1.0};
                }
            }
        }
    }

    /**
     * Generate a random stream of bits in a 2D array.
     * @param input The input array
     * @param numSamples Total number of bits in array
     * @param forceEvenDistribution Force an even distribution of ones and zeroes. If odd numbered, last is random.
     */
    private static void generateRandomBits(double[][] input, int numSamples, boolean forceEvenDistribution) {
        if (input.length != numSamples) throw new IllegalArgumentException("Input array not correctly sized");
        Random rng = Nd4j.getRandom();

        // Seed the input array with an equal distribution of ones and zeroes
        if (forceEvenDistribution) {
            int uniqueCount = numSamples / 2;
            for (int i = 0; i < uniqueCount; i++) {
                input[i] = new double[]{0.0};
            }
            for (int i = uniqueCount; i < uniqueCount * 2; i++) {
                input[i] = new double[]{1.0};
            }
            // Remainder (if any) is randomized
            if (numSamples % 2 != 0) {
                input[input.length - 1] = new double[]{rng.nextInt(2)};
            }
        } else {
            for (int i = 0; i < numSamples; i++) {
                input[i] = new double[]{rng.nextInt(2)};
            }
        }
    }

    /**
     * Flattens a 2D matrix
     * @param input A matrix of values
     * @return An array of values
     */
    public static double[] flattenArray(double[][] input) {
        int finalLength = 0;
        for (double[] d : input) finalLength += d.length;

        double[] output = new double[finalLength];

        int destPos = 0;
        for (double[] d : input) {
            System.arraycopy(d, 0, output, destPos, d.length);
            destPos += d.length;
        }

        return output;
    }

    /**
     * Naive Fisher-Yates shuffle for 1D array
     * @param data An array of doubles
     */
    public static void shuffle(double[] data) {
        Random rng = Nd4j.getRandom();
        for (int i = data.length - 1; i > 0; i--) {
            // Random.nextInt is from 0 (inclusive) to n (exclusive)
            int j = rng.nextInt(i + 1);
            double mark = data[j];
            data[j] = data[i];
            data[i] = mark;
        }
    }

    /**
     * Naive Fisher-Yates shuffle for 2D array
     * @param data A 2D array of doubles
     */
    public static void shuffle(double[][] data) {
        Random rng = Nd4j.getRandom();
        for (int i = data.length - 1; i > 0; i--) {
            // Random.nextInt is from 0 (inclusive) to n (exclusive)
            int j = rng.nextInt(i + 1);
            double[] mark = data[j];
            data[j] = data[i];
            data[i] = mark;
            // Shuffle inner array
            shuffle(data[i]);
        }
        // Shuffle last array
        shuffle(data[0]);
    }

    /**
     * Creates a temporal parity data set with an equal number of parity bits (target labels)
     * @param windowSize How many time steps backwards should be calculated
     * @param delay How many time steps backwards before calculation should begin
     * @param numSamples How many samples should the data set consist of.
     * @param oneHot Whether or not to use one-hot representation (e.g [0, 1], [1, 0], [0, 0, 0, 1, 0]) for class label
     * @return A Data object with input and target fields containing the data arrays.
     */
    public static Data createTemporalParityDataWithEvenParityBitDistribution(int windowSize, int delay,
                                                                          int numSamples, boolean oneHot) {
        Data data = new Data(numSamples);

        boolean equal = false;
        while (!equal) {
            createRandomTemporalParityData(data.input(), data.output(), windowSize, delay, numSamples, oneHot);
            double sum = 0;
            for (int i = 0; i < numSamples; i++) {
                sum += data.output()[i][0];
            }
            equal = numSamples / 2 - sum < 1.0;
        }

        return data;
    }

    public static List<Data> createTemporalParityDataWithEvenParityBitDistribution(int numDataSets, int windowSize,
                                                                                   int delay, int numSamples,
                                                                                   boolean oneHot) {
        if (numSamples % numDataSets != 0) {
            throw new IllegalArgumentException("Number of samples must be a factor of numDataSets");
        }
        ArrayList<Data> sets = new ArrayList<>();

        for (int i = 0; i < numDataSets; i++) {
            sets.add(createTemporalParityDataWithEvenParityBitDistribution(
                windowSize, delay, numSamples / numDataSets, oneHot
            ));
        }

        return sets;
    }

    public static Data createTemporalDensityDataWithEvenParityBitDistribution(int windowSize, int delay, int numSamples,
                                                                              boolean oneHot) {
        Data data = new Data(numSamples);

        boolean equal = false;
        while (!equal) {
            createRandomTemporalDensityData(data.input(), data.output(), windowSize, delay, numSamples, oneHot);
            double sum = 0;
            for (int i = 0; i < numSamples; i++) {
                sum += data.output()[i][0];
            }
            equal = numSamples / 2 - sum < 1.0;
        }

        return data;
    }

    public static List<Data> createTemporalDensityDataWithEvenParityBitDistribution(int numDataSets, int windowSize,
                                                                                    int delay, int numSamples,
                                                                                    boolean oneHot) {
        if (numSamples % numDataSets != 0) {
            throw new IllegalArgumentException("Number of samples must be a factor of numDataSets");
        }
        ArrayList<Data> sets = new ArrayList<>();

        for (int i = 0; i < numDataSets; i++) {
            sets.add(createTemporalDensityDataWithEvenParityBitDistribution(
                windowSize, delay, numSamples / numDataSets, oneHot
            ));
        }

        return sets;
    }

    /**
     * Merges the given arrays together into a new array.
     * @param arrays A variable number of double arrays.
     * @return A new, merged array of the total combined length of the argument arrays.
     */
    public static double[] merge(double[]... arrays) {
        int totalLength = 0;
        for (double[] a : arrays) totalLength += a.length;
        double[] output = new double[totalLength];

        int pointer = 0;
        for (double[] arr : arrays) {
            System.arraycopy(arr, 0, output, pointer, arr.length);
            pointer += arr.length;
        }

        return output;
    }

    /**
     * Converts underlying data of an INDArray instance to a two dimensional array.
     * @param data A INDArray instance with data.
     * @return A matrix of doubles.
     */
    public static double[] convertTo1DArray(INDArray data) {
        if (data.shape()[1] != 1) throw new IllegalArgumentException("INDArray shape must be 1 dimensional.");
        int rows = data.shape()[0];

        double[] arr = new double[rows];

        for (int x = 0; x < rows; x++) {
            arr[x] = data.getDouble(x);
        }

        return arr;
    }

    /**
     * Converts underlying data of an INDArray instance to a two dimensional array.
     * @param data A INDArray instance with data.
     * @return A matrix of doubles.
     */
    public static double[][] convertTo2DArray(INDArray data) {
        int rows = data.shape()[0];
        int columns = data.shape()[1];

        double[][] arr = new double[rows][columns];

        for (int y = 0; y < rows; y++) {
            arr[y] = new double[columns];
            for (int x = 0; x < columns; x++) {
                arr[y][x] = data.getDouble(y, x);
            }
        }

        return arr;
    }

    /**
     * Generates a series of real valued numbers that follow a sinusoidal pattern.
     * See http://www.gaussianwaves.com/2015/11/interpreting-fft-results-obtaining-magnitude-and-phase-information/
     * for more information about these arguments.
     *
     * @param frequency The frequency of the signal.
     * @param amplitude The maximum amplitude of the signal.
     * @param phase Offset the signal phase.
     * @param overSamplingFactor A factor that adjusts how many complete signal periods should be present.
     * @param seconds The number of seconds to capture the signal over
     * @return An array of real valued numbers in a sinusoidal pattern.
     */
    public static double[] generateSineSeries(double frequency, double amplitude, double phase,
                                              double overSamplingFactor, double seconds) {
        return generateSineOrCosine(true, frequency, amplitude, phase, overSamplingFactor, seconds);
    }

    /**
     * Generates a series of real valued numbers that follow a cosine pattern.
     * See http://www.gaussianwaves.com/2015/11/interpreting-fft-results-obtaining-magnitude-and-phase-information/
     * for more information about these arguments.
     *
     * @param frequency The frequency of the signal.
     * @param amplitude The maximum amplitude of the signal.
     * @param phase Offset the signal phase.
     * @param overSamplingFactor A factor that adjusts how many complete signal periods should be present.
     * @param seconds The number of seconds to capture the signal over
     * @return An array of real valued numbers in a cosine pattern.
     */
    public static double[] generateCosineSeries(double frequency, double amplitude, double phase,
                                                double overSamplingFactor, double seconds) {
        return generateSineOrCosine(false, frequency, amplitude, phase, overSamplingFactor, seconds);
    }

    /**
     * Generates a series of real valued numbers that follow a cosine pattern.
     * See http://www.gaussianwaves.com/2015/11/interpreting-fft-results-obtaining-magnitude-and-phase-information/
     * for more information about these arguments.
     *
     * @param sine True if a sine-pattern should be created, false if cosine.
     * @param frequency The frequency of the signal.
     * @param amplitude The maximum amplitude of the signal.
     * @param phase Offset the signal phase
     * @param overSamplingFactor A factor that adjusts how many complete signal periods should be present.
     * @param seconds The number of seconds to capture the signal over
     * @return An array of real valued numbers in a cosine pattern.
     */
    private static double[] generateSineOrCosine(boolean sine, double frequency, double amplitude,
                                                 double phase, double overSamplingFactor, double seconds) {
        double samplingRate = frequency * overSamplingFactor;
        int numSamples = (int) (samplingRate * seconds);

        double[] result = new double[numSamples];

        for (int i = 0; i < numSamples; i++) {
            double t = i / samplingRate;
            if (sine) result[i] = amplitude * Math.sin(2.0 * Math.PI * frequency * t + phase);
            else result[i] = amplitude * Math.cos(2.0 * Math.PI * frequency * t + phase);
        }

        return result;
    }

    /**
     * Adds uniform noise to a data array, given a maximum delta the data points can be changed.
     * @param data An array of real numbers.
     * @param maxDelta A maximum delta each data point can be changed.
     */
    public static void addUniformNoise(double[] data, double maxDelta) {
        for (int i = 0; i < data.length; i++) {
            data[i] += Nd4j.getRandom().nextDouble() * maxDelta * 2 - maxDelta;
        }
    }

    /**
     * Fill an array with the specified value.
     * @param d The value to fill this array with.
     * @param n The length of the resulting array.
     * @return An array of length N filled with the real number D.
     */
    public static double[] fillWith(double d, int n) {
        double[] arr = new double[n];

        for (int i = 0; i < n; i++) arr[i] = d;

        return arr;
    }

    /**
     * Adds statistical noise having a probability density function (PDF) equal to the normal distribution
     * (Gaussian distribution).
     * @param data An array of real numbers.
     * @param snr Signal to noise ratio, which determines how strong the noise is.
     */
    public static void addGaussianNoise(double[] data, double snr) {
        Random r = Nd4j.getRandom();
        DescriptiveStatistics ds = new DescriptiveStatistics();

        for (double d : data) {
            ds.addValue(d);
        }

        double stdDev = ds.getStandardDeviation();
        double mean = ds.getMean();

        for (int i = 0; i < data.length; i++) {
            // Add Gaussian noise adhering to a flat SNR
            data[i] += (r.nextGaussian() * stdDev + mean) * Math.abs(data[i] / snr);
        }
    }

    /**
     * Normalizes the values in an array to within the range of 0 to 1.
     * @param data An array of real numbers.
     */
    public static void normalizeToOne(double[] data) {
        double min = Double.MAX_VALUE;
        double max = Double.MIN_VALUE;

        for (double d : data) {
            if (d < min) min = d;
            else if (d > max) max = d;
        }

        for (int i = 0; i < data.length; i++) {
            data[i] = (data[i] - min) / (max - min);
        }
    }

    /**
     * Normalizes the data in the given array
     * @param data The array to be normalized.
     * @param maxSeen The globally maximum number seen.
     * @param minSeen The globally minimum number seen.
     */
    public static void normalizeToOne(double[] data, double maxSeen, double minSeen) {
        for (int i = 0; i < data.length; i++) {
            data[i] = (data[i] - minSeen) / (maxSeen - minSeen);
        }
    }

    /**
     * Normalizes the data in the given array to within the provided upper and lower bounds.
     * @param data An array of real numbers.
     * @param upper An upper bound to which to linearly scale the data.
     * @param lower A lower bound to which to linearly scale the data.
     */
    public static void normalize(double[] data, double upper, double lower) {
        double min = Double.MAX_VALUE;
        double max = Double.MIN_VALUE;

        for (double d : data) {
            if (d < min) min = d;
            else if (d > max) max = d;
        }

        for (int i = 0; i < data.length; i++) {
            data[i] = (upper - lower) / (max - min) * (data[i] - max) + upper;
        }
    }

    /**
     * Returns the column at the specified index of a 2D matrix.
     * @param c The column index
     * @param arr THe matrix
     * @return An array containing the values of the column at index 'c'
     */
    public static double[] getColumn(int c, double[][] arr) {
        return getColumn(c, arr.length, arr);
    }

    /**
     * Returns the n first values of the column at the specified index of a 2D matrix.
     * @param c The column index
     * @param n The number of rows to traverse
     * @param arr The matrix
     * @return An array containing the N first values of the column at index 'c'
     */
    public static double[] getColumn(int c, int n, double[][] arr) {
        double[] col = new double[n];

        for (int i = 0; i < n; i++) {
            col[i] = arr[i][c];
        }

        return col;
    }

    /**
     * Loads the NARMA10 data set, containing 10 samples of the Narma10 data set, each with 1000 data points.
     * @return A list of Data objects containing the Narma10 data set.
     */
    public static List<Data> generateNARMA10DataSet() {
        return loadNARMASet("/datasets/narma10.json", 16);
    }

    /**
     * Loads the NARMA10 data set, containing 10 samples of the Narma10 data set, each with 1000 data points.
     * @return A list of Data objects containing the Narma10 data set.
     */
    public static List<Data> generateNARMA30DataSet() {
        return loadNARMASet("/datasets/narma30.json", 10);
    }

    private static List<Data> loadNARMASet(String filename, int numSets) {
        Gson g = new Gson();
        InputStreamReader resourceStream = new InputStreamReader(DataFactory.class.getResourceAsStream(filename));
        Narma n = g.fromJson(resourceStream, Narma.class);

        ArrayList<Data> samples = new ArrayList<>();
        for (int i = 0; i < numSets; i++) {
            Data d = new Data(n.inputs[i].length);

            double[][] input = d.input();
            System.arraycopy(n.inputs[i], 0, input, 0, n.inputs[i].length);

            double[][] output = d.output();
            System.arraycopy(n.outputs[i], 0, output, 0, n.outputs[i].length);

            samples.add(d);
        }

        return samples;
    }
}
