package no.ntnu.cerebrum.medulla.data;

import no.ntnu.cerebrum.medulla.config.Config;
import org.apache.commons.math3.complex.Complex;

import javax.annotation.Nullable;
import java.util.HashSet;

/**
 * Created 2017-04-06.
 * <p>
 * cerebrum-medulla is licenced under the MIT licence.
 */
public class Evaluation {
    /**
     * Analyzes the real numbered output of some signal power spectrum, looking at the relationship between signal
     * strengths between adjacent frequency bins. A "state" is registered when there is an increasing or decreasing signal
     * within the min and max coefficient restraints when traversing the frequency bins.
     * @param spectrum The squared real valued output of a Fourier Transform.
     * @return The number of valid harmonic states fulfilling the constraints.
     */
    public static int numDistinctSpectrumHarmonics(double[] spectrum) {
        int states = 1;
        double maxPowerCoefficient = Config.getDouble("no.ntnu.cerebrum.medulla.dataset.evaluation.spectrum.maxPowerCoefficient");
        double minPowerCoefficient = Config.getDouble("no.ntnu.cerebrum.medulla.dataset.evaluation.spectrum.minPowerCoefficient");

        for (int x = 0; x < spectrum.length - 2; x++) {
            // Increasing
            double a = spectrum[x];
            double b = spectrum[x + 1];
            double c = spectrum[x + 2];

            if (b > 0 &&
                a > b &&
                (a / b) < maxPowerCoefficient &&
                (a / b) > minPowerCoefficient) {
                states++;
            } else if (c > 0 &&
                a > c &&
                (a / c) < maxPowerCoefficient &&
                (a / c) > minPowerCoefficient) {
                states++;
            }

            // Decreasing
            double d = spectrum[spectrum.length - x - 1];
            double e = spectrum[spectrum.length - x - 2];
            double f = spectrum[spectrum.length - x - 3];

            if (e > 0 &&
                d > e &&
                (d / e) < maxPowerCoefficient &&
                (d / e) > minPowerCoefficient) {
                states++;
            } else if (f > 0 &&
                d > f &&
                (d / f) < maxPowerCoefficient &&
                (d / f) > minPowerCoefficient) {
                states++;
            }
        }

        return states;
    }

    /**
     * Analyzes the real numbered output of some signal power spectrum, looking at the relationship between signal
     * strengths between adjacent frequency bins. A "state" is registered when there is an decreasing signal
     * within the min and max coefficient restraints when traversing the sine frequency bins (1, 3, 5 ... ).
     *
     * If the power spectrum does not represent a sine wave or any harmonics of a sine wave. It will return 0
     *
     * @param spectrum The squared real valued output of a Fourier Transform.
     * @return The number of valid harmonic sine states fulfilling the constraints.
     */
    public static int numDistinctSineSpectrumHarmonics(double[] spectrum) {
        int states = 1;
        double maxPowerCoefficient = Config.getDouble("no.ntnu.cerebrum.medulla.dataset.evaluation.spectrum.maxPowerCoefficient");
        double minPowerCoefficient = Config.getDouble("no.ntnu.cerebrum.medulla.dataset.evaluation.spectrum.minPowerCoefficient");

        // A signal in the DC bin has a frequency of 0.0
        // it is arguably not periodic.
        if (spectrum[0] > spectrum[1]) {
            return states;
        }

        for (int x = 1; x < spectrum.length - 2; x = x + 2) {
            // Increasing
            double a = spectrum[x];
            double b = spectrum[x + 2];

            if (b > 0 &&
                    a > b &&
                    (a / b) < maxPowerCoefficient &&
                    (a / b) > minPowerCoefficient) {
                states++;
            }
        }

        return states;
    }

    /**
     * Scans an array of real numbers to check whether or not it has converged towards a single number.
     * Number equality is checked backward from the last output, up to the optionally provided
     * number of steps. If number of steps is not provided, the last 20% of the array is checked by default.
     * @param data An array of real numbers.
     * @param numStepsBackward An optional number of steps to be checked for equality. Defaults to 20% of data length.
     * @param roundingThreshold A specified value for which a difference less than this number is considered zero.
     *                          Defaults to 0.0.
     * @return true if the last N numbers are equal, false otherwise.
     */
    public static boolean isNumericallyConverged(double[] data, @Nullable Integer numStepsBackward, @Nullable Double roundingThreshold) {
        if (roundingThreshold == null) roundingThreshold = 0.0;
        if (numStepsBackward == null) numStepsBackward = data.length / 5;
        if (numStepsBackward > data.length) throw new IllegalArgumentException("Number of steps cannot be larger than the data array!");

        for (int i = data.length - 1; i > Math.max(data.length - 1 - numStepsBackward, 0); i--) {
            if (Math.abs(data[i] - data[i-1]) > roundingThreshold) return false;
        }

        return true;
    }

    /**
     * Scans an array of real numbers to check whether or not it has converged towards a single number.
     * Number equality is checked backward from the last output, up to the optionally provided
     * number of steps. If number of steps is not provided, the last 20% of the array is checked by default.
     * @param data An array of real numbers.
     * @param numStepsBackward An optional number of steps to be checked for equality. Defaults to 20% of data length.
     * @return true if the last N numbers are equal, false otherwise.
     */
    public static boolean isNumericallyConverged(double[] data, @Nullable Integer numStepsBackward) {
        return isNumericallyConverged(data, numStepsBackward, null);
    }

    /**
     * Scans an array of real numbers to check whether or not it has converged towards a single number.
     * Number equality is checked backward and counting 20% of the total array length.
     * @param data An array of real numbers.
     * @return true if the last N numbers are equal, false otherwise.
     */
    public static boolean isNumericallyConverged(double[] data) {
        return isNumericallyConverged(data, null, null);
    }

    /**
     * Counts the number of unique values in a 1D array.
     * @param data An array of real numbers.
     * @return The number of distinct values found in the data array.
     */
    public static int numUniqueValues(double[] data) {
        HashSet<Double> seen = new HashSet<>();
        for (double d : data) seen.add(d);

        return seen.size();
    }

    /**
     * Counts the number of unique values in a 2D array.
     * @param data An array of real numbers.
     * @return The number of distinct values found in the data array.
     */
    public static int numUniqueValues(double[][] data) {
        HashSet<Double> seen = new HashSet<>();
        for (double[] row : data) {
            for (double d : row) seen.add(d);
        }

        return seen.size();
    }

    /**
     * Calculates the power spectrum density of a 1D signal in the frequency domain (Fourier Transformed).
     * @param fft An array of Complex numbers.
     * @param roundingThreshold Optional rounding threshold. Values below this threshold will be rounded to zero.
     * @return An array of real numbers representing the signal strength at each frequency bin.
     */
    public static double[] powerSpectrumDensity(Complex[] fft, @Nullable Double roundingThreshold) {
        if (roundingThreshold == null) roundingThreshold = 1.0E-6;
        double[] result = new double[fft.length];

        for (int x = 0; x < fft.length; x++) {
            double real = fft[x].getReal();
            double imag = fft[x].getImaginary();

            if (Math.abs(real) < roundingThreshold) real = 0.0;
            if (Math.abs(imag) < roundingThreshold) imag = 0.0;

            result[x] = Math.sqrt(Math.pow(imag, 2) + Math.pow(real, 2));
        }

        return result;
    }

    /**
     * Calculates the power spectrum density of a 2D signal in the frequency domain (Fourier Transformed).
     * @param fft An array of Complex numbers.
     * @param roundingThreshold Optional rounding threshold. Values below this threshold will be rounded to zero.
     * @return A 2D array of real numbers representing the signal strength at each point.
     */
    public static double[][] powerSpectrumDensity(Complex[][] fft, @Nullable Double roundingThreshold) {
        double[][] result = new double[fft.length][fft[0].length];

        for (int y = 0; y < fft.length; y++) {
            result[y] = powerSpectrumDensity(fft[y], roundingThreshold);
        }

        return result;
    }
}
