package no.ntnu.cerebrum.medulla.data;

import org.apache.commons.math3.complex.Complex;
import org.apache.commons.math3.exception.MathIllegalArgumentException;
import org.apache.commons.math3.transform.DftNormalization;
import org.apache.commons.math3.transform.FastFourierTransformer;
import org.apache.commons.math3.transform.TransformType;
import org.jtransforms.fft.DoubleFFT_2D;
import org.jtransforms.fft.RealFFTUtils_2D;

import javax.annotation.Nullable;
import java.util.Arrays;

/**
 * Created 2017-03-24.
 * <p>
 * cerebrum-medulla is licenced under the MIT licence.
 */
public class Fourier {
    /**
     * Creates an array consisting of the values along a specific column in a 2D array of double values.
     * @param data A 2D array of doubles
     * @param idx The column index from which to gather values.
     * @return A 1D array of double values along the "idx" column.
     */
    private static double[] arrayColumn(double[][] data, int idx) {
        double[] result = new double[data.length];
        for (int y = 0; y < data.length; y++) {
            result[y] = data[y][idx];
        }

        return result;
    }

    /**
     * Performs a 1D DFT (Using FFT) on a 1D input signal in the forward direction.
     * @param inputSignal An array of doubles. The input array will be zero padded up to the nearest power of two
     *                    in length before applying the FFT algorithm.
     * @return An array of Complex numbers.
     */
    public static Complex[] forwardDFT(double[] inputSignal) {
        FastFourierTransformer fft = new FastFourierTransformer(DftNormalization.STANDARD);
        try {
            return fft.transform(inputSignal, TransformType.FORWARD);
        } catch (MathIllegalArgumentException e) {
            // Check the next power of two that can hold all the samples.
            // Array will be zero-padded to fix FFT matrix multiplication
            int arraySize = new Double(Math.pow(2.0, Math.ceil(Math.log(inputSignal.length) / Math.log(2)))).intValue();

            return fft.transform(Arrays.copyOf(inputSignal, arraySize), TransformType.FORWARD);
        }
    }

    /**
     * Performs a 1D DFT (Using FFT) on a 2D input signal in the forward direction.
     * @param inputSignal A 2D array of doubles. The input array will be zero padded up to the nearest power of two
     *                    in length before applying the FFT algorithm.
     * @return An array of Complex numbers.
     */
    public static Complex[][] forwardDFT(double[][] inputSignal) {
        FastFourierTransformer fft = new FastFourierTransformer(DftNormalization.STANDARD);
        Complex[][] result = new Complex[inputSignal.length][inputSignal[0].length];

        // 1st pass (row-wise)
        for (int y = 0; y < inputSignal.length; y++) {
            result[y] = fft.transform(inputSignal[y], TransformType.FORWARD);
        }

        // 2nd pass (column-wise)
        for (int x = 0; x < inputSignal[0].length; x++) {
            Complex[] columnResult = fft.transform(arrayColumn(inputSignal, x), TransformType.FORWARD);
            for (int y = 0; y < columnResult.length; y++) {
                result[y][x] = columnResult[y];
            }
        }

        return result;
    }

    public static double[][] jTransformForwardFFT(double[][] data) {
        DoubleFFT_2D fft_2D = new DoubleFFT_2D(data.length, data[0].length);

        // Prepare the matrix for calculation by doubling the column size (For imaginary numbers)
        double[][] matrix = new double[data.length][data[0].length * 2];
        for (int i = 0; i < data.length; i++) {
            for (int j = 0; j < data[0].length; j++) {
                matrix[i][j*2] = data[i][j];
            }
        }

        fft_2D.realForward(matrix);

        return matrix;
    }

    public static double[][] jTransformPowerSpectrumDensity(double[][] rawData, @Nullable Double roundingThreshold) {
        if (roundingThreshold == null) roundingThreshold = 1.0E-6;

        RealFFTUtils_2D unPacker = new RealFFTUtils_2D(rawData.length, rawData[0].length);
        double[][] preparedData = new double[rawData.length][rawData[0].length / 2];

        for (int y = 0; y < preparedData.length; y++) {
            for (int x = 0; x < preparedData[0].length; x++) {
                // We take the root of the sum of squares of both the real and imaginary part of the transform
                preparedData[y][x] = Math.sqrt(
                    Math.pow(unPacker.unpack(y, x * 2, rawData), 2) +
                    Math.pow(unPacker.unpack(y, x * 2 + 1, rawData), 2)
                );

                // If signal power is below threshold, zero it out
                if (preparedData[y][x] < roundingThreshold) preparedData[y][x] = 0.0;
            }
        }

        return preparedData;
    }
}
