package no.ntnu.cerebrum.medulla.data;

import org.apache.commons.io.FilenameUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Created 2017-04-07.
 * <p>
 * cerebrum-medulla is licenced under the MIT licence.
 */
public class Storage {
    private static final Logger log = LoggerFactory.getLogger(Storage.class);

    public static final String BASE_PATH = "storage/";
    public static final String CANDIDATES = "candidates/";

    /**
     * Reads the binary contents of a file and interprets the bytes as a String.
     * Returns null if the file is not found or corrupted.
     * @param filePath The full path to the file.
     * @return The contents of the file as a String.
     */
    public static String readFromFile(String filePath) {
        return readFromFile(Paths.get(filePath));
    }

    /**
     * Reads the binary contents of a file and interprets the bytes as a String.
     * Returns null if the file is not found or corrupted.
     * @param filePath The full path to the file.
     * @return The contents of the file as a String.
     */
    public static String readFromFile(Path filePath) {
        try {
            return String.join("", Files.readAllLines(filePath));
        } catch (IOException e) {
            log.error(String.format(
                "Failed to read data from file: %s (%s - %s)",
                filePath,
                e.getMessage(),
                e.getCause().toString()
            ));
            e.printStackTrace();

            return null;
        }
    }

    /**
     * Opens a stream of Path objects (file paths) which are opened and read as Strings.
     * @param dir A base directory path from which to recursively read all file contents.
     * @return A list of strings containing the file data.
     */
    public static List<String> readAllFiles(String dir) {
        try (Stream<Path> paths = Files.walk(Paths.get(dir))) {
            return paths
                .filter(Files::isRegularFile)
                .map(Storage::readFromFile)
                .collect(Collectors.toList());
        } catch (IOException e) {
            log.error(String.format(
                "Failed to read directory: %s (%s - %s)",
                dir,
                e.getMessage(),
                e.getCause()
            ));
            return new ArrayList<>();
        }
    }

    /**
     * Opens a stream of Path objects (file paths) which are opened and read as Strings.
     * @param dir A base directory path from which to recursively read all files of the provided fileType.
     * @return A list of strings containing the file data.
     */
    public static List<String> readAllFiles(String dir, String fileType) {
        try (Stream<Path> paths = Files.walk(Paths.get(dir))) {
            return paths
                .filter(Files::isRegularFile)
                .filter(fp -> fp.toString().toLowerCase().endsWith(fileType.toLowerCase()))
                .map(Storage::readFromFile)
                .collect(Collectors.toList());
        } catch (IOException e) {
            log.error(String.format(
                "Failed to read directory: %s (%s - %s)",
                dir,
                e.getMessage(),
                e.getCause()
            ));
            return new ArrayList<>();
        }
    }

    /**
     * Searches for a file with the specified filename, recursively through subdirectories. Returns null if not found.
     * @param fileName Filename (including extension) to search for.
     * @return A Path object if the file is located, null otherwise.
     */
    public static Path search(String fileName) {
        try (Stream<Path> paths = Files.walk(Paths.get(BASE_PATH))) {
            Optional<Path> p = paths
                .filter(Files::isRegularFile)
                .filter(fp -> fp.toString().toLowerCase().contains(fileName.toLowerCase())).findFirst();

            if (p.isPresent()) return p.get();
        } catch (IOException e) {
            log.error(String.format(
                "Failed to read directory: %s (%s - %s)",
                fileName,
                e.getMessage(),
                e.getCause()
            ));
        }

        return null;
    }

    /**
     * Writes CSV data to a file in the appropriate storage folder. Parent directories will be created if they do
     * not exist.
     * @param fileName Filename without extension.
     * @param data The contents of the file.
     * @return true if the operation is successful, false otherwise.
     */
    public static boolean writeCSV(String fileName, String data) {
        String fullPath = BASE_PATH + CANDIDATES + fileName;
        if (!fullPath.contains(".csv")) fullPath += ".csv";

        createDirectoriesIfNotExists(fullPath);

        return writeToFile(fullPath, data);
    }

    /**
     * Writes JSON data to a file in the appropriate storage folder. Parent directories will be created if they do
     * not exist.
     * @param fileName Filename without extension.
     * @param data The contents of the file.
     * @return true if the operation is successful, false otherwise.
     */
    public static boolean writeJSON(String fileName, String data) {
        String fullPath = BASE_PATH + CANDIDATES + fileName;
        if (!fullPath.contains(".json")) fullPath += ".json";

        createDirectoriesIfNotExists(fullPath);

        return writeToFile(fullPath, data);
    }

    /**
     * Writes a string to a file. NB: This method assumes directories exist.
     * @param filePath The full path to the file being written to.
     * @param data The contents of the file.
     * @return true if the operation is successful, false otherwise.
     */
    private static boolean writeToFile(String filePath, String data) {
        Path fp = Paths.get(filePath);
        try (BufferedWriter writer = Files.newBufferedWriter(fp)) {
            writer.write(data);
            writer.flush();
            writer.close();

            return true;
        } catch (IOException e) {
            log.error(String.format(
                "Failed to write data to file: %s (%s - %s)",
                filePath,
                e.getMessage(),
                e.getCause().toString()
            ));
            e.printStackTrace();

            return false;
        }
    }

    /**
     * Ensures that all parent directories of the specified path are created on disk.
     * @param fullPath Full path to a directory or filename
     */
    private static void createDirectoriesIfNotExists(String fullPath) {
        Path path = Paths.get(FilenameUtils.getPath(fullPath));
        // If the working directory does not exist, create it
        if (!Files.exists(path)) {
            try {
                Files.createDirectories(path);
            } catch (IOException e) {
                log.error(String.format(
                    "Failed to create parent directories for: %s (%s)",
                    fullPath,
                    e.getMessage()
                ));
                e.printStackTrace();
            }
        }
    }
}
