package no.ntnu.cerebrum.medulla.reservoir;

import no.ntnu.cerebrum.medulla.config.Config;
import no.ntnu.cerebrum.medulla.data.DataFactory;
import no.ntnu.cerebrum.medulla.reservoir.interfaces.Reservoir;
import no.ntnu.cerebrum.medulla.reservoir.interfaces.ReservoirTrainingListener;
import no.ntnu.cerebrum.medulla.reservoir.interfaces.SMCGPProcessingListener;
import no.ntnu.cerebrum.medulla.reservoir.readout.MultiLayerRegression;
import no.ntnu.cerebrum.medulla.reservoir.readout.ReadoutLayer;
import no.ntnu.cerebrum.medulla.reservoir.readout.Score;
import no.ntnu.cerebrum.medulla.reservoir.readout.SingleLayerLinearRegression;
import org.apache.commons.math3.stat.descriptive.SummaryStatistics;
import org.nd4j.linalg.dataset.api.DataSet;

import java.util.ArrayList;
import java.util.concurrent.TimeoutException;

/**
 * Created 2017-02-16.
 * <p>
 * cerebrum-medulla is licenced under the MIT licence.
 */
public class SMCGPReservoir {
    // Statistics
    protected static SummaryStatistics ss = new SummaryStatistics();
    private ArrayList<ReservoirTrainingListener> scoreListeners;
    private ArrayList<SMCGPProcessingListener> smcgpProcessingListeners;
    
    public Reservoir reservoir;
    private ReadoutLayer readout;

    /**
     * Constructs a new SMCGP Reservoir, with the provided argument as the active SMCGP instance.
     * @param r A SMCGP Phenotype instance.
     */
    public SMCGPReservoir(Reservoir r) {
        if (r == null) throw new IllegalArgumentException("The reservoir cannot be 'null'!");

        reservoir = r;

        scoreListeners = new ArrayList<>();
        smcgpProcessingListeners = new ArrayList<>();
    }

    /**
     * Attach a readout layer to this SMCGPReservoir
     * @param r A ReadoutLayer instance
     * @return This instance of SMCGPReservoir
     */
    public SMCGPReservoir readoutLayer(ReadoutLayer r) {
        readout = r;
        scoreListeners = new ArrayList<>();

        r.init(
            Config.getDouble("no.ntnu.cerebrum.medulla.reservoir.readout.learningRate"),
            Config.getDouble("no.ntnu.cerebrum.medulla.reservoir.readout.momentum")
        );

        return this;
    }

    /**
     * Pre-process an array of inputs through the SMCGP reservoir, returning the raw output of the SMCGP output node(s).
     *
     * @param input An array of inputs/features from a data set.
     * @return An array containing the raw SMCGP output.
     * @throws TimeoutException If processing the output results of the SMCGP takes longer than specified in
     *                          "smcgp.timeout" in config.
     */
    public double[] preProcess(double[] input) throws TimeoutException {
        long startTime = System.currentTimeMillis();

        int numCachedTimeSteps = Config.getInt("no.ntnu.cerebrum.medulla.reservoir.numCachedTimeSteps");
        double[] inputBuffer = new double[numCachedTimeSteps + 1];
        double[] outputBuffer = new double[input.length];

        for (int i = 0; i < input.length; i++) {
            inputBuffer[0] = input[i];

            for (int j = 1; j <= numCachedTimeSteps; j++) {
                if (i - j < 0) break;
                inputBuffer[j] = outputBuffer[i - j];
            }

            double[] result = reservoir.fireSignal(inputBuffer);

            outputBuffer[i] = result[0];
        }

        // Report statistics to any registered listeners
        for (SMCGPProcessingListener spl : smcgpProcessingListeners) {
            spl.preProcessingComplete(input.length, System.currentTimeMillis() - startTime);
        }

        return outputBuffer;
    }

    /**
     * Pre-process an array of inputs through the SMCGP reservoir, returning the raw output of the SMCGP output node(s).
     *
     * @param input An 2D matrix of inputs/features from a data set.
     * @return A matrix containing the raw SMCGP output.
     * @throws TimeoutException If processing the output results of the SMCGP takes longer than specified in
     *                          "smcgp.timeout" in config.
     */
    public double[][] preProcess(double[][] input) throws TimeoutException {
        long startTime = System.currentTimeMillis();

        boolean useInputExpansion = Config.getBoolean("no.ntnu.cerebrum.medulla.reservoir.useInputExpansion");
        int numCachedTimeSteps = Config.getInt("no.ntnu.cerebrum.medulla.reservoir.numCachedTimeSteps");
        int outputs = Config.getInt("no.ntnu.cerebrum.medulla.smcgp.outputs");
        int inputExpansionFactor = Config.getInt("no.ntnu.cerebrum.medulla.reservoir.inputExpansionFactor");

        double[][] inputBuffer = new double[numCachedTimeSteps + 1][];
        double[][] outputBuffer = new double[input.length][outputs];

        for (int i = 0; i < input.length; i++) {
            // Use input dimensionality expansion if we only have a single input value
            if (input[i].length == 1 && useInputExpansion) inputBuffer[0] = DataFactory.fillWith(input[i][0], inputExpansionFactor);
            else inputBuffer[0] = input[i];

            for (int j = 1; j <= numCachedTimeSteps; j++) {
                if (i - j < 0) inputBuffer[j] = new double[outputs];
                else inputBuffer[j] = outputBuffer[i - j];
            }

            outputBuffer[i] = reservoir.fireSignal(DataFactory.flattenArray(inputBuffer));
            for (int k = 0; k < outputBuffer[i].length; k++) {
                if (!Double.isFinite(outputBuffer[i][k])) {
                    outputBuffer[i][k] = 0.0;
                }
            }
        }

        // Report statistics to any registered listeners
        for (SMCGPProcessingListener spl : smcgpProcessingListeners) {
            spl.preProcessingComplete(input.length, System.currentTimeMillis() - startTime);
        }

        return outputBuffer;
    }

    /**
     * Train the reservoir readout for a given amount of epochs, without performing intermediate evaluations.
     * This method does not use early fitness cutoffs, reporting statistics to accuracy score listeners,
     * stop conditions or intermediate accuracy testing during training.
     *
     * @param epochs The total number of epochs to be trained.
     * @param trainingSet The training set on which to train the readout layer.
     */
    public void train(int epochs, DataSet trainingSet) {
        if (readout == null) throw new IllegalStateException("No readout layer specified! Use .readoutLayer() first.");

        long startTime = System.currentTimeMillis();
        for (int x = 0; x < epochs; x++) {
            readout.train(trainingSet);
        }
        ss.addValue(System.currentTimeMillis() - startTime);
    }

    /**
     * Train the reservoir readout for a given number of epochs. Performs evaluation of training and test sets every N
     * epochs, based on the provided parameter. Reports statistics to registered ReservoirTrainingListeners.
     *
     * This method will return early if "reservoir.readout.useFitnessTrainingCutoff" is set to true in config
     * and fitness has not gone above the specified "quarterWayFitnessCutoff" or "halfWayFitnessCutoff" in config.
     *
     * Method will also return early if "reservoir.readout.stopCondition" is met.
     *
     * @param epochs The total number of epochs to be trained.
     * @param trainingSet The training set to train the readout layer on.
     * @param testSet The test set on which to measure the networks true accuracy.
     * @param evalEveryEpoch Perform evaluation of the network every this number of epochs.
     */
    public void train(int epochs, DataSet trainingSet, DataSet testSet, int evalEveryEpoch) {
        if (readout == null) throw new IllegalStateException("No readout layer specified! Use .readoutLayer() first.");

        long startTime = System.currentTimeMillis();

        double quarterWayCutoff = Config.getDouble("no.ntnu.cerebrum.medulla.reservoir.readout.quarterWayFitnessCutoff");
        double halfWayCutoff = Config.getDouble("no.ntnu.cerebrum.medulla.reservoir.readout.halfWayFitnessCutoff");

        for (int x = 0; x < epochs; x++) {
            readout.train(trainingSet);
            if (x % evalEveryEpoch == 0) {
                Score testSetScore = readout.test(testSet);
                Score trainingSetScore = readout.test(trainingSet);
                double testSetFitness = testSetScore.accuracy;
                double trainingSetFitness = trainingSetScore.accuracy;

                // Report statistics to fitness listeners
                for (ReservoirTrainingListener rtl : scoreListeners) {
                    if (readout.getClass().equals(SingleLayerLinearRegression.class) ||
                        readout.getClass().equals(MultiLayerRegression.class)) {
                        rtl.epochComplete(
                            x,
                            trainingSetScore.MSE,
                            trainingSetScore.RMSE,
                            trainingSetScore.NRMSE,
                            trainingSetScore.RNMSE,
                            trainingSetScore.MAE,
                            testSetScore.MSE,
                            testSetScore.RMSE,
                            testSetScore.NRMSE,
                            testSetScore.RNMSE,
                            testSetScore.MAE
                        );
                    }

                    rtl.epochComplete(x, testSetFitness, trainingSetFitness);
                }

                // If configured to throw give up if lousy fitness, check if we are above quarter way and half
                // way thresholds. Checks are only enforced within one evalEveryEpoch of quarter and half to prevent
                // breaking if a sudden drop occurs later on during training.
                if (Config.getBoolean("no.ntnu.cerebrum.medulla.reservoir.readout.useFitnessTrainingCutoff")) {
                    if (Math.abs(epochs / 4 - x) <= evalEveryEpoch &&
                        (testSetFitness < quarterWayCutoff && trainingSetFitness < quarterWayCutoff)) break;
                    if (Math.abs(epochs / 2 - x) <= evalEveryEpoch &&
                        (testSetFitness < halfWayCutoff && trainingSetFitness < halfWayCutoff)) break;
                }

                // Check if we have surpassed our stop condition
                if (Config.getDouble("no.ntnu.cerebrum.medulla.reservoir.readout.stopCondition") - testSetFitness <= 10E-8) break;
            }
        }
        ss.addValue(System.currentTimeMillis() - startTime);
    }

    /**
     * Performs an evaluation of the network accuracy, and an model accuracy is returned.
     * @param testSet The test data set on which to measure the network model accuracy.
     * @return The total accuracy of the network as a real number.
     */
    public Score test(DataSet testSet) {
        if (readout == null) throw new IllegalStateException("No readout layer specified! Use .readoutLayer() first.");

        return readout.test(testSet);
    }

    /**
     * Adds a listener object that will receive updates on each epoch with fitness values.
     * @param rtl An object implementing ReservoirTrainingListener interface.
     */
    public void addReservoirTrainingListener(ReservoirTrainingListener rtl) {
        if (!scoreListeners.contains(rtl)) scoreListeners.add(rtl);
    }

    /**
     * Removes a listener object from the ReservoirTraining registry.
     * @param rtl An object implementing ReservoirTrainingListener interface, currently registered to receive calls.
     */
    public void removeReservoirTrainingListener(ReservoirTrainingListener rtl) {
        if (scoreListeners.contains(rtl)) scoreListeners.remove(rtl);
    }

    /**
     * Adds a listener object that will receive updates on each SMCGP data pre-processing.
     * @param spl An object implementing the SMCGPProcessingListener interface.
     */
    public void addSMCGPPreProcessingListener(SMCGPProcessingListener spl) {
        if (!smcgpProcessingListeners.contains(spl)) smcgpProcessingListeners.add(spl);
    }

    /**
     * Removes a listener object from the SMCGPProcessing registry.
     * @param spl An object implementing the SMCGPProcessingListener interface.
     */
    public void removeSMCGPPreProcessingListener(SMCGPProcessingListener spl) {
        if (smcgpProcessingListeners.contains(spl)) smcgpProcessingListeners.remove(spl);
    }

    /**
     * Resets the global statistics counters
     */
    public static void resetStats() {
        ss = new SummaryStatistics();
    }

    /**
     * String representation of this SMCGPReservoir
     * @return SMCGPReservoir runtime statistics.
     */
    public String toString() {
        return stats();
    }

    /**
     * Get a summary of runtime statistics from this SMCGPReservoir.
     * @return A string containing min/mean/max/total training times.
     */
    public static String stats() {
        return String.format(
            "Reservoir{%.0f/%.0f/%.0f-%.0f ms}",
            ss.getMin(),
            ss.getMean(),
            ss.getMax(),
            ss.getSum()
        );
    }
}
