package no.ntnu.cerebrum.medulla.reservoir.readout;

import org.deeplearning4j.eval.RegressionEvaluation;
import org.deeplearning4j.nn.api.OptimizationAlgorithm;
import org.deeplearning4j.nn.conf.MultiLayerConfiguration;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.Updater;
import org.deeplearning4j.nn.conf.layers.OutputLayer;
import org.deeplearning4j.nn.multilayer.MultiLayerNetwork;
import org.deeplearning4j.nn.weights.WeightInit;
import org.nd4j.linalg.activations.Activation;
import org.nd4j.linalg.api.ndarray.INDArray;

import org.nd4j.linalg.dataset.api.DataSet;
import org.nd4j.linalg.dataset.api.iterator.DataSetIterator;
import org.nd4j.linalg.lossfunctions.LossFunctions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Created 2017-02-16.
 * <p>
 * cerebrum-medulla is licenced under the MIT licence.
 */
public class SingleLayerLinearRegression extends ReadoutLayer {
    protected Logger log = LoggerFactory.getLogger(SingleLayerLinearRegression.class);

    public SingleLayerLinearRegression(int nIn, int nOut) {
        super(nIn, nOut);
    }

    @Override
    public void init(double learningRate, double momentum) {
        log.debug("Initializing single layer regression readout layer...");
        MultiLayerConfiguration conf = new NeuralNetConfiguration.Builder()
            // Use stochastic gradient descent as an optimization algorithm
            .optimizationAlgo(OptimizationAlgorithm.STOCHASTIC_GRADIENT_DESCENT)
            // Apply updates for each iteration
            .iterations(1)
            .weightInit(WeightInit.SIGMOID_UNIFORM)
            // Specify learning rate and the rate of change of the learning rate.
            .learningRate(learningRate)
            .updater(Updater.SGD)
            .biasInit(0.3)
            .list()
            // Create output layer
            .layer(0, new OutputLayer.Builder(LossFunctions.LossFunction.MSE)
                .nIn(nIn)
                .nOut(nOut)
                .activation(Activation.IDENTITY)
                .build())
            // No pre train
            .pretrain(false).backprop(true)
            .build();

        net = new MultiLayerNetwork(conf);
        net.init();

        log.debug("Readout layer initialized: " + this);
    }

    @Override
    public double test(DataSetIterator testSet) {
        int correct = 0;

        while (testSet.hasNext()) {
            DataSet ds = testSet.next();

            INDArray labels = ds.getLabels();
            INDArray output = net.output(ds.getFeatures());

            double predicted = output.getDouble(0);
            double target = labels.getDouble(0);
            if (Double.isNaN(predicted)) {
                numericError++;
                return 0.1234567890E-6;
            }

            if (target < 0.5 && predicted < 0.5) correct++;
            else if (target >= 0.5 && predicted >= 0.5) correct++;
        }

        return (double)correct / testSet.totalExamples();
    }

    @Override
    public Score test(DataSet testSet) {
        RegressionEvaluation e = new RegressionEvaluation(1, 64);

        INDArray predictions = net.output(testSet.getFeatures());
        INDArray labels = testSet.getLabels();

        double labelStdDev = labels.stdNumber().doubleValue();
        double labelRange = labels.maxNumber().doubleValue() - labels.minNumber().doubleValue();

        e.eval(labels, predictions);

        Score s = new Score();
        s.MSE = e.meanSquaredError(0);
        s.RMSE = e.rootMeanSquaredError(0);
        s.NRMSE = e.rootMeanSquaredError(0) / labelRange;
        s.RNMSE = Math.sqrt(e.meanSquaredError(0) / labelStdDev);
        s.MAE = e.meanAbsoluteError(0);
        s.accuracy = 1.0 - s.RMSE;

        if (Double.isInfinite(e.meanSquaredError(0))) {
            s.accuracy = 0.0;
        }

        return s;
    }
}
