package no.ntnu.cerebrum.medulla.smcgp;


import no.ntnu.cerebrum.medulla.config.Config;
import no.ntnu.cerebrum.medulla.data.Data;
import no.ntnu.cerebrum.medulla.data.DataFactory;
import no.ntnu.cerebrum.medulla.data.Evaluation;
import no.ntnu.cerebrum.medulla.data.Fourier;
import no.ntnu.cerebrum.medulla.reservoir.SMCGPReservoir;
import no.ntnu.cerebrum.medulla.reservoir.interfaces.Reservoir;
import no.ntnu.cerebrum.medulla.reservoir.interfaces.SMCGPProcessingListener;
import no.ntnu.cerebrum.medulla.smcgp.evolution.EvolutionListener;
import static no.ntnu.cerebrum.medulla.smcgp.evolution.EvolutionListener.KillReason.*;
import no.ntnu.cerebrum.medulla.smcgp.evolution.Evolve;
import no.ntnu.cerebrum.medulla.smcgp.graph.action.Function;
import no.ntnu.cerebrum.medulla.smcgp.graph.action.FunctionLookup;
import no.ntnu.cerebrum.medulla.smcgp.graph.action.SMFunction;

import org.apache.commons.math3.complex.Complex;
import org.apache.log4j.Logger;
import org.nd4j.linalg.api.rng.*;
import org.nd4j.linalg.api.rng.Random;
import org.nd4j.linalg.factory.Nd4j;

import java.util.*;
import java.util.concurrent.TimeoutException;
import java.util.stream.Collectors;

public class Phenotype2 implements Reservoir {
    private Logger log = Logger.getLogger(Phenotype2.class);

    private Genotype2 genotype;
    private double[][] nodes;
    private ArrayList<Integer> outputNodes;
    private int generation;
    private double fitness;
    private double trainingSetFitness = 0.0;
    private int lastActiveNodeCount = 0;
    private int lastGraphSize = 0;
    private int lastGraphState = 0;

    private double[] input;

    // Input / Terminal
    private int inputPointer;

    // Self-modification
    private ArrayList<Integer> todoList;
    private boolean terminate;

    // Active Node
    private Set<Integer> activeNodes;

    // Timing
    private Long evaluationStart;

    // Node caching
    private Map<Integer, Double> nodeCache;

    // Internal statistics
    private ArrayList<SMCGPProcessingListener> smcgpProcessingListeners;
    private ArrayList<EvolutionListener> evolutionListeners;

    public Phenotype2(Genotype2 genotype) {
        this.genotype = genotype;
        this.nodes = genotype.copyGenes();
        this.generation = 1;
        this.inputPointer = 0;
        this.todoList = new ArrayList<>();
        this.activeNodes = new HashSet<>();
        this.smcgpProcessingListeners = new ArrayList<>();
        this.evolutionListeners = new ArrayList<>();
        this.terminate = false;
        assignOutputNodes();
    }

    public Phenotype2(Genotype2 genotype, int generation) {
        this.genotype = genotype;
        this.nodes = genotype.copyGenes();
        this.generation = generation;
        this.inputPointer = 0;
        this.todoList = new ArrayList<>();
        this.activeNodes = new HashSet<>();
        this.smcgpProcessingListeners = new ArrayList<>();
        this.evolutionListeners = new ArrayList<>();
        this.terminate = false;
        assignOutputNodes();
    }

    public double[] getOutput(double[] rawInput) throws TimeoutException {
        // Start timing
        this.evaluationStart = System.currentTimeMillis();

        // Set input
        this.input = rawInput;
        this.inputPointer = 0;

        // Clear active nodes
        activeNodes.clear();

        // Gather output
        double[] output = new double[outputNodes.size()];
        for (int outputIndex = 0; outputIndex < outputNodes.size(); outputIndex++) {
            if (Config.getBoolean("no.ntnu.cerebrum.medulla.smcgp.cache")) {
                nodeCache = new HashMap<>();
            }
            output[outputIndex] = readNode(outputNodes.get(outputIndex));
        }

        // After collecting output, do self-modify
        doTodoList();

        // The graph is now altered, re-assign outputs
        assignOutputNodes();

        // Reset timer
        this.evaluationStart = null;

        lastActiveNodeCount = activeNodes.size();
        lastGraphSize = nodes.length;

        return output;
    }

    public int getGeneration() {
        return generation;
    }

    public double getFitness() {
        return fitness;
    }

    public double getTrainingSetFitness() {
        return trainingSetFitness;
    }

    public void setFitness(double fitness) {
        this.fitness = fitness;
    }

    public void setTrainingSetFitness(double fitness) {
        this.trainingSetFitness = fitness;
    }

    public Genotype2 getGenotype() {
        return new Genotype2(this.genotype.copyGenes());
    }

    public double[][] getNodes() {
        return nodes;
    }

    public ArrayList<Integer> getOutputNodes() {
        return outputNodes;
    }

    public Set getActiveNodes() {
        return this.activeNodes;
    }

    public int getLastActiveNodeCount() {
        return lastActiveNodeCount;
    }

    public int getLastGraphSize() {
        return lastGraphSize;
    }

    public void addSMCGPProcessingListener(SMCGPProcessingListener spl) {
        if (!smcgpProcessingListeners.contains(spl)) smcgpProcessingListeners.add(spl);
    }

    public void removeSMCGPProcessingListener(SMCGPProcessingListener spl) {
        smcgpProcessingListeners.remove(spl);
    }

    /**
     * Register an object for evolve loop callbacks
     * @param listener An object implementing the EvolutionListener interface
     */
    public void addEvolutionListener(EvolutionListener listener) {
        if (!evolutionListeners.contains(listener)) evolutionListeners.add(listener);
    }

    /**
     * Remove an object from evolve loop callbacks
     * @param listener An object implementing the EvolutionListener interface, currently registered to this Evolve instance.
     */
    public void removeEvolutionListener(EvolutionListener listener) {
        evolutionListeners.remove(listener);
    }

    private void assignOutputNodes() {
        outputNodes = new ArrayList<>();
        int requiredOutputs = Config.getInt("no.ntnu.cerebrum.medulla.smcgp.outputs");

        // Check for output nodes from end to start
        for (int i = nodes.length - 1; i >= 0; i--) {
            // Check if node is output node
            if (nodes[i][nodes[i].length - 1] == 1.0) {
                outputNodes.add(i);
                requiredOutputs--;

                if (requiredOutputs == 0) {
                    break;
                }
            }
        }

        // If not enough output nodes, convert from end to start
        if (requiredOutputs > 0) {
            for (int i = nodes.length - 1; i >= 0; i--) {
                // Check if node is output node
                if (nodes[i][nodes[i].length - 1] != 1.0) {
                    outputNodes.add(i);
                    requiredOutputs--;

                    if (requiredOutputs == 0) {
                        break;
                    }
                }
            }
        }

        if (requiredOutputs > 0) {
            terminate = true;
            //System.out.println("Not enough nodes for desired output! Terminating individual");
        }

    }

    private double readNode(int absolutePosition) throws TimeoutException {
        // Check that we have not used to long time
        if (this.evaluationStart != null) {
            if (System.currentTimeMillis() - evaluationStart > Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.timeout")) {
                this.fitness = 0.0;
                this.terminate = true;
                throw new TimeoutException("Timeout");
            }
        }

        boolean useSimpleInputFunction = Config.getBoolean("no.ntnu.cerebrum.medulla.smcgp.useSimpleInputFunction");

        // Out of bounds - return Zero Input
        if (absolutePosition < 0 || absolutePosition >= nodes.length) {
            return 0.0;
        } else {
            // Store active node position
            activeNodes.add(absolutePosition);
        }

        // Node caching
        if (nodeCache != null && nodeCache.getOrDefault(absolutePosition, null) != null) {
            //System.out.println("Cached value!");
            return nodeCache.get(absolutePosition);
        }

        // Current node and function of node
        double[] node = nodes[absolutePosition];
        int nodeType = (int) node[0];

        // Return if input node
        if (FunctionLookup.isInput(nodeType)) {
            double terminalInput = 0;

            // Read input and adjust pointer
            switch (nodeType) {
                case 0:
                    terminalInput = input[inputPointer];
                    inputPointer++;
                    break;
                case 1:
                    terminalInput = input[inputPointer];
                    if (useSimpleInputFunction) inputPointer++;
                    else inputPointer--;

                    break;
                case 2:
                    terminalInput = input[inputPointer];
                    if (useSimpleInputFunction) inputPointer++;
                    else inputPointer = (int) Math.abs(inputPointer + node[3]) % input.length;

                    break;
            }

            // Wrap around pointer
            if (inputPointer < 0) {
                inputPointer = input.length - 1;
            } else if (inputPointer > input.length - 1) {
                inputPointer = 0;
            }

            // Caching
            if (Config.getBoolean("no.ntnu.cerebrum.medulla.smcgp.cache")) {
                nodeCache.put(absolutePosition, terminalInput);
            }

            return terminalInput;
        }

        // Connections and Input from connections
        double[] connectionPositions = Arrays.copyOfRange(node, 1, 3);
        double[] neighbourInput = new double[2];

        neighbourInput[0] = readNode((int) (absolutePosition - connectionPositions[0]));
        neighbourInput[1] = readNode((int) (absolutePosition - connectionPositions[1]));

        if (FunctionLookup.isSelfModifying(nodeType)) {

            // SM activation function -> when 2nd input is larger than 1st
            if (neighbourInput[1] > neighbourInput[0]) {
                // Add SM function to todoList
                todoList.add(absolutePosition);
            }

            // Caching
            if (Config.getBoolean("no.ntnu.cerebrum.medulla.smcgp.cache")) {
                nodeCache.put(absolutePosition, neighbourInput[1]);
            }

            // Pass along the second input
            return neighbourInput[1];
        } else if (FunctionLookup.isMathematical(nodeType)) {
            try {
                // Lazy load function and return result
                Function mathFunction = (Function) Class.forName(Config.get("no.ntnu.cerebrum.medulla.smcgp.functionMath") + FunctionLookup.getFunction(nodeType)).newInstance();
                double result = (double) mathFunction.doFunction(neighbourInput, node);

                // Caching
                if (Config.getBoolean("no.ntnu.cerebrum.medulla.smcgp.cache")) {
                    nodeCache.put(absolutePosition, result);
                }

                return result;

            } catch (InstantiationException e) {
                e.printStackTrace();
            } catch (IllegalAccessException e) {
                e.printStackTrace();
            } catch (ClassNotFoundException e) {
                e.printStackTrace();
            }
        }

        System.out.println("This should not happen... interesting!" + nodeType + FunctionLookup.getFunction(nodeType));
        return neighbourInput[1];
    }

    public void resetToGenotypeCopy() {
        this.nodes = this.genotype.copyGenes();
        this.inputPointer = 0;
        this.todoList = new ArrayList<>();
        this.activeNodes = new HashSet<>();
        this.fitness = 0.0;
        this.terminate = false;
        this.lastGraphState = 0;
        assignOutputNodes();
    }

    private void doTodoList() {
        Collections.sort(todoList);

        // Force hash if unset
        if (lastGraphState == 0) {
            lastGraphState = getActiveNodesHash();
            // Notify listeners of initial state
            for (SMCGPProcessingListener spl : smcgpProcessingListeners) {
                spl.iterationComplete(activeNodes.size(), hasActiveInput(), lastGraphState);
            }
        }

        for (int i = 0; i < Math.min(todoList.size(), Config.getInt("no.ntnu.cerebrum.medulla.smcgp.todoSize")); i++) {
            int position = todoList.get(i);
            double[] smNode = nodes[position];

            if (!FunctionLookup.isSelfModifying((int) smNode[0]) || smNode[0] > FunctionLookup.getOpRange()[1]) {
                System.out.println("Non-SM node in todoList or func out of op range: " + Arrays.toString(smNode));
            }

            try {
                // Lazy load the sm function through class loader
                int funcType = (int) smNode[0];
                SMFunction function = (SMFunction) Class.forName(Config.get("no.ntnu.cerebrum.medulla.smcgp.functionSM") + FunctionLookup.getFunction(funcType)).newInstance();
                this.nodes = (double[][]) function.doSMFunction(this.nodes, smNode, position);

                lastGraphState = getActiveNodesHash();

                // Report SM to stats listeners
                for (SMCGPProcessingListener spl : smcgpProcessingListeners) {
                    spl.selfModification(funcType);
                }

            } catch (InstantiationException e) {
                e.printStackTrace();
            } catch (IllegalAccessException e) {
                e.printStackTrace();
            } catch (ClassNotFoundException e) {
                e.printStackTrace();
            }
        }

        // Notify listeners that we are done with this iteration
        for (SMCGPProcessingListener spl : smcgpProcessingListeners) {
            spl.iterationComplete(activeNodes.size(), hasActiveInput(), lastGraphState);
        }

        // Clear todoList after all sm operations are done
        todoList.clear();
    }

    /**
     * This can be refactored away. It just delegates to the getOuput function.
     *
     * @param input A vector of input data to be sent into the SMGCP.
     * @return An output vector og size smcgp.outputs
     */
    @Override
    public double[] fireSignal(double[] input) throws TimeoutException {
        return getOutput(input);
    }

    /**
     * Reset this Phenotype to its primordial state, unaltered by iteration
     */
    @Override
    public void reset() {
        resetToGenotypeCopy();
    }

    /**
     * Builds a hashcode from the activeNodes list for use in topography comparison.
     * @return A hashcode
     */
    public int getActiveNodesHash() {
        return Arrays.deepHashCode(
            activeNodes
                .stream()
                .sorted()
                .filter(i -> i < nodes.length)
                .map(i -> nodes[i])
                .collect(Collectors.toList())
                .toArray()
        );
    }

    public double miniScore(int minStates, int numBatches, int numSamplesPerBatch) {
        int outputs = Config.getInt("no.ntnu.cerebrum.medulla.smcgp.outputs");
        org.nd4j.linalg.api.rng.Random rng = Nd4j.getRandom();
        double maxScore = numBatches * numSamplesPerBatch;
        HashSet<Integer> states = new HashSet<>();
        for (int i = 0; i < numBatches; i++) {
            for (int j = 0; j < numSamplesPerBatch; j++) {
                try {
                    double[] input = new double[outputs+1];
                    input[0] = rng.nextInt(2);
                    for (int x = 1; x < outputs + 1; x++) {
                        input[x] = rng.nextDouble();
                    }
                    fireSignal(input);
                    states.add(getActiveNodesHash());
                    if (!hasActiveInputAndActiveSM()) {
                        return (i * numSamplesPerBatch + j) / maxScore;
                    }
                } catch (TimeoutException e) {
                    return 0.0;
                }
            }
            if (states.size() < 4) return (i * numSamplesPerBatch) / maxScore;
        }
        if (states.size() < minStates) return 0.9;
        return 1.0;
    }

    /**
     * Returns early if when iterated, SM functions disappear from active nodes tree.
     * @return A fitness value.
     */
    public double strictIterativeFitness() {
        long startTime = System.currentTimeMillis();
        // Timeout in milliseconds
        int timeout = 4000;
        // Total number of data sets
        int numRuns = 32;
        // Size of each data set
        int samplesPerRun = 64;

        double f;
        Random rng = Nd4j.getRandom();
        HashSet<Integer> states = new HashSet<>();
        HashSet<Double> outputs = new HashSet<>();

        int[][] attractors = new int[numRuns][samplesPerRun];

        int minStates = Config.getInt("no.ntnu.cerebrum.medulla.smcgp.phenotype.minTopographicStates");
        int maxStates = Config.getInt("no.ntnu.cerebrum.medulla.smcgp.phenotype.maxTopographicStates");
        double statesPenalty = 0.0;
        int numOutputs = Config.getInt("no.ntnu.cerebrum.medulla.smcgp.outputs");
        int numCachedTimeSteps = Config.getInt("no.ntnu.cerebrum.medulla.reservoir.numCachedTimeSteps");
        double minInputNodePercentage = Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.phenotype.minInputNodePercentage");
        int inputExpansionFactor = Config.getInt("no.ntnu.cerebrum.medulla.reservoir.inputExpansionFactor");

        double[] sineData = DataFactory.generateSineSeries(
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.frequency"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.amplitude"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.phase"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.overSamplingFactor"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.seconds")
        );
        // Make a 2D version if we are using multi-output preProcessing
        double[][] sineData2D = new double[sineData.length][1];
        for (int j = 0; j < sineData.length; j++) {
            sineData2D[j][0] = sineData[j];
        }

        for (int x = 1; x < numRuns + 1; x++) {
            // If we have exceeded the timeout, break early to not clog up the population with expensive phenotypes
            if (System.currentTimeMillis() - startTime > timeout) return x * 0.9;

            double[] lastOutput = new double[numOutputs];
            double[][] rawOutput = new double[samplesPerRun][numOutputs];
            try {
                boolean stable = true;
                for (int i = 0; i < samplesPerRun; i++) {
                    // If we are caching more than one time step, merge the Z time steps into a flattened last output array
                    if (numCachedTimeSteps > 1) {
                        for (int z = i - 2; z >= i - numCachedTimeSteps; z--) {
                            if (z < 0) break;
                            lastOutput = DataFactory.merge(
                                lastOutput,
                                rawOutput[z]
                            );
                        }
                    }

                    // Verify that we have the correct number of output values in the cache
                    assert lastOutput.length == numOutputs * numCachedTimeSteps;

                    if (Config.getBoolean("no.ntnu.cerebrum.medulla.reservoir.useInputExpansion")) {
                        lastOutput = fireSignal(DataFactory.merge(DataFactory.fillWith(rng.nextInt(), inputExpansionFactor), lastOutput));
                    } else {
                        lastOutput = fireSignal(DataFactory.merge(new double[]{rng.nextInt()}, lastOutput));
                    }

                    rawOutput[i] = lastOutput;

                    // Store the outputs for later analysis of unique outputs
                    for (double d : lastOutput) outputs.add(d);

                    attractors[x-1][i] = getActiveNodesHash();
                    states.add(attractors[x-1][i]);

                    if (i > samplesPerRun - Config.getInt("no.ntnu.cerebrum.medulla.smcgp.phenotype.pointAttractorStepsBack") &&
                        attractors[x-1][i] != attractors[x-1][i-1]) stable = false;

                    if (!hasActiveInputAndActiveSM()) {
                        if (states.size() < minStates) statesPenalty = minStates - states.size();
                        else if (states.size() > maxStates) statesPenalty = states.size() - maxStates;

                        f = 2.0 * x + (i / (double) samplesPerRun) - statesPenalty;
                        for (EvolutionListener listener : evolutionListeners) listener.fitnessEvaluationComplete(f, x * i + i, INPUT_OR_SM);

                        return f;
                    }
                }
                if (states.size() < 4 || stable) {
                    for (EvolutionListener listener : evolutionListeners) listener.fitnessEvaluationComplete(x, x * samplesPerRun, FEW_STATES_OR_STABLE);
                    return x;
                }

                int input_nodes = 0;
                for (double[] node : nodes) {
                    if (node[0] < 3) input_nodes++;
                }
                if (input_nodes < nodes.length * minInputNodePercentage) {
                    for (EvolutionListener listener : evolutionListeners) listener.fitnessEvaluationComplete(x, x * samplesPerRun, INPUT_PERCENTAGE);
                    return x * 0.95;
                }
            } catch (TimeoutException e) {
                for (EvolutionListener listener : evolutionListeners) listener.fitnessEvaluationComplete(x, x * samplesPerRun, TIMEOUT);
                return x;
            } finally {
                reset();
            }

            // Verify that we have varying output for all of the runs...
            if (Config.getBoolean("no.ntnu.cerebrum.medulla.smcgp.phenotype.useNumericalConvergenceCheck")) {
                int convergedColumns = 0;
                for (int j = 0; j < numOutputs; j++) {
                    if (Evaluation.isNumericallyConverged(
                        DataFactory.getColumn(j, rawOutput),
                        Config.getInt("no.ntnu.cerebrum.medulla.smcgp.phenotype.numericalConvergenceStepsBack"),
                        Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.phenotype.numericalConvergenceRoundingThreshold"))) {

                        convergedColumns++;
                    }

                }

                // We want to have usable output in at least two thirds of the number of outputs
                if (numOutputs - convergedColumns < numOutputs * 0.66) {
                    f = 2.0 * x - 5.0;
                    for (EvolutionListener listener : evolutionListeners) listener.fitnessEvaluationComplete(f, x * samplesPerRun, CONVERGED);

                    return f;
                }
            }

            // If we use more than 1s on one data set, abort, since it's too expensive for this purpose..
            if (System.currentTimeMillis() - startTime > timeout) {
                log.debug(
                    "Strict iterative fitness timed out after a run that took " + (System.currentTimeMillis() - startTime) + "ms"
                );

                f = 2.0 * x;
                for (EvolutionListener listener : evolutionListeners) listener.fitnessEvaluationComplete(f, x * samplesPerRun, TIMEOUT);

                return f;
            }
        }

        if (states.size() < minStates) statesPenalty = minStates - states.size();
        else if (states.size() > maxStates) statesPenalty = 2.0;

        int attractorLength = 0;
        for (int i = samplesPerRun - 1; i > samplesPerRun / 4 - 1; i--) {
            for (int y = 0; y < numRuns - 1; y++) {
                if (attractors[y][i] != attractors[y+1][i]) {
                    attractorLength = i;
                    break;
                }
            }
        }

        if (attractorLength == 0) {
            f = 64 - statesPenalty - 16;
            for (EvolutionListener listener : evolutionListeners) listener.fitnessEvaluationComplete(f, numRuns * samplesPerRun, ATTRACTOR_LENGTH);

            return f;
        }
        if (outputs.size() < Config.getInt("no.ntnu.cerebrum.medulla.smcgp.phenotype.minUniqueValues")) {
            f = 64 - statesPenalty - 8;
            for (EvolutionListener listener : evolutionListeners) listener.fitnessEvaluationComplete(f, numRuns * samplesPerRun, UNIQUE_VALUES);

            return f;
        }

        // Send a sine wave signal through a reservoir with this Phenotype, and inspect the power spectrum of the output...
        SMCGPReservoir r = new SMCGPReservoir(this);
        double[][] sineResults;

        try {
            sineResults = r.preProcess(sineData2D);
            reset();
        } catch (TimeoutException e) {
            reset();

            for (EvolutionListener listener : evolutionListeners) listener.fitnessEvaluationComplete(0.0, numRuns * samplesPerRun, TIMEOUT);

            return 0.0;
        }

        double rt = Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.phenotype.powerSpectrumRoundingThreshold");

        int frequencySpikes = 0;
        double[] spectrum;

        for (int k = 0; k < numOutputs; k++) {
            spectrum = Evaluation.powerSpectrumDensity(
                Fourier.forwardDFT(DataFactory.getColumn(k, sineData.length / 2, sineResults)),
                rt
            );
            frequencySpikes += Evaluation.numDistinctSpectrumHarmonics(spectrum);
        }

        spectrum = Evaluation.powerSpectrumDensity(
            Fourier.forwardDFT(Arrays.copyOf(sineResults[0], sineResults.length / 2)),
            rt
        );
        frequencySpikes = Evaluation.numDistinctSpectrumHarmonics(spectrum);

        // Just reduce a bit based on the number of possible harmonics found from spectrum,
        // but put a small emphasis on it
        int minPowerSpectrumStates = Config.getInt("no.ntnu.cerebrum.medulla.smcgp.phenotype.minPowerSpectrumStates");
        if (frequencySpikes < minPowerSpectrumStates) {
            f = 64.0 - statesPenalty - (minPowerSpectrumStates - frequencySpikes);
            for (EvolutionListener listener : evolutionListeners) listener.fitnessEvaluationComplete(f, numRuns * samplesPerRun + sineData.length * numOutputs, SPECTRUM_HARMONICS);

            return f;
        }

        for (EvolutionListener listener : evolutionListeners) listener.fitnessEvaluationComplete(64 - statesPenalty, numRuns * samplesPerRun + sineData.length * numOutputs, NONE);

        return 64.0 - statesPenalty;
    }

    public double[] getThroughputDryRun(double[] data) {

        // Set up SMCGPReservoir with readout layer
        SMCGPReservoir r = new SMCGPReservoir(this);

        try {
            double[] result = r.preProcess(data);
            reset();
            return result;
        } catch (TimeoutException e) {
            System.out.println("timeout....");
            reset();
            return new double[]{0.0};
        }
    }

    public int getStates() {
        // Make reservoir and create standard data set.
        SMCGPReservoir r = new SMCGPReservoir(this);
        double[] data = DataFactory.generateSineSeries(
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.frequency"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.amplitude"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.phase"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.overSamplingFactor"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.seconds")
        );

        try {
            double[] output = r.preProcess(data);
            Complex[] result = Fourier.forwardDFT(output);
            // We drop the second half the output since we are only dealing with real valued numbers
            double[] spectrum = Evaluation.powerSpectrumDensity(Arrays.copyOf(result, result.length / 2), 10E-8);
            int states = Evaluation.numDistinctSpectrumHarmonics(spectrum);

            resetToGenotypeCopy();

            return states;

        } catch (TimeoutException e) {
            resetToGenotypeCopy();
            return 0;
        }
    }

    /**
     * Iterate through the activeNodes set to check for input nodes.
     * @return true if an input node is in the active nodes set, false otherwise.
     */
    public boolean hasActiveInput() {
        for (Integer i : activeNodes) {
            if (i < nodes.length && FunctionLookup.isInput((int) nodes[i][0])) return true;
        }

        return false;
    }

    /**
     * Iterate through the activeNodes set to check for SM nodes.
     * @return true if an SM node is in the active nodes set, false otherwise.
     */
    public boolean hasActiveSM() {
        for (Integer i : activeNodes) {
            if (i < nodes.length && FunctionLookup.isSelfModifying((int) nodes[i][0])) return true;
        }

        return false;
    }

    /**
     * Iterate through the activeNodes set to check for SM nodes.
     * @return true if an SM node is in the active nodes set, false otherwise.
     */
    public boolean hasActiveInputAndActiveSM() {
        boolean hasInput = false;
        boolean hasSM = false;

        for (Integer i : activeNodes) {
            if (i < nodes.length) {
                int funcType = (int) nodes[i][0];
                if (FunctionLookup.isInput(funcType)) hasInput = true;
                if (FunctionLookup.isSelfModifying(funcType)) hasSM = true;
            }
        }

        return hasInput && hasSM;
    }

    public boolean isTerminated() {
        return terminate;
    }

    public String getUuid() {
        return genotype.getUuid();
    }

    public void setUuid(String uuid) {
        genotype.setUuid(uuid);
    }

    public String toString() {
        return String.format("[Fit: %s, Gen: %s]", this.fitness, this.generation);
    }
}
