package no.ntnu.cerebrum.medulla.smcgp.evolution;

import no.ntnu.cerebrum.medulla.common.EvolveCaller;
import no.ntnu.cerebrum.medulla.config.Config;
import no.ntnu.cerebrum.medulla.data.Data;
import no.ntnu.cerebrum.medulla.data.DataFactory;
import no.ntnu.cerebrum.medulla.smcgp.Genotype2;
import no.ntnu.cerebrum.medulla.smcgp.Phenotype2;
import no.ntnu.cerebrum.medulla.smcgp.utility.NumberGenerator;
import org.apache.commons.math3.stat.descriptive.SummaryStatistics;
import org.apache.log4j.PropertyConfigurator;
import org.slf4j.Logger;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.concurrent.*;

public class Evolve {

    // Logging
    protected static Logger log = org.slf4j.LoggerFactory.getLogger(Evolve.class);
    public static ExecutorService es = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());
    protected ArrayList<Phenotype2> population;
    protected EvolveCaller observer;
    protected int currentEpoch;
    private ArrayList<EvolutionListener> evolutionListeners;

    // Fitness comparator
    public static Comparator<Phenotype2> cmp = Comparator.comparing(Phenotype2::getFitness, Comparator.reverseOrder())
            .thenComparing(Phenotype2::getGeneration, Comparator.reverseOrder());

    // Elitism
    protected Phenotype2 bestIndividual;

    // Training data containers
    private double[][] trainingData;
    private double[][] trainingLabels;

    // Test data containers
    private double[][] testData;
    private double[][] testLabels;

    public static double[] noisySineData;

    // Statistics
    private SummaryStatistics ss = new SummaryStatistics();

    // Parent Selection
    private ArrayList<Phenotype2> parents;
    private String selectionScheme = Config.get("no.ntnu.cerebrum.medulla.smcgp.ea.parentSelectionScheme");
    private int parentPoolSize = Config.getInt("no.ntnu.cerebrum.medulla.smcgp.ea.parentPool");

    public Evolve() {
        generateRandomTemporalDensityData();
        this.evolutionListeners = new ArrayList<>();
        noisySineData = DataFactory.generateSineSeries(
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.frequency"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.amplitude"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.phase"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.overSamplingFactor"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.seconds")
        );
        // DataFactory.addGaussianNoise(noisySineData, 5.0);
    }

    public Evolve(EvolveCaller caller) {
        generateRandomTemporalDensityData();
        this.observer = caller;
        this.evolutionListeners = new ArrayList<>();

        noisySineData = DataFactory.generateSineSeries(
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.frequency"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.amplitude"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.phase"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.overSamplingFactor"),
            Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.signal.seconds")
        );
        // DataFactory.addGaussianNoise(noisySineData, 5.0);
    }


    /**
     * Clears the whole population
     */
    protected void clearPopulation() {
        if (this.population != null) {
            population.clear();
        } else {
            population = new ArrayList<>();
        }
    }

    /**
     * Schedules the provided argument runnable for immediate invocation in the executor pool
     *
     * @param r An object implementing the Runnable interface.
     */
    public void scheduleJob(Runnable r) {
        es.execute(r);
    }

    /**
     * Randomly create initial population
     */
    protected void generateRandomInitialPopulation() {
        log.info("Adding initial phenotypes...");

        // Clear population
        clearPopulation();

        // Generate random
        int initialSize = Config.getInt("no.ntnu.cerebrum.medulla.smcgp.ea.initialPopulation");
        while (population.size() < initialSize) {
            population.add(new Phenotype2(new Genotype2()));
        }

        log.info(String.format("Initial population of %d added.", population.size()));
    }

    /**
     * Register an object for evolve loop callbacks
     * @param listener An object implementing the EvolutionListener interface
     */
    public void addEvolutionListener(EvolutionListener listener) {
        if (!evolutionListeners.contains(listener)) evolutionListeners.add(listener);
    }

    /**
     * Remove an object from evolve loop callbacks
     * @param listener An object implementing the EvolutionListener interface, currently registered to this Evolve instance.
     */
    public void removeEvolutionListener(EvolutionListener listener) {
        evolutionListeners.remove(listener);
    }

    /**
     * Generate the next generation of children
     *
     * @param master the gender-neutral parent
     */
    protected void generateOffspring(Genotype2 master) {
        // Clear population current population
        clearPopulation();

        // What is our current mutation rate?
        double mutationProbability = Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.ea.mutationRate");

        // Elitism - keep the best individual
        if (Config.getBoolean("no.ntnu.cerebrum.medulla.smcgp.ea.elitism")) {
            population.add(new Phenotype2(bestIndividual.getGenotype(), bestIndividual.getGeneration()));
        }

        // Create offspring from parents
        while (population.size() < Config.getInt("no.ntnu.cerebrum.medulla.smcgp.ea.poolSize")) {
            if (Config.getBoolean("no.ntnu.cerebrum.medulla.smcgp.ea.crossover")) {
                for (int i = 1; i < parents.size(); i++) {
                    Phenotype2 mother = parents.get(i - 1);
                    Phenotype2 father = parents.get(i);

                    // Breed parents
                    Genotype2 offspring = Genotype2.crossover(mother.getGenotype(), father.getGenotype());

                    // Mutate offspring
                    offspring.mutate(mutationProbability, true);

                    // Develop child
                    Phenotype2 child = new Phenotype2(offspring, mother.getGeneration() + 1);

                    population.add(child);
                }

            } else {
                for (Phenotype2 parent : parents) {

                    // Copy the genotype of the best individual
                    Genotype2 mutated = new Genotype2(parent.getGenotype().copyGenes());

                    // Mutation chance
                    mutated.mutate(mutationProbability, true);

                    // Develop child
                    Phenotype2 offspring = new Phenotype2(mutated, parent.getGeneration() + 1);

                    population.add(offspring);
                }
            }
        }
    }

    /**
     * Evaluates population with checks per iteration, breaking early if SM is iterated away...
     */
    public void evaluatePopulation6() {
        resetStats();

        List<Future> tasks = new ArrayList<>();
        for (Phenotype2 phenotype : population) {
            tasks.add(es.submit(() -> {
                for (EvolutionListener listener : evolutionListeners) phenotype.addEvolutionListener(listener);
                phenotype.setFitness(phenotype.strictIterativeFitness());
                for (EvolutionListener listener : evolutionListeners) phenotype.removeEvolutionListener(listener);
            }));
        }

        for (Future f : tasks) {
            try {
                f.get();
            } catch (InterruptedException e) {
                e.printStackTrace();
            } catch (ExecutionException e) {
                e.printStackTrace();
            }
        }
    }


    /**
     * Select parents from the population
     */
    protected void parentSelection() {
        // Find the current best individual - Highest fitness and Generation in current population
        if (bestIndividual == null) bestIndividual = population.get(0);
        for (int i = 0; i < population.size(); i++) {
            Phenotype2 current = population.get(i);

            // Collect some stats for this epoch
            ss.addValue(current.getFitness());
            // Actual comparison with all-time best individual
            if (current.getFitness() > bestIndividual.getFitness() || (current.getFitness() == bestIndividual.getFitness() &&
                    current.getGeneration() > bestIndividual.getGeneration())) {
                bestIndividual = current;
            }
        }

        // Create new parent pool
        parents = new ArrayList<>();

        // Elitism - keep the best individual
        if (Config.getBoolean("no.ntnu.cerebrum.medulla.smcgp.ea.elitism")) {
            parents.add(new Phenotype2(bestIndividual.getGenotype(), bestIndividual.getGeneration()));
        }

        // Selection Scheme
        switch (selectionScheme) {
            case "tournament":
                while (parents.size() < parentPoolSize) {

                    // Gather tournament participants
                    ArrayList<Phenotype2> tournament = new ArrayList<>();
                    int tournamentBest = 0;
                    for (int i = 0; i < Config.getInt("no.ntnu.cerebrum.medulla.smcgp.ea.tournamentSize"); i++) {
                        tournament.add(population.get(NumberGenerator.generateRandomInt(0, population.size() - 1)));
                        if (tournament.get(i).getFitness() >= tournament.get(tournamentBest).getFitness()) {
                            tournamentBest = i;
                        }
                    }

                    // Add the best individual from the tournament
                    parents.add(tournament.get(tournamentBest));
                }
                break;
            case "best":
                int parentIndex = parents.size();
                population.sort(cmp);
                while (parentIndex < parentPoolSize) {
                    parents.add(population.get(parentIndex));
                    parentIndex++;
                }
                break;
            case "1+X":
                // If elitism, this is already ensued
                if (parents.size() <= 0) {
                    parents.add(new Phenotype2(bestIndividual.getGenotype()));
                }
                break;
        }
    }

    public Phenotype2 evolve() {
        // 1. Create initial population
        generateRandomInitialPopulation();

        // 2. Evaluate all individuals N times
        evaluatePopulation6();

        // 3. Pick best individual
        parentSelection();

        // Evolutionary Strategy
        int maxEpochs = Config.getInt("no.ntnu.cerebrum.medulla.smcgp.ea.maxEpochs");
        currentEpoch = 0;
        while (currentEpoch < maxEpochs) {
            // Generate offspring for the best individuals genes
            generateOffspring(bestIndividual.getGenotype());

            // Send population to reservoir for evaluation
            evaluatePopulation6();

            // Pick the best individual - stores best in bestIndividual
            // Based on fitness(-) and age(+)
            parentSelection();

            log.info(String.format(
                    "[%d] Best: Phenotype{%.2f-%s/%s}, %s",
                    currentEpoch,
                    bestIndividual.getFitness(),
                    bestIndividual.getLastGraphSize(),
                    bestIndividual.getLastActiveNodeCount(),
                    stats()
            ));

            // If we have no errors - break
            if (bestIndividual.getFitness() >= Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.ea.stopCondition")) {
                break;
            }

            currentEpoch++;
        }

        log.info(String.format("Evolution done in %s epochs, Best: %s", currentEpoch, bestIndividual.toString()));

        if (observer != null) {
            observer.evolutionCompleted();
        }

        // Return best individual
        return bestIndividual;
    }

    public ArrayList<Phenotype2> getPopulation() {
        return population;
    }

    public int getCurrentEpoch() {
        return this.currentEpoch;
    }

    private void generateRandomTemporalParityData() {
        generateRandomTemporalParityTrainingData();
        generateRandomTemporalParityTestData();
    }

    private void generateRandomTemporalDensityData() {
        generateRandomTemporalDensityTrainingData();
        generateRandomTemporalDensityTestData();
    }

    private void generateRandomTemporalParityTrainingData() {
        int trainingSamples = Config.getInt("dataset.trainingSamples");
        int windowSize = Config.getInt("dataset.temporal.windowSize");
        int signalDelay = Config.getInt("dataset.temporal.delay");
        boolean useOneHotRepresentation = Config.getBoolean("dataset.temporal.useOneHotRepresentation");

        Data d = DataFactory.createTemporalParityDataWithEvenParityBitDistribution(windowSize, signalDelay, trainingSamples, useOneHotRepresentation);
        trainingData = d.input();
        trainingLabels = d.output();
    }

    private void generateRandomTemporalDensityTrainingData() {
        int trainingSamples = Config.getInt("dataset.trainingSamples");
        // Training data containers
        trainingData = new double[trainingSamples][];
        trainingLabels = new double[trainingSamples][];

        generateRandomTemporalDensityData(trainingData, trainingLabels);
    }

    private void generateRandomTemporalParityTestData() {
        int testSamples = Config.getInt("dataset.testSamples");
        int windowSize = Config.getInt("dataset.temporal.windowSize");
        int signalDelay = Config.getInt("dataset.temporal.delay");
        boolean useOneHotRepresentation = Config.getBoolean("dataset.temporal.useOneHotRepresentation");

        Data d = DataFactory.createTemporalParityDataWithEvenParityBitDistribution(windowSize, signalDelay, testSamples, useOneHotRepresentation);
        testData = d.input();
        testLabels = d.output();
    }

    private void generateRandomTemporalDensityTestData() {
        int testSamples = Config.getInt("dataset.testSamples");
        // Test data containers
        testData = new double[testSamples][];
        testLabels = new double[testSamples][];

        generateRandomTemporalDensityData(testData, testLabels);
    }

    private void generateRandomTemporalParityData(double[][] input, double[][] target) {
        int windowSize = Config.getInt("dataset.temporal.windowSize");
        int signalDelay = Config.getInt("dataset.temporal.delay");
        boolean useOneHotRepresentation = Config.getBoolean("dataset.temporal.useOneHotRepresentation");

        DataFactory.createRandomTemporalParityData(input, target, windowSize, signalDelay, input.length, useOneHotRepresentation);
    }

    private void generateRandomTemporalDensityData(double[][] input, double[][] target) {
        int windowSize = Config.getInt("dataset.temporal.windowSize");
        int signalDelay = Config.getInt("dataset.temporal.delay");
        boolean useOneHotRepresentation = Config.getBoolean("dataset.temporal.useOneHotRepresentation");

        DataFactory.createRandomTemporalDensityData(input, target, windowSize, signalDelay, input.length, useOneHotRepresentation);
    }

    public void resetStats() {
        ss = new SummaryStatistics();
    }

    public String stats() {
        return String.format(
            "Population{%.3f/%.3f/%.3f-%.4f}",
            ss.getMax(),
            ss.getMean(),
            ss.getMin(),
            ss.getStandardDeviation()
        );
    }
}
