package no.ntnu.cerebrum.medulla.smcgp.graph.action;

import com.google.common.collect.ImmutableSet;
import com.google.common.reflect.ClassPath;
import no.ntnu.cerebrum.medulla.config.Config;
import org.nd4j.linalg.api.rng.Random;
import org.nd4j.linalg.factory.Nd4j;
import org.slf4j.Logger;

import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;

public class FunctionLookup {

    private static Map<Integer, String> functionTable;
    private static int[] opRange;
    private static int[] mathRange;
    private static int[] smRange;

    protected static Logger log = org.slf4j.LoggerFactory.getLogger(FunctionLookup.class);
    protected static Random rng = Nd4j.getRandom();

    /**
     * static method that assigns classes to the lookup table
     * used to identify which function a node should use
     */
    static {
        functionTable = new HashMap<>();
        opRange = new int[2];
        // Input & SMCGP specific
        functionTable.put(0, "Input");
        functionTable.put(1, "Input Decrement");
        functionTable.put(2, "Input Skip");

        int lookUpPosition = 3;

        // Mathematical
        ArrayList<String> mathFunctions = getAllClasses(Config.get("smcgp.functionMath"));
        mathFunctions = applyClassWhiteList(mathFunctions, Config.getList("smcgp.functionMath.whiteList"));
        Collections.sort(mathFunctions);

        mathRange = new int[2];
        mathRange[0] = lookUpPosition;
        for (String mathFunc : mathFunctions) {
            functionTable.put(lookUpPosition, mathFunc);
            lookUpPosition++;
        }
        mathRange[1] = lookUpPosition - 1;

        // Self Modifying
        ArrayList<String> smFunctions = getAllClasses(Config.get("smcgp.functionSM"));
        Collections.sort(smFunctions);

        smRange = new int[2];
        smRange[0] = lookUpPosition;
        for (String smFunc : smFunctions) {
            functionTable.put(lookUpPosition, smFunc);
            lookUpPosition++;
        }
        smRange[1] = lookUpPosition - 1;

        // Store range
        opRange[0] = (int) functionTable.keySet().toArray()[0];
        opRange[1] = (int) functionTable.keySet().toArray()[functionTable.keySet().toArray().length - 1];

        log.info("SM Classes loaded: " + Arrays.toString(smFunctions.toArray()));
        log.info("Math Classes loaded: " + Arrays.toString(mathFunctions.toArray()));
    }

    public static String getFunction(int functionIndex) {
        if (functionTable.containsKey(functionIndex)) {
            return functionTable.get(functionIndex);
        } else {
            log.error("Lookup function at " + functionIndex + " does not EXIST!");
        }
        return "";
    }

    public static int getIndexForFunction(String functionName) {
        for (int key : functionTable.keySet()) {
            if (functionTable.get(key).equals(functionName)) {
                return key;
            }
        }
        return -1;
    }

    public static int[] getOpRange() {
        return opRange;
    }

    public static int[] getMathRange() {
        return mathRange;
    }

    public static int[] getSmRange() {
        return smRange;
    }

    /**
     * Lookup for smcgp functions based on first chromosome value
     */
    public static boolean isMathematical(int functionIndex) {
        return functionIndex >= mathRange[0] && functionIndex <= mathRange[1];
    }

    /**
     * Lookup for smcgp functions based on first chromosome value
     */
    public static boolean isSelfModifying(int functionIndex) {
        return functionIndex >= smRange[0] && functionIndex <= smRange[1];
    }

    public static boolean isInput(int functionIndex) {
        return functionIndex >= 0 && functionIndex <= 2;
    }

    /**
     * Locates all classes within package
     *
     * @param pckgname - the package to hunt for classes
     */
    public static ArrayList<String> getAllClasses(String pckgname) {
        ArrayList<String> classes = new ArrayList<>();

        // Remove trailing '.'
        String _package;
        if (pckgname.charAt(pckgname.length() - 1) == '.') {
            _package = pckgname.substring(0, pckgname.length() - 1);
        } else {
            _package = pckgname;
        }

        // Load classes from classpath
        ClassLoader loader = Thread.currentThread().getContextClassLoader();
        try {
            ClassPath clp = ClassPath.from(loader);
            ImmutableSet<ClassPath.ClassInfo> classesInfo = clp.getTopLevelClasses(_package);
            classes.addAll(classesInfo.stream().map(ClassPath.ClassInfo::getSimpleName).collect(Collectors.toList()));
        } catch (IOException e) {
            e.printStackTrace();
        }

        return classes;
    }

    /**
     * Method for returning the intersect of all classes and desired classes
     *
     * @param classes   the list of all classes
     * @param whiteList the list of the allowed classes
     * @return
     */
    public static ArrayList<String> applyClassWhiteList(ArrayList<String> classes, String[] whiteList) {
        if (whiteList != null && whiteList.length > 0 && !whiteList[0].toLowerCase().equals("all")) {
            ArrayList<String> allowed = new ArrayList(Arrays.asList(whiteList));
            System.out.println(Arrays.toString(allowed.toArray()));
            List<String> intersect = classes.stream().filter(allowed::contains).collect(Collectors.toList());
            log.info("White-listed math functions:");
            System.out.println("White-listed math functions:");
            intersect.stream().forEach(log::info);
            return new ArrayList<>(intersect);
        } else {
            log.info("White-listed math functions: ALL");
        }
        return classes;
    }

    public static int getRandomInputFunction() {
        return rng.nextInt(3);
    }

    public static int getRandomMathematicalFunction() {
        return mathRange[0] + rng.nextInt(mathRange[1] + 1);    // nextInt's are non-inclusive
    }

    public static int getRandomSMFunction() {
        return smRange[0] + rng.nextInt(smRange[1] + 1);    // nextInt's are non-inclusive
    }

    public static int getRandomFunction() {
        return opRange[0] + rng.nextInt(opRange[1] + 1);    // nextInt's are non-inclusive
    }

    public static int getWeightedRandomFunction() {
        double rand = rng.nextDouble();
        // TODO: FIX, goes out of bounds!
        if (rand <= 0.25) {
            return getRandomInputFunction();
        } else if (rand > 0.25 && rand < 0.75) {
            return getRandomMathematicalFunction();
        } else {
            return getRandomSMFunction();
        }
    }

}
