package no.ntnu.cerebrum.medulla.telemetry;

import no.ntnu.cerebrum.medulla.config.Config;
import no.ntnu.cerebrum.medulla.reservoir.interfaces.Reservoir;
import no.ntnu.cerebrum.medulla.reservoir.interfaces.ReservoirTrainingListener;
import no.ntnu.cerebrum.medulla.reservoir.interfaces.SMCGPProcessingListener;
import no.ntnu.cerebrum.medulla.smcgp.graph.action.FunctionLookup;
import org.apache.commons.math3.stat.descriptive.SummaryStatistics;
import org.apache.log4j.Logger;

import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Created on 08.04.2017.
 */
public class ReservoirStatsCollector implements ReservoirTrainingListener, SMCGPProcessingListener {
    private Logger log = Logger.getLogger(ReservoirStatsCollector.class);
    private SummaryStatistics trainingTime;
    private SummaryStatistics preProcessingTime;
    private SummaryStatistics activeNodesSize;
    private SummaryStatistics trainingSetMSE;
    private SummaryStatistics trainingSetRMSE;
    private SummaryStatistics trainingSetNRMSE;
    private SummaryStatistics trainingSetRNMSE;
    private SummaryStatistics trainingSetMAE;
    private SummaryStatistics testSetMSE;
    private SummaryStatistics testSetRMSE;
    private SummaryStatistics testSetNRMSE;
    private SummaryStatistics testSetRNMSE;
    private SummaryStatistics testSetMAE;
    private ArrayList<Double> trainingSetFitnessHistory;
    private ArrayList<Double> testSetFitnessHistory;
    private HashMap<Integer, Integer> selfModificationTypeCount;
    private HashSet<Integer> activeNodeStates;
    private ArrayList<Integer> activeNodeStateHistory;
    private int totalInputsProcessed;
    private int inputPresentAfterIteration;
    private int fitnessEvalEveryEpoch;
    private int epochsTrained;
    private long lastEpochTime;
    private double trainingSetFitness;
    private double testSetFitness;

    public ReservoirStatsCollector() {
        trainingTime = new SummaryStatistics();
        preProcessingTime = new SummaryStatistics();
        activeNodesSize = new SummaryStatistics();
        trainingSetMSE = new SummaryStatistics();
        trainingSetRMSE = new SummaryStatistics();
        trainingSetNRMSE = new SummaryStatistics();
        trainingSetRNMSE = new SummaryStatistics();
        trainingSetMAE = new SummaryStatistics();
        testSetMSE = new SummaryStatistics();
        testSetRMSE = new SummaryStatistics();
        testSetNRMSE = new SummaryStatistics();
        testSetRNMSE = new SummaryStatistics();
        testSetMAE = new SummaryStatistics();
        totalInputsProcessed = 0;
        inputPresentAfterIteration = 0;
        trainingSetFitnessHistory = new ArrayList<>();
        testSetFitnessHistory = new ArrayList<>();
        selfModificationTypeCount = new HashMap<>();
        activeNodeStates = new HashSet<>();
        activeNodeStateHistory = new ArrayList<>();
        fitnessEvalEveryEpoch = Config.getInt("no.ntnu.cerebrum.medulla.reservoir.evaluateEveryNumEpochs");
        lastEpochTime = System.currentTimeMillis();
    }

    /**
     * Retrieve the final fitness score reported for the training set.
     * @return The fitness score.
     */
    public double getTrainingSetFitness() {
        return trainingSetFitness;
    }

    /**
     * Retrieve the fitness history of the training set.
     * @return An ArrayList of fitness values.
     */
    public ArrayList<Double> getTrainingSetFitnessHistory() {
        return trainingSetFitnessHistory;
    }

    /**
     * Retrieve the final fitness score reported for the test set.
     * @return The fitness score.
     */
    public double getTestSetFitness() {
        return testSetFitness;
    }

    /**
     * Retrieve the fitness history of the test set.
     * @return An ArrayList of fitness values.
     */
    public ArrayList<Double> getTestSetFitnessHistory() {
        return testSetFitnessHistory;
    }

    /**
     * Retrieve average training time per epoch, based on mean time between test results and test interval from config.
     * @return The average training time per epoch.
     */
    public double getAverageTrainingTimePerEpoch() {
        return trainingTime.getMean() / fitnessEvalEveryEpoch;
    }

    /**
     * Retrieve the average pre-processing time. NB: Does not take into account that input sizes may vary.
     * @return The current average pre-processing time.
     */
    public double getAveragePreProcessingTime() {
        return preProcessingTime.getMean();
    }

    /**
     * Retrieve the current average number of epochs trained per second
     * @return The average number of epochs trained per second.
     */
    public double getNumEpochsPerSecond() {
        return 1.0 / (trainingTime.getMean() / (1000.0 * fitnessEvalEveryEpoch));
    }

    /**
     * Retrieve the total time spent training the readout layer.
     * @return The current total training time.
     */
    public double getTotalTrainingTime() {
        return trainingTime.getSum() / 1000.0;
    }

    /**
     * Retrieve the total number of inputs that have been processed through the SMCGP Phenotype.
     * @return The current total number of processed inputs.
     */
    public int getTotalInputsProcessed() {
        return totalInputsProcessed;
    }

    /**
     * Retrieve the total time spent pre-processing inputs through the SMCGP.
     * @return The current total pre-processing time.
     */
    public double getTotalPreProcessingTime() {
        return preProcessingTime.getSum();
    }

    /**
     * Retrieve the last epoch number registered on this stats object.
     * @return The number of epochs trained.
     */
    public int getNumEpochsTrained() {
        return epochsTrained;
    }

    /**
     * Retrieve the minimum number of active nodes registered.
     * @return The min number of active nodes in use in the phenotype.
     */
    public double getMinActiveNodes() {
        return activeNodesSize.getMin();
    }

    /**
     * Retrieve the mean number of active nodes registered.
     * @return The mean number of active nodes in use in the phenotype.
     */
    public double getMeanActiveNodes() {
        return activeNodesSize.getMean();
    }

    /**
     * Retrieve the max number of active nodes registered.
     * @return The max number of active nodes in use in the phenotype.
     */
    public double getMaxActiveNodes() {
        return activeNodesSize.getMax();
    }

    /**
     * Retrieve the total number of SM operations seen by this collector.
     * @return The total number of applied SM operations.
     */
    public int getNumSMFunctionsApplied() {
        int i = 0;

        for (Integer x : selfModificationTypeCount.values()) {
            i += x;
        }

        return i;
    }

    /**
     * Retrieve a summary of SM nodes that have been activated, with the total number of activations per function type.
     * @return A SM node activation summary.
     */
    public String getSelfModificationNodeActivationCounts() {
        StringBuilder sb = new StringBuilder();

        for (Integer k : selfModificationTypeCount.keySet()) {
            sb.append(String.format("%s: %d ", FunctionLookup.getFunction(k), selfModificationTypeCount.get(k)));
        }

        return sb.toString();
    }

    /**
     * Retrieve the total number of unique states among the active nodes in a phenotypoe
     * @return The total number of unique states seen.
     */
    public int getNumActiveNodeStates() {
        return activeNodeStates.size() + 1;
    }

    /**
     * Retrieve the entire active node state history seen by this collector.
     * @return The entire active node state history.
     */
    public ArrayList<Integer> getActiveNodeStateHistory() {
        return activeNodeStateHistory;
    }

    /**
     * Retrieve the entire active node state history seen by this collector as a primitive array.
     * @return The entire active node state history.
     */
    public int[] getActiveNodeStateHistoryAsArray() {
        int[] activeNodeStates = new int[activeNodeStateHistory.size()];
        for (int i = 0; i < activeNodeStates.length; i++) activeNodeStates[i] = activeNodeStateHistory.get(i);

        return activeNodeStates;
    }

    /**
     * Retrieve the factor of which after an SM operation has been applied, the Phenotype still possesses an active
     * input node in the node tree.
     * @return The percentage of SM operations maintaining an active input node.
     */
    public double getActiveInputPercentage() {
        double numSmApplied = getNumSMFunctionsApplied();

        if (numSmApplied == 0) return 0.0;

        return inputPresentAfterIteration / totalInputsProcessed;
    }

    public double getTestSetMSE() { return testSetMSE.getMin(); }
    public double getTestSetRMSE() { return testSetRMSE.getMin(); }
    public double getTestSetNRMSE() { return testSetNRMSE.getMin(); }
    public double getTestSetRNMSE() { return testSetRNMSE.getMin(); }
    public double getTestSetMAE() { return testSetMAE.getMin(); }
    public double getTrainingSetMSE() { return trainingSetMSE.getMin(); }
    public double getTrainingSetRMSE() { return trainingSetRMSE.getMin(); }
    public double getTrainingSetNRMSE() { return trainingSetNRMSE.getMin(); }
    public double getTrainingSetRNMSE() { return trainingSetRNMSE.getMin(); }
    public double getTrainingSetMAE() { return trainingSetMAE.getMin(); }

    @Override
    public void epochComplete(int epoch, double testSetFitness, double trainingSetFitness) {
        this.trainingSetFitnessHistory.add(trainingSetFitness);
        this.testSetFitnessHistory.add(testSetFitness);
        this.testSetFitness = testSetFitness;
        this.trainingSetFitness = trainingSetFitness;
        this.epochsTrained = epoch;

        // Store time since last call
        trainingTime.addValue(System.currentTimeMillis() - lastEpochTime);
        lastEpochTime = System.currentTimeMillis();

        log.info(String.format(
            "[%d] Accuracy: %.6f (%.6f) Time: %.1fs (%.1f e/s)",
            epoch,
            testSetFitness,
            trainingSetFitness,
            getTotalTrainingTime(),
            getNumEpochsPerSecond()
        ));
    }

    @Override
    public void epochComplete(int epoch, double trainingSetMSE, double trainingSetRMSE, double trainingSetNRMSE, double trainingSetRNMSE, double trainingSetMAE,
                              double testSetMSE, double testSetRMSE, double testSetNRMSE, double testSetRNMSE, double testSetMAE) {
        this.trainingSetMSE.addValue(trainingSetMSE);
        this.trainingSetRMSE.addValue(trainingSetRMSE);
        this.trainingSetNRMSE.addValue(trainingSetNRMSE);
        this.trainingSetRNMSE.addValue(trainingSetRNMSE);
        this.trainingSetMAE.addValue(trainingSetMAE);
        this.testSetMSE.addValue(testSetMSE);
        this.testSetRMSE.addValue(testSetRMSE);
        this.testSetNRMSE.addValue(testSetNRMSE);
        this.testSetRNMSE.addValue(testSetRNMSE);
        this.testSetMAE.addValue(testSetMAE);

        log.info(String.format(
            "[%d] TestSet: MSE: %.6f RMSE: %.6f NRMSE: %.6f RNMSE: %.6f MAE: %.6f",
            epoch,
            testSetMSE,
            testSetRMSE,
            testSetNRMSE,
            testSetRNMSE,
            testSetMAE
        ));
        log.info(String.format(
            "[%d] TrnSet:  MSE: %.6f RMSE: %.6f NRMSE: %.6f RNMSE: %.6f MAE: %.6f",
            epoch,
            trainingSetMSE,
            trainingSetRMSE,
            trainingSetNRMSE,
            trainingSetRNMSE,
            trainingSetMAE
        ));
    }

    @Override
    public void preProcessingComplete(int numInputs, long totalTime) {
        totalInputsProcessed += numInputs;
        preProcessingTime.addValue(totalTime);
    }

    @Override
    public void selfModification(int funcType) {
        selfModificationTypeCount.putIfAbsent(funcType, 0);
        selfModificationTypeCount.put(funcType, selfModificationTypeCount.get(funcType) + 1);
    }

    @Override
    public void iterationComplete(int numActiveNodes, boolean hasActiveInput, int activeNodesHashCode) {
        activeNodesSize.addValue(numActiveNodes);
        activeNodeStates.add(activeNodesHashCode);
        activeNodeStateHistory.add(activeNodesHashCode);

        if (hasActiveInput) {
            inputPresentAfterIteration++;
        }
    }
}
